/*
   (C) Copyright 1996  Michael Stenns
   
   Permission to use, copy, modify, and distribute this program without fee 
   is hereby granted, provided that this copyright and permission notice 
   appear on all copies and supporting documentation.

   All sourcefiles starting with "gd" are part or Thomas Boutell's graphic
   library gd1.2. gd 1.2 is copyright 1994, 1995, Quest Protein Database 
   Center, Cold Spring Harbor Labs.

   If you want use the gd1.2 library, remark that the files provided with
   this package are slightly modified. You should get the original set from
   http://www.boutell.com/gd/ .

*/
/***************************************************************************/

/*
 * this program analyzes lines from a standard httpd log file
 * (common log file format)
 *
*/

#define MAIN_MODUL        /* needed in accstat.h */


#include "accstat.h"


/* string constants */


/* declarations */


/* prototypes */


/***************************************************************************
 * Main program
 ***************************************************************************/

int main (int argc, char *argv[])
{
CommandLineType * pCommandLineOptions = NULL;
ConfigFileType  * pConfigFileOptions  = NULL;
StandardListType * StandardLists;

 eval_cpu_time (NULL); /* init time */
 StandardLists = InitStandardLists ();
 pCommandLineOptions = ParseCommandLine (argc, argv);
 pConfigFileOptions = ReadConfigFile (pCommandLineOptions, StandardLists);
 ReadHistoryFile (pConfigFileOptions, StandardLists);
 AdjustFromHistory (StandardLists, pCommandLineOptions, pConfigFileOptions);
 CreateLists (StandardLists, pCommandLineOptions);
 CreateSpecialLists (StandardLists, pCommandLineOptions, pConfigFileOptions);
 CreateImages (StandardLists, pCommandLineOptions, pConfigFileOptions);
 CreateOutputFiles (StandardLists, pCommandLineOptions, pConfigFileOptions);
  
 WriteHistoryFile (pConfigFileOptions, StandardLists);
 SummeryReport (StandardLists, pCommandLineOptions);
 printf ("\n cpu time used: %.1f seconds\n",((double)eval_cpu_time(""))/CLOCKS_PER_SEC);

 exit (EXIT_SUCCESS);
 return 0;
}





/***************************************************************************
 * CreateLists reads the specified logfiles and creates
 * the standard lists (date, site, item, method, protocol and status)
 ***************************************************************************/
void CreateLists (StandardListType * pSList, 
                  CommandLineType * pCmdLinOptions)
{
FILE *AccesslogFile;                /* httpd logfile */
char *pszRecord, *pszRecordCopy = NULL;     /* line from httpd logfile */
RecordParseType *pRecParse; 
int iDateNum = 0;               /* index of current date */
int iCurrentFile;

/*
 * check, if logfiles specified
*/
 if (!pCmdLinOptions->NumLogFiles)
 {
   fputs (ACCFILEFAIL, stderr); /* no logfile specified */
   fprintf (stderr, INIT_FAIL, DEF_QU);
   exit (EXIT_FAILURE);
 } 
/*
 * initialize list's
*/

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));
 if (pCmdLinOptions->ExtractFile != NULL)  /* extract to file enabled */
 {
    pszRecordCopy = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));
 }
 pRecParse = my_calloc (1, sizeof (RecordParseType));
 pRecParse->BeginPeriod = pCmdLinOptions->since;
 pRecParse->EndPeriod = pCmdLinOptions->before;

/*
 * start of the logfile's read loop
*/
 for (iCurrentFile = 0;
      iCurrentFile < pCmdLinOptions->NumLogFiles;
      iCurrentFile++)
 {
 AccesslogFile = fopen (pCmdLinOptions->LogFileList[iCurrentFile], 
                        LOGFILE_OPEN_MODE);
 if (AccesslogFile == NULL)
 {
      fputs (ACCFILEFAIL, stderr);
      exit (EXIT_FAILURE);
 }

/*
 * now read the logfile
*/

 while (fgets(pszRecord,DEFAULT_RECORD_LENGTH,AccesslogFile) != NULL)
 {
  pSList->NumRecord++;
  if (pszRecordCopy) strcpy (pszRecordCopy, pszRecord);
  if (!pCmdLinOptions->iCaseSensitive) StringToLower (pszRecord);
  if (ParseRecord (pszRecord, pRecParse) && 
      CheckCommandLists (pRecParse, pCmdLinOptions))
  {
    /* copy all valid lines to extract file */
    if (pszRecordCopy) fputs (pszRecordCopy,pCmdLinOptions->ExtractFile);
    /* convert sites allways to lowercase */
    if (pCmdLinOptions->iCaseSensitive)  StringToLower (pRecParse->pszSite);

    iDateNum = DateListInsert (pSList->DateList, pRecParse->tDate, pRecParse->Bytes);
    StringListInsert (pSList->MethodList, pRecParse->pszMethod,iDateNum);
    StringListInsert (pSList->ProtocolList, pRecParse->pszProtocol,iDateNum);
    StringListInsert (pSList->StatusList, pRecParse->pszStatusCode,iDateNum);
    StringListInsert (pSList->AgentList, pRecParse->pszAgent,iDateNum);
    StringListInsert (pSList->ItemList, pRecParse->pszItem,iDateNum);
    StringListInsert (pSList->SiteList, pRecParse->pszSite,iDateNum);
    TimeListInsert (pSList->TimeList, pRecParse->pszTime);
  }
 }

 fclose (AccesslogFile);
 }    /* end of the logfile's read loop */
 free(pszRecord);
 free(pRecParse);
 if (pCmdLinOptions->ExtractFile) fclose (pCmdLinOptions->ExtractFile); 
#if COUNT_ALLOC_MEMORY >= 2
 eval_cpu_time ("elapsed time in scanning logfiles");
#endif
 if (!pSList->SiteList->NumMembers)
 {
   fprintf (stderr, CREATELIST_NONE_READ);
   exit (EXIT_SUCCESS);
 }
 return;
}


/**************************************************************************
 *
 * checks the record stored in pszRecord against option read from
 * the command line.
 *
 *************************************************************************/

int CheckCommandLists (RecordParseType *pszRecord, CommandLineType *pOptions)
{
 int Valid = FALSE;
 int iSite=2,iMethod=2,iItem=2,iProtocol=2,iStatus=2,iAgent=2;
 int i;



/*
 *  check against include list
*/
 for (i=0;i < pOptions->NumIncludeMembers; i++)
 {
   switch (pOptions->IncludeList[i]->Key)
   {
    case SITE:
       if (iSite)
       {
         iSite = wild_strcmp (pszRecord->pszSite,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case METHOD:
       if (iMethod)
       {
         iMethod = wild_strcmp (pszRecord->pszMethod,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case ITEM:
       if (iItem)
       {
         iItem = wild_strcmp (pszRecord->pszItem,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case PROTOCOL:
       if (iProtocol)
       {
         iProtocol = wild_strcmp (pszRecord->pszProtocol,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case STATUS:
       if (iStatus)
       {
         iStatus = wild_strcmp (pszRecord->pszStatusCode,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case AGENT:
       if (iStatus)
       {
         iAgent = wild_strcmp (pszRecord->pszAgent,
                          pOptions->IncludeList[i]->Val.pszString);
       }
       break;
    case NO_KEY:
       break;
   }
 }
 if (iSite == 2) iSite = 0;
 if (iMethod == 2) iMethod = 0;
 if (iItem == 2) iItem = 0;
 if (iProtocol == 2) iProtocol = 0;
 if (iStatus == 2) iStatus = 0;
 if (iAgent == 2) iAgent = 0;

 if (!iSite && !iMethod && !iItem && !iProtocol && !iStatus && !iAgent) Valid = TRUE;

/*
 *  check against exclude list
*/

 for (i=0; Valid && (i < pOptions->NumExcludeMembers); i++)
 {
  switch (pOptions->ExcludeList[i]->Key)
  {
   case SITE:
      if (!wild_strcmp (pszRecord->pszSite,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case METHOD:
      if (!wild_strcmp (pszRecord->pszMethod,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case ITEM:
      if (!wild_strcmp (pszRecord->pszItem,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case PROTOCOL:
      if (!wild_strcmp (pszRecord->pszProtocol,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case STATUS:
      if (!wild_strcmp (pszRecord->pszStatusCode,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case AGENT:
      if (!wild_strcmp (pszRecord->pszAgent,
                         pOptions->ExcludeList[i]->Val.pszString))
      {
         Valid = FALSE;
      }
      break;
   case NO_KEY:
      break;
  }
 }

/*
 *  check against alias list
*/

 if (Valid)
 {
  for (i = 0; (i < pOptions->NumAliasMembers); i++)
  {
   switch (pOptions->AliasList[i]->Key)
   {
    case SITE:
       if (!wild_strcmp (pszRecord->pszSite,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszSite = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case METHOD:
       if (!wild_strcmp (pszRecord->pszMethod,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszMethod = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case ITEM:
       if (!wild_strcmp (pszRecord->pszItem,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszItem = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case PROTOCOL:
       if (!wild_strcmp (pszRecord->pszProtocol,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszProtocol = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case STATUS:
       if (!wild_strcmp (pszRecord->pszStatusCode,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszStatusCode = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case AGENT:
       if (!wild_strcmp (pszRecord->pszAgent,
                          pOptions->AliasList[i]->Val.pszString))
       {
         pszRecord->pszAgent = pOptions->AliasList[i]->pszAlias;
       }
       break;
    case NO_KEY:
       break;
   }
  }
 }

 return Valid;
}

/* *************************************************************************
 * ParseRecord parses a record from the so called Common Logfile Format.
 * It returns pointers to the interesting parts of the string in 
 * the structur pRecParse.
 * Expected format of the record: 
 *
 *    remotehost rfc931 authuser  [date] "request" status bytes "referer" "user-agent"
 *
 * Part of the Record      Returnvalue
 * ===============================================
 *  remotehost             pRecParse->pszSite
 *  rfc931                          -
 *  authuser                        -
 *  [date]                 pRecParse->tDate, ->pszTime
 *  "request"              pRecParse->pszMethod, ->pszItem, ->pszProtocol
 *  status                 pRecParse->pszStatusCode 
 *  bytes                  pRecParse->Bytes
 *  "referer"              pRecParse->pszReferer 
 *  "user -agent"          pRecParse->pszAgent 
 *
 *************************************************************************/

int ParseRecord (char * Record, RecordParseType * pRecParse)
{
static char *pszMonth = "JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC";
static char  szCode[4];
int iDay, iMonth, iYear;
char *pszTemp, *pszDate, *pszRest;
int i;


 /* now parse Record */
  pRecParse->pszSite = Record;      /* remotehost rfc931 authuser */
  if (!(pszDate = strchr (Record,  '['))) return FALSE;   /* [date] */
  /* "request" status bytes "referer" "user-agent" */
  if (!(pszRest = strchr (pszDate, ']'))) return FALSE;  

  *pszDate = *pszRest = '\0';
  pszDate++; pszRest++;
  
  pRecParse->pszTime = strchr (pszDate, ':');
  if (!pRecParse->pszTime) return FALSE;
  *pRecParse->pszTime = '\0'; pRecParse->pszTime++;

  /* now parse the date string (dd/mmm/yyyy) */
  if (!(pszTemp = strtok (pszDate, "/"))) return FALSE;  /* dd */
  iDay = atoi (pszTemp);

  if (!(pszDate = strtok (NULL, "/"))) return FALSE;      /* mmm */

  if (!(pszTemp = strtok (NULL, "/"))) return FALSE;     /* yyyy */
  iYear = atoi (pszTemp);

  for (pszTemp=pszDate; *pszTemp; pszTemp++) *pszTemp = toupper (*pszTemp);
  if (!(pszTemp = strstr(pszMonth,pszDate))) return FALSE;
  iMonth = (pszTemp - pszMonth)/3 + 1;

  pRecParse->tDate = DaysSince1970 ( iDay,  iMonth, iYear) *  SECONDS_OF_A_DAY;

 /*
  *  check time interval
  */

 if ( (pRecParse->tDate < pRecParse->BeginPeriod ) || 
      (pRecParse->tDate >= pRecParse->EndPeriod) ) return FALSE ;

  /* parse the rest */
  pszTemp = strchr (pRecParse->pszSite, ' ');
  if (pszTemp) *pszTemp = '\0';
  pszTemp = strchr (pRecParse->pszTime, ' ');
  if (pszTemp) *pszTemp = '\0';

  /* pszRest = "request" status bytes "referer" "user-agent" */
  pszRest = Truncate_String (pszRest);
  i = strlen (pszRest);
  if (pszRest[--i] == '\"')                         /* combined logfile format */
  {
    pszRest[i] = '\0';
    if (!(pszTemp = strrchr (pszRest, '\"'))) return FALSE; *pszTemp = '\0';
    pRecParse->pszAgent = pszTemp + 1;

    if (!(pszTemp = strrchr (pszRest, '\"'))) return FALSE; *pszTemp = '\0';
    if (!(pszTemp = strrchr (pszRest, '\"'))) return FALSE; *pszTemp = '\0';
    pRecParse->pszReferer = pszTemp + 1;
  }
  else                                              /* common logfile format */
  {
    pRecParse->pszReferer = pRecParse->pszAgent = NULL;
  }
  /* pszRest = "request" status bytes */
  pRecParse->pszMethod = pszRest + 1; /* "request" = method item protocol */
  if (!(pszTemp = strrchr (pszRest, '\"'))) return FALSE; *pszTemp = '\0';

  pszRest = pszTemp + 1; /* pszRest = status bytes */
  while (*pszRest && !isdigit(*pszRest)) pszRest++;
  if (!(pszTemp = strchr (pszRest, ' '))) return FALSE; 
  *pszTemp = '\0'; pszTemp++; /* bytes */

  i = abs (atoi(pszRest));  /* status */
  sprintf (szCode,"%3.3u",i);
  pRecParse->pszStatusCode = szCode;

  if (i >= 400)  /* parse rest only if status indicates successful transfer */
  {
    pRecParse->pszMethod   = NULL;        /* ignore */
    pRecParse->pszItem     = NULL;        /* ignore */
    pRecParse->pszProtocol = NULL;        /* ignore */
    pRecParse->Bytes       = 0;
    return TRUE;
  }

  pRecParse->Bytes = atoi (pszTemp);
  if (!(pszTemp = strchr (pRecParse->pszMethod, ' '))) return FALSE; 
  *pszTemp = '\0';
  pRecParse->pszItem = pszTemp + 1;

  if (!(pszTemp = strrchr (pRecParse->pszItem, ' ')))
  {
    pRecParse->pszProtocol = "http/0.9";
  }
  else
  {
    *pszTemp = '\0';
    pRecParse->pszProtocol = pszTemp + 1;
    if (!*pRecParse->pszProtocol) pRecParse->pszProtocol = "http/0.9";
  }
  
  /* remove search string from item */
  if ((pszTemp = strchr (pRecParse->pszItem, '?'))) *pszTemp = '\0';

  return TRUE;
}


/**************************************************************************
 *************************************************************************/
void usage ()
{
/* CONFIG,EXTRAKT,OUTPUT,INCLUDE,EXCLUDE,SINCE,BEFORE,SHOW */
char *Qualifier[]={QUALIFIER_LIST,NULL};

 printf (" \nusage: \n\n  accstat qualifier parameter\n\n");
 printf ("  valid qualifier are:\n\n");
 printf (" %s%s=filename                 (default: none)\n",DEF_QU,Qualifier[CONFIG]);
 printf (" %s%s=filename                       (default: none)\n",DEF_QU,Qualifier[EXTRAKT]);
 printf (" %s%s=filename                        (default: none (=stdout))\n",DEF_QU,Qualifier[OUTPUT]);
 printf (" %s%s=key=val1,val2,key=val3,val4... (default: *)\n",DEF_QU,Qualifier[INCLUDE]);
 printf (" %s%s=key=val1,val2,key=val3,val4... (default: none)\n",DEF_QU,Qualifier[EXCLUDE]);
 printf (" %s%s=date                             (default: 01-Jan-1970)\n",DEF_QU,Qualifier[SINCE]);
 printf (" %s%s=date                            (default: today)\n",DEF_QU,Qualifier[BEFORE]);
 printf (" %s%s=all={a,nn},key={a,nn},...         (default: none)\n",DEF_QU,Qualifier[SHOW]);
 printf ("  valid keys are  site, item, status, method or protocol\n");
 printf ("  valid dates are dd-mmm-jjjj, mm/dd/jjjj, dd.mm.jjjj,\n");
 printf ("                  ddd (delta time), yesterday, today or tomorrow\n\n");
 printf ("  valid parameter are:   logfile1 logfile2 ...\n");
 exit (EXIT_FAILURE);
} 
