#ifndef ACCSTAT_H
#define ACCSTAT_H

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define ACCSTAT_VERSION "1.0"

#define TRUE 1
#define FALSE 0

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC 100
#endif

#ifndef COUNT_ALLOC_MEMORY
#define COUNT_ALLOC_MEMORY 2
#endif

#ifndef DAY_OFFSET /* defines first day of a week (1-Jan-1970 was a thursday) */
#define DAY_OFFSET 3 /* friday = 6, saturday = 5, ..., thursday = 0 */
#endif

/* list of valid qualifiers
 *
 * for each member in "QUALIFIER_LIST" exists one correspondenting member
 * in "enum QUALIFIER"
*/

#if 0
#ifdef GERMAN
#define QUALIFIER_LIST "konfiguration","extrakt","ausgabe",\
"mit","ohne","bestaetigen",\
"seit","vor","zeige","bestaetigung","hilfe"
#endif
#endif

#ifndef QUALIFIER_LIST
#define QUALIFIER_LIST "configuration","extract","output",\
"include","exclude",\
"since","before","show","help"
#endif

/* string constants */


#ifdef GERMAN
#define AMBIGUOUS_WARNING "Warnung, Qualifier \"%s\" ist nicht eindeutig\n"
#define UNKNOWN_WARNING "Warnung, Qualifier \"%s\" ist nicht bekannt\n"
#define CONFIG_FILE_OPENFAIL "Fehler, Konfigurationsdatei \"%s\" nicht gefunden\
oder kann nicht geoeffnet werden\n"
#define EXTRAKT_FILE_OPENFAIL "Warnung, kann \"%s\" nicht zum Schreiben oeffnen\n"
#define LOG_FILE_OPENERR "Warnung, kann Logdatei \"%s\" nicht zum Lesen oeffnen\n"
#define ACCFILEFAIL "Fehler, Logdatei kann nicht geoeffnet werden\n"
#define ALLOCFAIL   "Fehler, Zuwenig Speicher\n"
#define DATE_CHECK_FAIL "Fehler, Datum in \"before\" Option muss juenger als \
in der \"since\" Option sein\n" 
#define INIT_FAIL "Info,  \"accstat %shelp\" fuer Erlaeuterung der Qualifier eingeben\n"
#define THOUSANDS_SEP '.'
#endif


#ifndef PLURAL_SIGN
#define PLURAL_SIGN "s"
#endif

#ifndef THOUSANDS_SEP
#define THOUSANDS_SEP ','
#endif

#ifndef PERIOD_PLACEOLDER  /* original date string placeholder */
#define PERIOD_PLACEHOLDER "!nn!"
#endif

#ifndef PERIOD_REPLACEMENT           /* needs to be a 6 byte string */
#define PERIOD_REPLACEMENT "!nnnn!"
#endif

#ifndef DEFAULT_TABLE_LENGHT
#define DEFAULT_TABLE_LENGHT 5
#endif

#ifndef DEFAULT_CLICKABLE_LIST
#define DEFAULT_CLICKABLE_LIST "*.htm*,*/"
#endif

#ifndef ACCFILEFAIL
#define ACCFILEFAIL "ERROR,   cannot open logfile\n"
#endif

#ifndef ALLOCFAIL
#define ALLOCFAIL "ERROR,   insufficient memory\n"
#endif

#ifndef AMBIGUOUS_WARNING
#define AMBIGUOUS_WARNING "WARNING, ambiguous qualifier \"%s\"\n"
#endif

#ifndef UNKNOWN_WARNING
#define UNKNOWN_WARNING "WARNING, unknown qualifier \"%s\"\n"
#endif

#ifndef CONFIG_FILE_OPENFAIL
#define CONFIG_FILE_OPENFAIL "ERROR,   configuration file \"%s\" inaccessible\n"
#endif

#ifndef EXTRAKT_FILE_OPENFAIL
#define EXTRAKT_FILE_OPENFAIL "WARNING, cannot open \"%s\" for write\n"
#endif

#ifndef EXTRAKT_FILE_EXIST
#define EXTRAKT_FILE_EXIST "ERROR,   extract file \"%s\" has the wrong format\n"
#endif

#ifndef EXTRACT_FILE_STARTLINE
#define EXTRACT_FILE_STARTLINE "#Extract file from httpd logfile ACCSTAT Ver. 1.0\n"
#endif

#ifndef LOG_FILE_OPENERR
#define LOG_FILE_OPENERR "WARNING, cannot open logfile \"%s\" for read\n"
#endif

#ifndef OUTPUT_REOPEN_FAIL
#define OUTPUT_REOPEN_FAIL "ERROR,   cannot redirect output to \"%s\"\n"
#endif

#ifndef INTERNAL_ERROR_FATAL
#define INTERNAL_ERROR_FATAL "sorry, internal program error\n"
#endif

#ifndef CREATELIST_NONE_READ
#define CREATELIST_NONE_READ "WARNING, no valid record read\n"
#endif

#ifndef DATE_CHECK_FAIL
#define DATE_CHECK_FAIL "ERROR,   before date may not be lower or equal than since date\n"
#endif


#ifndef INIT_FAIL
#define INIT_FAIL "INFO,    try \"accstat %shelp\" for help on qualifiers\n"
#endif

#ifndef SUMMARY_TIME_FORMAT
#define SUMMARY_TIME_FORMAT "%a %b %d %Y"
#endif

#ifndef SUMMARY_REPORT_TITLE_FORMAT
#define SUMMARY_REPORT_TITLE_FORMAT "\n Summary Report from %s until %s\n"
#endif

#ifndef SUMMARY_REPORT_TITLE_1DAY_FORMAT
#define SUMMARY_REPORT_TITLE_1DAY_FORMAT "\n Summary Report from %s\n"
#endif

#ifndef CONFIG_SYNTAX_MISSING_ENDSECTION
#define CONFIG_SYNTAX_MISSING_ENDSECTION \
"WARNING, missing end of section \"%s\" in configuration file\n"
#endif

#ifndef SPECIALLIST_SYNTAX_ERROR
#define SPECIALLIST_SYNTAX_ERROR "WARNING, Syntax error in definition of list \"%s\"\n"
#endif

#ifndef DEFINETABLE_SYNTAX_ERROR
#define DEFINETABLE_SYNTAX_ERROR "WARNING, Syntax error in definition of table \"%s\"\n"
#endif

#ifndef DEFINETABLE_TYPE_ERROR
#define DEFINETABLE_TYPE_ERROR "         \"%s\" is neither a special nor a standard list\n"
#endif

#ifndef OUTPUTFILE_SYNTAX_ERROR
#define OUTPUTFILE_SYNTAX_ERROR  "WARNING, Syntax error in definition of output file \"%s\"\n"
#endif

#ifndef CONFIG_ERROR_NOSECTION
#define CONFIG_ERROR_NOSECTION  "         Section was not found or is empty\n"
#endif

#ifndef SPECIALLIST_INVALID_KEY
#define SPECIALLIST_INVALID_KEY "         \"%s\" is not a valid key\n"
#endif

#ifndef AMBIGUOUS_KEY_WARNING
#define AMBIGUOUS_KEY_WARNING "WARNING, ambiguous key \"%s\"\n"
#endif

#ifndef OUTFIL_OPEN_ERROR
#define OUTFIL_OPEN_ERROR "WARNING, can't open output file \"%s\"\n"
#endif

#ifndef OUTFIL_SYNTAX_ERROR
#define OUTFIL_SYNTAX_ERROR "WARNING, syntax error in line %u of file \"%s\"\n"
#endif

#ifndef OUTFIL_TABLE_NOSPEC
#define OUTFIL_TABLE_NOSPEC "         no tablename specified \n"
#endif

#ifndef DEFINETABLE_NOLAYOUT
#define DEFINETABLE_NOLAYOUT "         missing layout information\n"
#endif

#ifndef DEFINETABLE_VALUE_CHECK
#define DEFINETABLE_VALUE_CHECK "         checking condition \"%s\" on value \"%s\" \n"
#endif

#ifndef OUTFIL_MAXLEN_ERROR
#define OUTFIL_MAXLEN_ERROR "ERROR,  lenght of path \"%s\" and filename \"%s\" \
is graeter than the allowed value of %u\n"
#endif

#ifndef OUTFILE_UNKNOWN_TAG
#define OUTFILE_UNKNOWN_TAG "WARNING unknown Tag \"%s\" with Value \"%s\"\n"
#endif

#ifndef HISTFILE_OPENWRITE_ERROR
#define HISTFILE_OPENWRITE_ERROR "WARNING, cannot open history file \"%s\" for write\n"
#endif

#ifndef HISTORY_FILE_STARTLINE
#define HISTORY_FILE_STARTLINE "# Historyfile ACCSTAT Ver. 1.0\n"
#endif

#ifndef HISTORY_FILE_TIMEFORMAT
#define HISTORY_FILE_TIMEFORMAT "%d-%b-%Y\0"
#endif

#ifndef M_LOG10E 
#define M_LOG10E        0.43429448190325182765
#endif

#ifndef SUNDAY
#define SUNDAY "Sunday"
#endif

#ifndef MONDAY
#define MONDAY "Monday"
#endif

#ifndef TUESDAY
#define TUESDAY "Tuesday"
#endif

#ifndef WEDNESDAY
#define WEDNESDAY "Wednesday"
#endif

#ifndef THURSDAY
#define THURSDAY "Thursday"
#endif

#ifndef FRIDAY
#define FRIDAY "Friday"
#endif

#ifndef SATURDAY
#define SATURDAY "Saturday"
#endif

#define DEFAULT_RECORD_LENGTH 1024

#define DEFAULT_SITE_NUM 2048
#define DEFAULT_DATE_NUM 40
#define DEFAULT_DATE_EXTEND 21
#define DEFAULT_TIME_NUM 1
#define DEFAULT_METHOD_NUM 10
#define DEFAULT_ITEM_NUM 1024
#define DEFAULT_PROTOCOL_NUM 10
#define DEFAULT_STATUS_NUM 20
#define DEFAULT_AGENT_NUM 20
#define STRINGLIST_SORT_INTERVAL 200
#define TIMLIST_DIM (24*6)

#define MEMPAGE_EXTEND    1024
#define MEMPAGE_PAGESIZE (1024 * 32 * sizeof (char))

#define SECONDS_OF_A_DAY    86400      /* (60 * 60 * 24) */
#define SECONDS_OF_A_WEEK  604800      /* (60 * 60 * 24 * 7) */

#define DEFAULT_SECTIONLIST_EXTEND 20

#ifndef DEFAULT_LOGFILE_INCREMENT
#define DEFAULT_LOGFILE_INCREMENT 10
#endif

#ifndef DEFAULT_SPECIALLIST_EXTEND
#define DEFAULT_SPECIALLIST_EXTEND 10
#endif

#define DEFAULT_COUNT_AR_DIM 12

#ifndef SUCCESS
#define SUCCESS 0
#endif

#define ERR_UNKNOWNTAG  1
#define ERR_NOSECTION   2
#define ERR_NOTABLE     3
#define ERR_INVALIDLIST 4

/**************************************************************************
 * define system or compiler specific stuff here
 *************************************************************************/


#if !defined UNIX && !defined __DJGPP__
#if  defined unix || defined __unix || defined __unix__ 
#define UNIX
#endif
#endif

#ifdef UNIX     
#define DEF_QU "-"
#define CASE_SENSITIVE 1
#else
#define DEF_QU "/"
#define CASE_SENSITIVE 0
#endif

#ifdef VAXC
typedef double longDouble;
#else
typedef long double longDouble;
#endif

#if defined VMS && (defined VAXC || defined __GNUC__)
#if FILENAME_MAX < 60 /* the VAX/VMS default value is quite short */
#undef FILENAME_MAX
#define FILENAME_MAX 80
#endif
#endif

#ifndef LOGFILE_OPEN_MODE
#ifdef VMS
#ifdef __ALPHA
#define LOGFILE_OPEN_MODE "r","shr=upd","mbc=64","mbf=64"
#else  /* not __ALPHA */
#define LOGFILE_OPEN_MODE "r","shr=upd","mbc=32","mbf=32"
#endif /* __ALPHA */
#else  /* not VMS */
#define LOGFILE_OPEN_MODE "r"
#endif /* VMS */
#endif /* LOGFILE_OPEN_MODE */

typedef unsigned int uInt;
typedef unsigned long int uLong;

/**************************************************************************
 * end of system or compiler specific stuff
 *************************************************************************/

/**************************************************************************
 * declarations
 *************************************************************************/
#define KEYS_NUM_MEMBERS 7
#ifndef KEYSTRINGS
#define KEYSTRINGS "site","method","item","protocol","status","agent"
#define MAXKEYLEN sizeof "protocol"
#endif
#ifndef MAXKEYLEN
#define MAXKEYLEN 10
#endif
enum KEYS {SITE, METHOD, ITEM, PROTOCOL, STATUS, AGENT, NO_KEY};

enum QUALIFIER {CONFIG,EXTRAKT,OUTPUT,INCLUDE,EXCLUDE,
                SINCE,BEFORE,SHOW,HELP,NO_QUALIFIER,PARAMETER};
enum DATE_MODE {NO_DATE, PER_DAY, PER_WEEK, PER_MONTH};
enum TABLE_TYPE {UNKNOWN_TABLE, STRINGLIST_TABLE, DATELIST_TABLE};
enum SORT_TYPE  {SORT_NUMERICAL, SORT_ALPHABETICAL, SORT_NOSORT};

typedef enum KEYS KeyType;
typedef enum DATE_MODE DateType;
typedef enum SORT_TYPE SortType;

typedef struct {uLong low;
                uLong high;
               }ByteCountType;


typedef struct {char * pszStr;                /* pointer to the string */
                uInt  * uiCountAr;            /* array of counters, 
                                                 one for each time interval */
                uInt Count;                   /* contains the count for a 
                                                 evaluated period */
               } StringCountType;


typedef struct {size_t NumMembers;    /* number of List members */
                size_t MaxMembers;    /* current size of List */
                size_t DefaultExtend; /* default extend of List */
                size_t DimCountAr;    /* dimension of counter array */
                char * Name;          /* Pointer to the name if the list */
                KeyType Key;          /* kind of data */
                SortType Sort;        /* numerical or alphabetical */
                int ShowMembers;      /* number to output */
                int    iLast;         /* used only in StringListInsert */
                StringCountType ** List;    /* pointer to List */
               } StringListType;


typedef struct {size_t NumMembers;    /* number of List members */
                size_t MaxMembers;    /* current size of List */
                size_t DefaultExtend; /* default extend of List */
                int Offset;           /* defines first day of a week */
                time_t * Date;        /* array of Date values*/
                uInt * uiCountAr;     /* hits per day */
                uInt * uiByteAr;      /* bytes per day */
               } DateListType;


union ValueUnion { char * pszString;
                   unsigned int uiNum;
                 };

typedef struct { KeyType Key;             /* kind of data */
                 union ValueUnion Val;    /* either a string or a number */
               } ConfigListType;

typedef struct { KeyType Key;             /* kind of data */
                 union ValueUnion Val;    /* either a string or a number */
                 char * pszAlias;         /* Pointer to the alias name */
               } AliasListType;


typedef struct { char * Token;            /* string, may contain wildcarts */
                 char * Name;             /* points to the appropriate name */
               } SpecialListType;


typedef struct { FILE * ConfigFile;       /* configuration file */
                 FILE * ExtractFile;      /* file to extract to */
                 char * pszOutputFile;    /* name of the output file */
                 int    iCaseSensitive;   /* true for unix, else false */
                 size_t NumIncludeMembers;
                 size_t NumExcludeMembers;
                 size_t NumAliasMembers;
                 size_t NumHideMembers;
                 size_t NumLogFiles;
                 ConfigListType ** IncludeList;
                 ConfigListType ** ExcludeList;
                 AliasListType ** AliasList;
                 ConfigListType ** HideList;  /* do not show in output list */
                 char ** AliasNameList;   /* list of the alias names */
                 char ** AliasUrlList;    /* list of the substitution urls */
                 char ** LogFileList;     /* names of all logfiles specified */
                 time_t since;            /* oldest valid date */
                 time_t before;           /* newest valid date */
                 SortType Sort[KEYS_NUM_MEMBERS + 1];   /* how to sort */
                 int ShowMembers[KEYS_NUM_MEMBERS + 1]; /* entrys in ouput table */
                 enum DATE_MODE DateMode; /* defines kind of table for date list */
               } CommandLineType;


typedef struct { char * Name;        /* name of the section in conf file */
                 KeyType Key;        /* kind of data */
                 long int Offset;    /* offset reportet by ftell */
                 size_t Lenght;      /* number of lines in section */
               } SectionDescriptType;

typedef struct { char * SpecialList; /* name of the special list or NULL*/
                 SectionDescriptType * FileSection;
               } TableDescriptType;

typedef struct { char * FileName;
                 enum DATE_MODE Period;
                 uInt MaxFiles;
                 SectionDescriptType * FileSection;
               } OutputFileDescriptType;


typedef struct { char * OutPath;                  /* path files are written to */
                 char * HistoryFile;              /* history file */
                 size_t NumSpecialLists;          /* number of special lists */
                 size_t NumLayoutLists;           /* number of table layout sections */
                 size_t NumOutFiles;              /* number of output files  */
                 size_t NumImages;                /* number of images to create */
                 size_t NumTables;                /* number of requested tables */
                 SectionDescriptType ** SpecialLists;   /* list of special lists */
                 SectionDescriptType ** LayoutList;     /* list of table layouts */
                 OutputFileDescriptType ** OutFileList; /* list of ouput files */
                 OutputFileDescriptType ** ImageList;   /* list of images to create */
                 TableDescriptType ** TableList;  /* list of requestet tables */
                 enum DATE_MODE DateMode;         /* max. requested Period */
               } ConfigFileType;

typedef struct {DateListType * DateList;  /* bytes, counts per day from logfile */
                DateListType * HistoryDateList; /* same from history file */
                StringListType * SiteList;
                StringListType * ItemList;
                StringListType * MethodList;
                StringListType * ProtocolList;
                StringListType * StatusList;
                StringListType * AgentList;
                StringListType * WeekDayList;
                StringListType ** SpecialLists; /* additional lists specified 
                                                   in the configuration file  */
                uInt             TimeList[TIMLIST_DIM]; /* 10 min resolution */
                size_t NumSpecialLists;
                unsigned int NumRecord;         /* number of records read */
               } StandardListType;



typedef struct { char * pszSite;
                 char * pszTime;
                 char * pszMethod;
                 char * pszItem;
                 char * pszProtocol;
                 char * pszStatusCode;
                 char * pszReferer;
                 char * pszAgent;
                 time_t tDate;
                 unsigned int Bytes;
                 time_t BeginPeriod;
                 time_t EndPeriod;
               } RecordParseType;


typedef struct {char * Tag;               /* describes insertion tag of type */
                char * Value;             /* <!--Accstat Tag = Value --> */
                char * RestOfRecord;      /* in output file defination */
               }TagInfoType; 

typedef struct { void * base;
                 void * ptr;
                 int    size;
               } MemPageType;



/* 
 * TableHeadInfoType contains the static infos known for a given table
 * while the infos in TableRowInfoType are only valid for one row of a table.
 */
typedef struct { enum TABLE_TYPE Type;
                 char * Name;
                 char * DateFormat;
                 size_t NumMembers;
                 size_t ValidMembers;
                 size_t ShowMembers;
                 uInt   TotalCount;
                 uInt   MaxCount;
                 ByteCountType MaxBytes;
                 ByteCountType TotalBytes;
               } TableHeadInfoType;

typedef struct { char * Item;
                 char * Url;
                 uInt   Current;
                 uInt   CurrentCount;
                 ByteCountType CurrentBytes;
               } TableRowInfoType;


/* global variables */

#if COUNT_ALLOC_MEMORY >= 1
#undef EXTERN
#ifdef MAIN_MODUL
#define EXTERN
#else
#define EXTERN extern
#endif /* MAIN_MODUL */
EXTERN unsigned int AllocMemoryCount;
EXTERN unsigned int AllocMemoryBlocks;
#define COUNT_ALLOC_MEMORY_FORMAT \
        "%u bytes memory in %u blocks allocated\n"
#endif /* COUNT_ALLOC_MEMORY */


/* prototypes */

/* 
 * =========================
 * functions in accstat.c 
 * =========================
*/

#undef EXTERN
#ifdef MAIN_MODUL
#define EXTERN
#else
#define EXTERN extern
#endif /* modul main */


void CreateLists (StandardListType * pSList, 
                  CommandLineType * pCommandLineOptions);

int CheckCommandLists (RecordParseType *pszRecord, CommandLineType *pOptions);



/* 
 * =========================
 * functions in modules.c 
 * =========================
*/

int ParseRecord (char *Record, RecordParseType *pRecParse);
int DaysSince1970 ( int Day, int Month, int Year);
void TimeListInsert (uInt TimeList[TIMLIST_DIM], char * pszTime);
void StringListInsert (StringListType *pL, const char *pszStr, int iDateNum);
int  DateListInsert (DateListType *pDatL, time_t Date, unsigned int uiBytes);

int ItemListStrComp (const void *e1, const void *e2);
int StringListCountComp (const void *e1, const void *e2);
int DateListComp (const void *e1, const void *e2);
int StringListBsearchComp (const void *ckey, const void *celem);

enum QUALIFIER CheckQualifier(char * pszArg, char **  pszOption);
char * StringToLower (char * pszString);
char * Truncate_String (char * pszString);
char * InsertThousandsSeparator (unsigned long ulLow, unsigned long ulHigh);
unsigned int nDigits (unsigned long ulValue);
time_t DateFromString_t (const char * pszString);
int wild_strcmp (const char *s1, const char *s2);
int NumberOfPeriods (DateListType * DateList, enum DATE_MODE DateMode);
time_t MinDate (DateListType * DateList);
time_t MaxDate (DateListType * DateList);
time_t mkDay (struct tm *tptr);

CommandLineType * ParseCommandLine (int argc, char *argv[]);

void SetShowDefault (CommandLineType * pOptions, char * pszOption);
void SetShowKey (CommandLineType * pOptions, ConfigListType * Member);
StringListType * SelectStringList (StandardListType * pSList, 
                                          KeyType Key, char * pszSpecialList);

char * string_alloc (const char * pszString);
StringCountType * StringCountAlloc (StringCountType * pStringCount);
uInt * CountArrayAlloc (uInt * uiCountArray, size_t OldSize, size_t AddSize);
MemPageType * GetMemoryPage (void * ptr, size_t SizeAllocMem);

void *my_malloc (size_t size);    /* stops program on error */
void *my_calloc (size_t nmemb, size_t size); /* stops program on error */
void *my_realloc (void *oldptr, size_t size);
void *my_recalloc (void *oldptr, size_t old_size, size_t add_size);

char ** AllocCharList (void * oldptr, size_t *old, size_t add);
AliasListType ** AllocAliasList (void * oldptr, size_t *old, size_t add);
ConfigListType ** AllocConfigList (void * oldptr, size_t *old, size_t add);
SectionDescriptType ** AllocSectionDescriptList 
                          (void * oldptr, size_t *old, size_t add);
TableDescriptType ** AllocTableDescriptList 
                          (void * oldptr, size_t *old, size_t add);
OutputFileDescriptType ** AllocOutputFileDescriptList 
                          (void * oldptr, size_t *old, size_t add);
SpecialListType ** AllocSpecialList (void * oldptr, size_t *old, size_t add);
StringCountType ** AllocStringCountList (void * oldptr, size_t *old, size_t add);

void ByteCountAdd (ByteCountType * NumBytes, uLong low, uLong high);
void ByteCountMax (ByteCountType * MaxVal, ByteCountType * Val);
double ByteCountProz (ByteCountType * Val, ByteCountType * All);
void EvalDateListPeriod (int * Indices, uInt NumIndices, 
                                DateListType * DateList,
                                unsigned int * TotalCount, 
                                ByteCountType * TotalBytes);
uLong CountStringListMembers (StringListType * StringList, char * pszMask);
uLong StringListMaxValue (StringListType *pSL);
uLong StringListTotalCount (StringListType *pSL);
void StringListAdd (StringListType *pSL, int * Indices, int NumIndices,
                                                       uInt MaxIndices);
char * MakeDateString (time_t Date, enum DATE_MODE DateMode, int WeekOffset);
int PrintDate (char * Buffer, int BufLen, time_t Date, char * pszFormat, int WeekOffset);
clock_t eval_cpu_time (char * OutString);

/* 
 * =========================
 * functions in output.c 
 * =========================
*/

void SummeryReport (StandardListType * pStdList, CommandLineType * pCmdLinOpts);
void SummeryReportPrint (FILE * outfile, StandardListType * pStdList, 
                                                CommandLineType * pCmdLinOpts, 
                                                time_t StartTime, time_t EndTime);
void StringListPrint (FILE * outfile, StringListType *pSL,
                      unsigned int MaxWidth, ConfigListType ** Hidelist);
void DateListPrint (FILE * outfile, DateListType *pDateList,
                           uInt MaxRows, uInt MaxWidth, enum DATE_MODE DateMode);
char * ParseConfigLine (char * pszLine);
size_t NumValidStringListMembers (StringListType *  pStringList,
                     ConfigListType ** HideList,  size_t * MaxValidValue);
int CheckAgainstHideList (ConfigListType ** HideList, KeyType Key,
                                 char * pszString);
char * GetStatusCode (char * pszCode);
time_t GetPeriodIndices (int * Indices, size_t MaxIndices, size_t NumPeriod,
                      DateListType * DateList, enum DATE_MODE DateMode);
void GetStartEndPeriod (time_t CurrentDate, int NumPeriod, 
                       enum DATE_MODE DateMode, int WeekOffset,
                       time_t * StartPeriod, time_t * EndPeriod);
void StandardListsSetPeriod (StandardListType * pSList, int * Indices);
void ReadHistoryFile (ConfigFileType * pConfigOpts, StandardListType * pSList);
void WriteHistoryFile (ConfigFileType * pConfigOpts, StandardListType * pSList);
int ParseOutFileRecord (char * pszString, TagInfoType * pTagInfo);
TableDescriptType * CheckTableTag (size_t NumTables, 
                                   TableDescriptType ** TableList, 
                                   char * pszTableName);
char * TagPosition (char * String, size_t * PosLen, size_t * TagLen, 
                    unsigned int * FieldWidth);
char * CheckUrlAlias(char * pszString, char * pszUrlPattern, int NumUrlPattern, 
                     char ** AliasNameList, char ** AliasUrlList);
char * CheckUrlFile (char * pszFilename, char * pszDate, char * pszOutPath);
int SetOutputFilename (uInt * pNumPeriod, uInt * pNumFiles, char * FileName,
                       time_t * StartPeriod, time_t * EndPeriod,
                       OutputFileDescriptType * FileDescript,
                       StandardListType * pSList);
void CreateOutputFiles (StandardListType * pSList, 
                       CommandLineType  * pCommandLineOptions,
                       ConfigFileType   * pConfigFileOptions);

int  PrintOutfilRecord (FILE * outfil, char * pszRecord,
                        time_t StartPeriod,  time_t EndPeriod,
                        StandardListType * pSList, 
                        CommandLineType  * pCmdLinOpts,
                        ConfigFileType   * pConfigOpts);

void InsertTable (FILE * outfil,
                  time_t StartPeriod,  time_t EndPeriod,
                  TagInfoType       * pTagInfo,
                  TableDescriptType * pTable,
                  StandardListType  * pSList, 
                  CommandLineType   * pCmdLinOpts,
                  ConfigFileType    * pConfigOpts);

void * InitTableInfo (TableHeadInfoType * Table,
                      StandardListType  * pSList,
                      ConfigListType   ** HideList,
                      char * pszSpecialList, KeyType Key,
                      int ShowMembers,
                      enum DATE_MODE DateMode);


void PrintTableLine (TableHeadInfoType * tblHead, TableRowInfoType * tblRow, 
                     char * outLine, char * inLine);

/* 
 * =========================
 * functions in config.c 
 * =========================
*/

ConfigFileType * ReadConfigFile (CommandLineType * pCmdLinOpts, 
                                        StandardListType * pSList);
SectionDescriptType ** CheckConfigFileSyntax (FILE * ConfigFile, 
                                     size_t * pMaxSections,  size_t MaxLines);
char * Truncate_Config_File_Record (char * pszRecord);
char * InsertAliasNameList (char ** AliasNameList, char ** AliasUrlList,
                                   char * pszString, char * pszUrl);
void InsertAliasList (AliasListType ** List, size_t * numelem,
                             KeyType Key, char * pszToken, char * pszAlias);
size_t ConfigFileSeek (FILE * ConfigFile, SectionDescriptType ** pzSections,
                              size_t NumSections, const char * pszName);
size_t CheckNumberOfTokens (char * pszString, char cSeparator);
int FillConfigListToken (char * pszToken, ConfigListType * ListMember);
KeyType TranslateStandardListKey (const char * pszKey);
void InsertConfigList (ConfigListType ** List, size_t * numelem,
                       ConfigListType * Member);
void CreateSpecialLists (StandardListType * pSList, 
                         CommandLineType  * pCommandLineOptions,
                         ConfigFileType   * pConfigFileOptions);
void FreeSectionDescriptTypeList (SectionDescriptType ** pzSections, 
                                                 size_t NumSections);
StandardListType *  InitStandardLists (void);
void AdjustFromHistory (StandardListType * pSList, 
                        CommandLineType  * pCmdLinOpts,
                        ConfigFileType   * pConfigOpts);
void usage (void);

/* 
 * =========================
   functions in graphics.c
 * =========================
*/

void CreateImages   (StandardListType * pSList, 
                     CommandLineType  * pCmdLinOpts,
                     ConfigFileType   * pConfigOpts);

#endif  /* ACCSTAT_H */
