/*
   (C) Copyright 1996  Michael Stenns
   
   Permission to use, copy, modify, and distribute this program without fee 
   is hereby granted, provided that this copyright and permission notice 
   appear on all copies and supporting documentation.

   All sourcefiles starting with "gd" are part or Thomas Boutell's graphic
   library gd1.2. gd 1.2 is copyright 1994, 1995, Quest Protein Database 
   Center, Cold Spring Harbor Labs.

   If you want use the gd1.2 library, remark that the files provided with
   this package are slightly modified. You should get the original set from
   http://www.boutell.com/gd/ .

*/
/***************************************************************************/
#define CONFIG_MODUL


#include "accstat.h"


/* string constants */

/* declarations */

/* prototypes */

ConfigListType ** FillConfigListFromSection (FILE * ConfigFile,
                                             ConfigListType ** ConfigList, 
                                             size_t * NumMembers, 
                                             int iCaseSensitive, 
                                             size_t SectionLenght);

DateListType * SetHistoryList (DateListType * DateList, 
                               DateListType * HistoryList);


/**************************************************************************
 * ReadConfigFile reads the configfile specified on the commandline
 *************************************************************************/
ConfigFileType * ReadConfigFile (CommandLineType * pCmdLinOpts, 
                                 StandardListType * pSList)
{
ConfigFileType * pConfigFileOpts  = NULL;     /* return value */
size_t MaxSpecialLists = 0;  /* current lenght of the SpecialLists
                                array in pConfigFileOpts */

SectionDescriptType ** pzSections = NULL; /* Null pointer terminated list */
size_t NumLines, NumValidLines;
size_t NumSections = 0;
size_t MaxLogfiles = pCmdLinOpts->NumLogFiles;
size_t MaxOutFiles = 0, MaxImages = 0, MaxTables = 0, MaxLayoutSections = 0;
size_t SectionIndex;
ConfigListType ListMember ;
char * pszRecord;   /* one line from the file */
char * pszToken, * pszAlias;
FILE * CheckFile;
int i = 0;

 /*
  * do some initialisations
  */

 pConfigFileOpts = my_calloc (1, sizeof (ConfigFileType));
 pConfigFileOpts->OutPath      = NULL;
 pConfigFileOpts->HistoryFile  = NULL;
 pConfigFileOpts->SpecialLists = NULL;
 pConfigFileOpts->LayoutList   = NULL;
 pConfigFileOpts->OutFileList  = NULL;
 pConfigFileOpts->ImageList    = NULL;
 pConfigFileOpts->TableList    = NULL;
 pConfigFileOpts->DateMode     = NO_DATE;

 if (!pCmdLinOpts->ConfigFile) return pConfigFileOpts; /* nothing to do */


 pzSections = CheckConfigFileSyntax (pCmdLinOpts->ConfigFile, &NumSections, 0);

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));


 /*
  * fill the Include, Exclude, Alias and Hide lists
  */

 /*
  *  Include list 
  */

 if ((NumLines = ConfigFileSeek(pCmdLinOpts->ConfigFile,pzSections,
                                NumSections,"IncludeList")))
 {
   pCmdLinOpts->IncludeList = FillConfigListFromSection 
                                  (pCmdLinOpts->ConfigFile, 
                                   pCmdLinOpts->IncludeList,
                                  &pCmdLinOpts->NumIncludeMembers,
                                   pCmdLinOpts->iCaseSensitive, NumLines);
 }
 
 /*
  *  Exclude list 
  */

 if ((NumLines = ConfigFileSeek(pCmdLinOpts->ConfigFile,pzSections,
                                NumSections,"ExcludeList")))
 {
   pCmdLinOpts->ExcludeList = FillConfigListFromSection 
                                  (pCmdLinOpts->ConfigFile, 
                                   pCmdLinOpts->ExcludeList,
                                  &pCmdLinOpts->NumExcludeMembers,
                                   pCmdLinOpts->iCaseSensitive, NumLines);
 }
 
 /*
  *  Alias list 
  *  line syntax: key = alias = val1,val2,... [= url]
  */

 if ((NumLines = ConfigFileSeek(pCmdLinOpts->ConfigFile,pzSections,
                                NumSections, "AliasList")))
 {
   char * pszKey;
   char * pszAlias;
   char * pszUrl;
   size_t NumTokens = 0;

   for (i = 0, NumValidLines = 1; i < NumLines; i++)
   {
     fgets (pszRecord, DEFAULT_RECORD_LENGTH, pCmdLinOpts->ConfigFile);
     pszKey   = Truncate_String(pszRecord);
     pszAlias = ParseConfigLine (pszKey);
     pszToken = ParseConfigLine (pszAlias);
     pszUrl   = ParseConfigLine (pszToken);
     if (pszKey && pszAlias && pszToken && strlen(pszToken))
     {
       NumValidLines++;
       NumTokens += CheckNumberOfTokens (pszToken,',');
     }
   }
   if (NumTokens)
   {
     size_t iTmp = pCmdLinOpts->NumAliasMembers;
     pCmdLinOpts->AliasList = AllocAliasList (pCmdLinOpts->AliasList, 
                                              &iTmp, NumTokens);
     iTmp = 0;
     pCmdLinOpts->AliasNameList = AllocCharList (pCmdLinOpts->AliasNameList,
                                                   &iTmp, NumValidLines);
     iTmp = 0;
     pCmdLinOpts->AliasUrlList = AllocCharList (pCmdLinOpts->AliasUrlList,
                                                  &iTmp, NumValidLines);
     ConfigFileSeek(pCmdLinOpts->ConfigFile, pzSections, NumSections, "AliasList");
     for (i = 0, NumTokens = 0; i < NumLines; i++)
     {
       fgets (pszRecord, DEFAULT_RECORD_LENGTH, pCmdLinOpts->ConfigFile);
       pszKey   = Truncate_String(pszRecord);
       pszAlias = ParseConfigLine (pszKey);
       pszToken = ParseConfigLine (pszAlias);
       pszUrl   = ParseConfigLine (pszToken);
       if (!pszToken) continue; /* skip to next record */
       pszAlias = InsertAliasNameList (pCmdLinOpts->AliasNameList,
                                       pCmdLinOpts->AliasUrlList, 
                                       pszAlias, pszUrl);
       ListMember.Key = NO_KEY;
       FillConfigListToken (pszKey, &ListMember);
       if (!pCmdLinOpts->iCaseSensitive) StringToLower (pszToken);
       for (pszToken = strtok(pszToken,",");pszToken;
            pszToken = strtok(NULL,","))
       {
          InsertAliasList ( pCmdLinOpts->AliasList,
                           &pCmdLinOpts->NumAliasMembers, ListMember.Key, 
                            pszToken, pszAlias);
       }
     }
   }
 }

 /*
  *  Hide list 
  */

 if ((NumLines = ConfigFileSeek(pCmdLinOpts->ConfigFile,pzSections,
                                NumSections, "HideList")))
 {
   pCmdLinOpts->HideList = FillConfigListFromSection 
                                  (pCmdLinOpts->ConfigFile, 
                                   pCmdLinOpts->HideList,
                                  &pCmdLinOpts->NumHideMembers,
                                   TRUE, NumLines);
 }
 

 /*
  *  process the option section of the configuration file 
  *
  *  currently known entrys are LogFile, SpecialList
  *  
  *
  */

 if ((NumLines = ConfigFileSeek(pCmdLinOpts->ConfigFile,pzSections,
                                NumSections, "Options")))
 {
   for (i = 0; i < NumLines; i++)
   {
     fgets (pszRecord, DEFAULT_RECORD_LENGTH, pCmdLinOpts->ConfigFile);
     pszAlias = Truncate_Config_File_Record (pszRecord);
     /*
      * each meaningful line in section Options has the format
      * Option = Text
      * pszAlias points to Option, pszToken to Text
      */
     pszToken = ParseConfigLine (pszAlias);
     StringToLower (pszAlias);

     if (!pszToken)
       ;  /* currently do nothing */
     else if (!strcmp (pszAlias,"logfile"))                           /* LogFile */
     {
       CheckFile = fopen (pszToken, LOGFILE_OPEN_MODE);
       if (CheckFile == 0)
       {
         fprintf (stderr,LOG_FILE_OPENERR,pszToken);
       }
       else
       {
         fclose (CheckFile);
         if (pCmdLinOpts->NumLogFiles >= MaxLogfiles)
         {        /* extend list */
           pCmdLinOpts->LogFileList = AllocCharList (
                                      pCmdLinOpts->LogFileList,
                                      (size_t *) &MaxLogfiles, 
                                      DEFAULT_LOGFILE_INCREMENT);
         }
         pCmdLinOpts->LogFileList[pCmdLinOpts->NumLogFiles] =  my_malloc 
                                     ((strlen(pszToken) + 1) * sizeof(char));
         strcpy(pCmdLinOpts->LogFileList[pCmdLinOpts->NumLogFiles],pszToken);
         pCmdLinOpts->NumLogFiles++;
       }
     }
     else if (!strcmp (pszAlias,"weekstart") && strlen(pszToken) >= 3)
     {                                                         /* WeekStart */
       char * pszWeekDay;
       char * pszDays = "thuwedtuemonsunsatfri";

       pszToken[3] = '\0';
       pszToken = StringToLower (pszToken);
       pszWeekDay = strstr (pszDays, pszToken);
       if (pszWeekDay)
       {
         pSList->DateList->Offset = (pszWeekDay - pszDays) / 3;
       }
     }
     else if (!strcmp (pszAlias,"outputpath") && !pConfigFileOpts->OutPath) 
     {                                                        /* Output path */
       pConfigFileOpts->OutPath = my_malloc ((strlen(pszToken)+1) * sizeof(char));
       strcpy (pConfigFileOpts->OutPath, pszToken);
     }
     else if (!strcmp (pszAlias,"historyfile") && !pConfigFileOpts->HistoryFile) 
     {                                                        /* HistoryFile*/
       pConfigFileOpts->HistoryFile = my_malloc ((strlen(pszToken)+1) * sizeof(char));
       strcpy (pConfigFileOpts->HistoryFile, pszToken);
     }
     else if (!strcmp (pszAlias,"definetablelayout"))  /* DefineTableLayout */
     {
       char * pszLayoutSection;

       pszLayoutSection = pszToken;
       if (!(SectionIndex = 
             ConfigFileSeek(NULL, pzSections, NumSections, pszLayoutSection)))
       {                            /* section not found */
         fprintf (stderr, DEFINETABLE_SYNTAX_ERROR, pszLayoutSection);
         fprintf (stderr, CONFIG_ERROR_NOSECTION);
       }
       else
       {
         SectionIndex--; /* section number => section index */
         if (pConfigFileOpts->NumLayoutLists >= MaxLayoutSections)
         {                          /* we must extend the list */
           pConfigFileOpts->LayoutList = AllocSectionDescriptList
                                           (pConfigFileOpts->LayoutList,
                                            &MaxLayoutSections,
                                            DEFAULT_SPECIALLIST_EXTEND);
         }
         pConfigFileOpts->LayoutList[pConfigFileOpts->NumLayoutLists]
                                            = pzSections[SectionIndex];
         pzSections[SectionIndex] = NULL;
         pConfigFileOpts->NumLayoutLists++;
       }
     }
     else if (!strcmp (pszAlias,"definetable"))               /* DefineTable */
     {
       /* DefineTable = TableName = Type */
       char * pszTableName;
       char * pszType;

       pszTableName = pszToken;
       pszType  = ParseConfigLine (pszTableName);
       if (pszType)
       {
         if ((SectionIndex = ConfigFileSeek 
                             (NULL, pzSections, NumSections, pszTableName)))
         {
           SectionIndex--; /* section number => section index */
           if (pConfigFileOpts->NumTables >= MaxTables)   /* extend the list */
           {
             pConfigFileOpts->TableList = AllocTableDescriptList (
                                               pConfigFileOpts->TableList,
                                               &MaxTables, 
                                               DEFAULT_SPECIALLIST_EXTEND);
           }
           pConfigFileOpts->TableList[pConfigFileOpts->NumTables] = 
                               my_calloc (1, sizeof (TableDescriptType));
           pConfigFileOpts->TableList[pConfigFileOpts->NumTables]->SpecialList = 
                          my_malloc ((strlen(pszType) + 1) * sizeof (char));
           strcpy (pConfigFileOpts->TableList[pConfigFileOpts->NumTables]
                                  ->SpecialList,pszType);

           pConfigFileOpts->TableList[pConfigFileOpts->NumTables]->FileSection =
                                                    pzSections[SectionIndex];
           pzSections[SectionIndex] = NULL;
           pConfigFileOpts->NumTables++;
         }
         else
         {
           fprintf (stderr, DEFINETABLE_SYNTAX_ERROR, pszTableName);
           fprintf (stderr, CONFIG_ERROR_NOSECTION);
         }

       }
     }
     else if (!strcmp (pszAlias,"speciallist"))               /* SpecialList */
     {
       /* SpecialList = ListName = Key */
       char * pszKey;
       char * pszListName;

       pszListName = pszToken;
       pszKey = ParseConfigLine (pszListName);
       if (pszKey)
       {
         pszKey = StringToLower (pszKey);
         ListMember.Key = NO_KEY;
         FillConfigListToken (pszKey, &ListMember);
         if (!(SectionIndex = 
               ConfigFileSeek(NULL, pzSections, NumSections, pszListName)))
         {                            /* section not found */
           fprintf (stderr, SPECIALLIST_SYNTAX_ERROR, pszListName);
           fprintf (stderr, CONFIG_ERROR_NOSECTION);
         }
         else if (ListMember.Key == NO_KEY)
         {                            /* no valid key specified */
           fprintf (stderr, SPECIALLIST_SYNTAX_ERROR, pszListName);
           fprintf (stderr, SPECIALLIST_INVALID_KEY, pszKey);
         }
         else
         {                            /* all seems ok */
           SectionIndex--; /* section number => section index */
           if (pConfigFileOpts->NumSpecialLists >= MaxSpecialLists)
           {                          /* we must extend the list */
             pConfigFileOpts->SpecialLists = AllocSectionDescriptList
                                             (pConfigFileOpts->SpecialLists,
                                              &MaxSpecialLists,
                                              DEFAULT_SPECIALLIST_EXTEND);
           }
           pConfigFileOpts->SpecialLists[pConfigFileOpts->NumSpecialLists]
                                              = pzSections[SectionIndex];
           pzSections[SectionIndex] = NULL;
           pConfigFileOpts->SpecialLists[pConfigFileOpts->NumSpecialLists]->Key
                                              = ListMember.Key;
           pConfigFileOpts->NumSpecialLists++;
         }
       }
       else
       {
         fprintf (stderr, SPECIALLIST_SYNTAX_ERROR, pszListName);
       }
     }
     else if (!strcmp (pszAlias,"outputfile") ||             /* Output files */
              !strcmp (pszAlias,"createimage"))              /* Images */
     {
       /* OutputFile  = Section = Filename = Period = MaxFiles */
       /* CreateImage = Section = Filename = Period = MaxFiles */
       OutputFileDescriptType ** *pFileList;
       size_t *pNumFiles;
       size_t *pMaxFiles;
       char * pszSectionName;
       char * pszFileName;
       char * pszPeriod;
       char * pszMaxFiles;
       int    iFileNameLen;

       if (!strcmp (pszAlias,"outputfile"))
       {
         pFileList = &pConfigFileOpts->OutFileList;
         pNumFiles = &pConfigFileOpts->NumOutFiles;
         pMaxFiles = &MaxOutFiles;
       }
       else
       {
         pFileList = &pConfigFileOpts->ImageList;
         pNumFiles = &pConfigFileOpts->NumImages;
         pMaxFiles = &MaxImages;
       }
       pszSectionName = pszToken;
       pszFileName = ParseConfigLine (pszSectionName);
       pszPeriod   = ParseConfigLine (pszFileName);
       pszMaxFiles = ParseConfigLine (pszPeriod);
       if (pszFileName)
       {
         if (!(SectionIndex = 
               ConfigFileSeek(NULL, pzSections, NumSections, pszSectionName)))
         {                            /* section not found */
           fprintf (stderr, OUTPUTFILE_SYNTAX_ERROR, pszSectionName);
           fprintf (stderr, CONFIG_ERROR_NOSECTION);
         }
         else
         {
           SectionIndex--; /* section number => section index */
           if (*pNumFiles >= *pMaxFiles)   /* extend the list */
           {
             *pFileList = AllocOutputFileDescriptList ( *pFileList, pMaxFiles, 
                                                  DEFAULT_SPECIALLIST_EXTEND);
           }
           (*pFileList)[*pNumFiles] = my_calloc (1, sizeof (OutputFileDescriptType));
           (*pFileList)[*pNumFiles]->FileSection = pzSections[SectionIndex];
           pzSections[SectionIndex] = NULL;
           (*pFileList)[*pNumFiles]->Period = NO_DATE;
           if (pszPeriod)
           {
             enum DATE_MODE DateMode = NO_DATE;

             pszPeriod = StringToLower (pszPeriod);
             if (pszPeriod[0] == 'd')
               DateMode = PER_DAY;
             else if (pszPeriod[0] == 'w')
               DateMode = PER_WEEK;
             else if (pszPeriod[0] == 'm')
               DateMode = PER_MONTH;
             else
               pszPeriod = NULL;
             (*pFileList)[*pNumFiles]->Period = DateMode;
             if (pszMaxFiles)
               (*pFileList)[*pNumFiles]->MaxFiles = atoi (pszMaxFiles);
             if (DateMode > pConfigFileOpts->DateMode) 
               pConfigFileOpts->DateMode = DateMode;
           }
           iFileNameLen = strlen(pszFileName) + 1;
           if (pszPeriod && (pszToken = strstr (pszFileName, PERIOD_PLACEHOLDER)))
           {
             iFileNameLen += 2;
             *pszToken = '\0';
           }
           else
           {
             pszToken = NULL;
           }
           (*pFileList)[*pNumFiles]->FileName = 
                          my_malloc (iFileNameLen * sizeof (char));
           strcpy ((*pFileList)[*pNumFiles]->FileName,pszFileName);
           if (pszToken)
           {
             char * pszOutfil = (*pFileList)[*pNumFiles]->FileName;
             strcat (pszOutfil, PERIOD_REPLACEMENT);
             pszToken += 4;
             strcat (pszOutfil, pszToken);
           }
           (*pNumFiles)++;
         }
       }
       else
       {
         fprintf (stderr, OUTPUTFILE_SYNTAX_ERROR, pszSectionName);
       }
     }

   }
 }

 /*
  * make final checks
  */

 for (i = 0; i < pConfigFileOpts->NumTables ; i++)
 {
   StringListType * StringList;
   /*
    * at this point we can check if the table points to a standard, a special
    * list of if the syntax is wrong
    */
   /* some "special" stringlists like weekdaylist are already defined */
   StringList = SelectStringList (pSList, NO_KEY, 
                                  pConfigFileOpts->TableList[i]->SpecialList);
   SectionIndex = ConfigFileSeek (NULL, pConfigFileOpts->SpecialLists, 
                                        pConfigFileOpts->NumSpecialLists, 
                                  pConfigFileOpts->TableList[i]->SpecialList);
   if (!SectionIndex && !StringList)
   {      /* should be a standard list */
     char * pszKey;

     pszKey = StringToLower (pConfigFileOpts->TableList[i]->SpecialList);
     ListMember.Key = NO_KEY;
     FillConfigListToken (pszKey, &ListMember);
     if (!strcmp (pszKey,"date"));
     else if (ListMember.Key == NO_KEY)  /* syntax error! */
     {
       SectionDescriptType ** pSection = NULL;

       fprintf (stderr, DEFINETABLE_SYNTAX_ERROR, 
                pConfigFileOpts->TableList[i]->FileSection->Name);
       fprintf (stderr, DEFINETABLE_TYPE_ERROR, pszKey);
       free (pConfigFileOpts->TableList[i]->SpecialList);
       pSection = &pConfigFileOpts->TableList[i]->FileSection;
       FreeSectionDescriptTypeList (pSection, 1);
       free (pConfigFileOpts->TableList[i]);
       pConfigFileOpts->TableList[i] = NULL;
     }
     else
     {
       free (pConfigFileOpts->TableList[i]->SpecialList);
       pConfigFileOpts->TableList[i]->SpecialList = NULL;
       pConfigFileOpts->TableList[i]->FileSection->Key = ListMember.Key;
     }
   }   
 }

 if (!pConfigFileOpts->OutPath) pConfigFileOpts->OutPath = "";

 /*
  * free memory
  */

 free (pszRecord);
 FreeSectionDescriptTypeList (pzSections, NumSections);
 free (pzSections);
#if COUNT_ALLOC_MEMORY >= 2
fprintf (stderr,"filters: Include %u, Exclude %u, Alias %u, Hide %u \n",
(uInt)pCmdLinOpts->NumIncludeMembers,(uInt)pCmdLinOpts->NumExcludeMembers,
(uInt)pCmdLinOpts->NumAliasMembers,  (uInt)pCmdLinOpts->NumHideMembers);
#endif
#if COUNT_ALLOC_MEMORY >= 2
 eval_cpu_time ("elapsed time in config");
#endif
 return pConfigFileOpts;
}


/**************************************************************************
 * FillConfigListFromSection fills a ConfigList from a section in the
 * configuration file. The section has SectionLenght lines and the current
 * file pointer must point to the first line.
 * The created ConfigList will be null pointer terminated, special for
 * needs of the HideList.
 *************************************************************************/
ConfigListType ** FillConfigListFromSection (FILE * ConfigFile,
                                             ConfigListType ** ConfigList, 
                                             size_t * NumMembers, 
                                             int iCaseSensitive, 
                                             size_t SectionLenght)
{
char Record[DEFAULT_RECORD_LENGTH];
char * pszToken;
long int Offset = ftell (ConfigFile);
ConfigListType ListMember;
int NumTokens, i;

 /*
  * line syntax: key=val1,val2,...
  */
 for (i = 0, NumTokens = 0; i < SectionLenght; i++)
 {
   fgets (Record, DEFAULT_RECORD_LENGTH, ConfigFile);
   pszToken = Truncate_String(Record);
   if (pszToken[0] != '#') NumTokens += CheckNumberOfTokens (pszToken,',');
 }
 if (NumTokens)
 {
   size_t iTmp = *NumMembers;
   ConfigList = AllocConfigList (ConfigList, &iTmp, NumTokens + 1);
   fseek (ConfigFile, Offset, SEEK_SET);
   for (i = 0, NumTokens = 0; i < SectionLenght; i++)
   {
     fgets (Record, DEFAULT_RECORD_LENGTH, ConfigFile);
     pszToken = Truncate_String(Record);
     if (pszToken[0] == '#') continue; /* skip to next record */
     if (!iCaseSensitive) StringToLower (pszToken);
     ListMember.Key = ITEM;
     for (pszToken = strtok(pszToken,","); pszToken; pszToken = strtok(NULL,","))
     {
       if (FillConfigListToken (Truncate_String(pszToken), &ListMember))
       {
          InsertConfigList (ConfigList, NumMembers, &ListMember);
       }
     }
   }
 }

 return ConfigList;
}

/**************************************************************************
 * frees a list of type SectionDescriptType
 *************************************************************************/
void FreeSectionDescriptTypeList 
     (SectionDescriptType ** pzSections, size_t NumSections)
{
int i;

 for (i = 0; i < NumSections; i++)
 {
   if (pzSections[i])
   {
     if (pzSections[i]->Name) free (pzSections[i]->Name);
     free (pzSections[i]);
   }
 }
 return;
}


/**************************************************************************
 * ConfigFileSeek positions the file pointer to the section specified in
 * pszName if it exists in pzSections.
 * return value: Number of lines in section.
 * If ConfigFile is NULL, only the existence of the section is checked.
 * return value: Number of section
 *************************************************************************/
size_t ConfigFileSeek (FILE * ConfigFile, SectionDescriptType ** pzSections,
                       size_t NumSections, const char * pszName)
{
size_t NumLines = 0;
int i = 0, Status = FALSE;

 for (i = 0; (i < NumSections) && !Status; i++)
 {
   if (pzSections[i] && !strcmp (pszName,pzSections[i]->Name))
   {
     if (ConfigFile)
     {
       Status = fseek (ConfigFile, pzSections[i]->Offset, SEEK_SET);
       if (Status) 
       {
         Status = FALSE;
       }
       else
       {
          Status = TRUE;
          NumLines = pzSections[i]->Lenght;
       }
     }
     else
     {
       Status = TRUE;
       /* Number of section (0 means no section found) */
       /* (empty sections aren't from interest) */
       if (pzSections[i]->Lenght) NumLines = i + 1; 
     }
   }
 }
 return NumLines;
}


/**************************************************************************
 * CheckConfigFileSyntax makes some basic checks on the syntax of the
 * configuration file. It also stores the name and positions of the found
 * sections in the list pSections.
 * If MaxLines is not specified, the whole file is checked
 *************************************************************************/
SectionDescriptType ** CheckConfigFileSyntax (FILE * ConfigFile,
                                              size_t * pMaxSections,
                                              size_t MaxLines)
{
SectionDescriptType ** pSections = NULL;
char * pszRecord;   /* one line from the file*/
char * pszTruncRecord;
char  szSectionName[80] = {'\0'}; /* Name of a Section. Each Section starts */
size_t NumSections = 0;           /* with a "Begin" an end with a "End" */
size_t NumLinesOfSection = 0;
size_t NumRecords = 0;
long int Offset = 0;


 if (!ConfigFile) return  pSections;  /* no file specified */

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));
 if (!MaxLines) rewind (ConfigFile); 

 while (fgets(pszRecord,DEFAULT_RECORD_LENGTH,ConfigFile))
 {
   NumRecords++;
   pszTruncRecord = Truncate_Config_File_Record (pszRecord); /*remove whitespace chars*/
   if (!*szSectionName && !strncmp (pszTruncRecord, "Begin", 5))
   { /* begin of a section found */
     char * pszSectionName = Truncate_String (&pszTruncRecord[5]);
     if (pszSectionName && *pszSectionName)
     {
       strncpy (szSectionName,pszSectionName,sizeof szSectionName);
       Offset = ftell (ConfigFile);
     }
   }
   else if (*szSectionName && !strncmp (pszTruncRecord   ,"End",3)  && 
            !strcmp (Truncate_String (&pszTruncRecord[3]),szSectionName))
   { /* end of a section found */
     if (NumSections >= *pMaxSections)
     {    /* allocate more memory */
       pSections =  AllocSectionDescriptList
                        (pSections, pMaxSections, DEFAULT_SECTIONLIST_EXTEND);
     }
     pSections[NumSections] =  my_calloc (1 ,sizeof(SectionDescriptType));
     pSections[NumSections]->Name = my_malloc ((strlen(szSectionName) + 1) 
                                                * sizeof(char));
     strcpy (pSections[NumSections]->Name,szSectionName);
     pSections[NumSections]->Key = NO_KEY;
     pSections[NumSections]->Offset = Offset;
     pSections[NumSections]->Lenght = NumLinesOfSection;
     /* init for next cycle */
     NumSections++;
     NumLinesOfSection = 0;
     szSectionName[0] = '\0';
   }
   else if (*szSectionName)
   {
     NumLinesOfSection++;
   }
   if (MaxLines && (NumRecords >= MaxLines)) break;
 } 
 if (szSectionName[0])
 {
   printf (CONFIG_SYNTAX_MISSING_ENDSECTION,szSectionName);
 }
 free (pszRecord);
 *pMaxSections = NumSections;
 return pSections;
}



/**************************************************************************
 * Truncate_Config_File_Record truncates a record from the configuration
 * file. Comments and leading or trailing whitespace characters are removed.
 *************************************************************************/
char * Truncate_Config_File_Record (char * pszRecord)
{
char * pszTemp = NULL;

 if (pszRecord)
 {
   if ((pszTemp = strchr (pszRecord,'#'))) *pszTemp = '\0';
   pszRecord = Truncate_String (pszRecord);
 }
 return pszRecord;
}



/**************************************************************************
 *  accstat qualifier parameter
 *
 *  valid qualifier are:
 *
 * /configuration=filename                 (default: none)
 * /extract=filename                       (default: none)
 * /output=filename                        (default: none (=stdout))
 * /include=key=val1,val2,key=val3,val4... (default: *)
 * /exclude=key=val1,val2,key=val3,val4... (default: none)
 * /since=date                             (default: 01-Jan-1970)
 * /before=date                            (default: today)
 * /show=all={a,nn},key={a,nn},...         (default: none)
 *
 *  valid keys are  site, item, status, method or protocol
 *  valid dates are dd-mmm-jjjj, mm/dd/jjjj, dd.mm.jjjj,
 *                  ddd (delta time), yesterday, today or tomorrow
 *
 *  valid parameter are:   logfile1 logfile2 ...
 *
 *************************************************************************/
CommandLineType * ParseCommandLine (int argc, char *argv[])
{
CommandLineType * pOptions;
char * pszArg;  /* copy of an argument */
char * pszOption; /* option part of qualifier */
char * pszToken;
char * pszConfigFile = NULL; /* name of the config file */
char * pszExtractFile = NULL; /* name of extract file */
enum QUALIFIER Argument;
FILE * CheckFile;

ConfigListType ListMember ;


size_t MaxLogFiles=0;
size_t NumberOfTokens;
int Fail=FALSE;
int iLenght=0;
int i, iTmp;

 pOptions = my_calloc (1, sizeof (CommandLineType));

 pOptions->IncludeList   = NULL;
 pOptions->ExcludeList   = NULL;
 pOptions->AliasList     = NULL;
 pOptions->HideList      = NULL;
 pOptions->AliasNameList = NULL;
 pOptions->AliasUrlList  = NULL;
#ifdef UNIX
 pOptions->iCaseSensitive = TRUE;
#else
 pOptions->iCaseSensitive = FALSE;
#endif

 /*  determine maximum argument lenght and number of logfiles specified */
 for (i=1; i < argc; i++)
 {
   iLenght = (iLenght >= strlen(argv[i])) ? iLenght : strlen(argv[i]);
   if (     argv[i][0] != '-' 
#ifndef UNIX
          &&  argv[i][0] != '/'
#endif /* !UNIX */
      )
   {
     MaxLogFiles++;
   }
 }
 iTmp = 0;
 pOptions->LogFileList = AllocCharList (pOptions->LogFileList,
                         (size_t *)&iTmp,MaxLogFiles);
 pszArg = (char *) my_malloc ((size_t) (iLenght+1) * sizeof(char) );
 
/*
 *  now check the arguments 
*/

 for (i=1; i < argc; i++)    
 {
   strcpy (pszArg, argv[i]);
   Argument=CheckQualifier(pszArg, &pszOption);
   switch (Argument)
   {
    case NO_QUALIFIER:
       Fail = TRUE;
       break;
    case PARAMETER:
       CheckFile = fopen (pszArg, LOGFILE_OPEN_MODE);
       if (CheckFile == 0)
       {
          fprintf (stderr,LOG_FILE_OPENERR,pszArg);
       }
       else
       {
          fclose (CheckFile);
          pOptions->LogFileList[pOptions->NumLogFiles] = (char *)  my_malloc 
                    ((strlen(pszArg) + 1) * sizeof(char));
          strcpy(pOptions->LogFileList[pOptions->NumLogFiles],pszArg);
          pOptions->NumLogFiles++;
       }
       break;
    case CONFIG:
       if (!pszConfigFile && pszOption)
       {
         pszConfigFile = (char *)  my_malloc 
                      ((strlen(pszOption) + 1) * sizeof(char));
         strcpy(pszConfigFile,pszOption);
       }
       break;
    case EXTRAKT:
       if (!pszExtractFile && pszOption)
       {
          pszExtractFile = (char *)  my_malloc 
                      ((strlen(pszOption) + 1) * sizeof(char));
          strcpy(pszExtractFile,pszOption);
       }
       break;
    case OUTPUT:
       if (pszOption)
       {
          CheckFile = fopen (pszOption, "a");
          if (CheckFile)
          {
            fclose (CheckFile);
            CheckFile = freopen (pszOption, "a", stdout);
          }
          if (!CheckFile)
          {
            Fail = TRUE;
            fprintf (stderr,OUTPUT_REOPEN_FAIL,pszOption);
          }
       }
       break;
    case INCLUDE:
       if (!pOptions->iCaseSensitive) StringToLower (pszOption);
       NumberOfTokens = CheckNumberOfTokens (pszOption,',');
       iTmp = 0;
       pOptions->IncludeList = AllocConfigList (pOptions->IncludeList,
                                          (size_t *)&iTmp,NumberOfTokens);
       ListMember.Key = ITEM;
       for (pszToken = strtok(pszOption,",");pszToken;
            pszToken = strtok(NULL,","))
       {
         if (FillConfigListToken (pszToken, &ListMember))
         {
            InsertConfigList (pOptions->IncludeList,
                             &pOptions->NumIncludeMembers,&ListMember);
         }
       }
       break;
    case EXCLUDE:
       if (!pOptions->iCaseSensitive) StringToLower (pszOption);
       NumberOfTokens = CheckNumberOfTokens (pszOption,',');
       iTmp = 0;
       pOptions->ExcludeList = AllocConfigList (pOptions->ExcludeList,
                                          (size_t *)&iTmp,NumberOfTokens);
       ListMember.Key = ITEM;
       for (pszToken = strtok(pszOption,",");pszToken;
            pszToken = strtok(NULL,","))
       {
         if (FillConfigListToken (pszToken, &ListMember))
         {
            InsertConfigList (pOptions->ExcludeList,
                             &pOptions->NumExcludeMembers,&ListMember);
         }
       }
       break;
    case SINCE:
       pOptions->since = DateFromString_t (StringToLower(pszOption));
       break;
    case BEFORE:
       pOptions->before = DateFromString_t (StringToLower(pszOption));
       break;
    case SHOW:
       if (!pszOption)
       {
         SetShowDefault (pOptions, "all");
       }
       else
       {
         StringToLower(pszOption);
         for (pszToken = strtok(pszOption,",");pszToken;
              pszToken = strtok(NULL,","))
         {
           SetShowDefault (pOptions, pszToken); /* look for "all" */
           ListMember.Key = NO_KEY;
           if (FillConfigListToken (pszToken, &ListMember))
           {
              SetShowKey (pOptions,&ListMember);
           }
         }
       }
       break;
    case HELP:
       usage ();
       break;
    default:
       puts (INTERNAL_ERROR_FATAL);
       exit (EXIT_FAILURE);
   }
 }
 
 /* 
  * all Qualifier read, now make some additional checks
 */

 if (pszConfigFile) /* Config file specified?*/
 {
    pOptions->ConfigFile = fopen (pszConfigFile, "r");
    if (pOptions->ConfigFile == NULL)
    {
       fprintf (stderr,CONFIG_FILE_OPENFAIL,pszConfigFile);
       Fail = TRUE;
    }
    free (pszConfigFile);    
 } 

/*
 * checking the extract file; if it already exist,
 * the first line should be equal to EXTRACT_FILE_STARTLINE
 */

 if (pszExtractFile) /* Extract file specified?*/
 {
    CheckFile = fopen (pszExtractFile, "r");
    pszOption =  NULL;
    if (CheckFile)
    {
      pszOption = my_malloc ( 46 * sizeof(char));
      fgets (pszOption, 44, CheckFile);
      fclose (CheckFile);
      if (strncmp (pszOption, EXTRACT_FILE_STARTLINE, 40))
      {   /* not equal */
         CheckFile = stdout;    /* not NULL means wrong file structure */
         fprintf (stderr,EXTRAKT_FILE_EXIST,pszExtractFile);
         Fail = TRUE;
         free (pszOption);
         pszOption = NULL;
      }
      else
      {   /* equal */
         CheckFile = NULL; /* file is fine */
      }
    }
    if (CheckFile == NULL)
    {
      pOptions->ExtractFile = fopen (pszExtractFile, "a");
      if (pOptions->ExtractFile == NULL)
      {
         fprintf (stderr,EXTRAKT_FILE_OPENFAIL,pszExtractFile);
         Fail = TRUE;
      }
      else
      {
         if (pszOption)
         {
            free (pszOption);
         }
         else
         {        /* new file */
            fputs (EXTRACT_FILE_STARTLINE, pOptions->ExtractFile);
         }
      }
    }
    free (pszExtractFile);    
 } 


/*
 * date checks
*/

 if (pOptions->since == DateFromString_t("today"))
 {
   pOptions->before = DateFromString_t("tomorrow");
 }

 if (pOptions->before == 0) pOptions->before = DateFromString_t("today");
 
 if (pOptions->before <= pOptions->since)
 {
   fputs (DATE_CHECK_FAIL,stderr);
   Fail = TRUE;
 }
 free (pszArg);
 if (Fail)
 {
    fprintf (stderr, INIT_FAIL, DEF_QU);
    exit (EXIT_FAILURE);
 }
 return pOptions;
}

/**************************************************************************
 * SetShowDefault set the default values for the command line option
 * SHOW (what to list etc.)
 *************************************************************************/
void SetShowDefault (CommandLineType * pOptions, char * pszString)
{
enum DATE_MODE DateMode = NO_DATE;
SortType Sort = SORT_NUMERICAL;
int ShowMembers = 0;    /* default: show nothing */
char * pszTemp;
int i;

 if (pszString)
 {
   if (!strncmp (pszString, "all", 3))
   {
     ShowMembers = DEFAULT_TABLE_LENGHT ;     /*  entrys in each table */
     if (((pszString[3] == '=') || (pszString[3] == ':')) &&
          (strlen(pszString) > 4))
     {
       if (pszString[4] == 'a')   /* all in alphabetical order */
       {
         ShowMembers = -1;
         Sort = SORT_ALPHABETICAL;
       }
       else /* should be a number */
       {
         ShowMembers = atoi (&pszString[4]);
         Sort = SORT_NUMERICAL;
       }
     }
   }
   else if (!strncmp (pszString, "date", 4))
   {
     DateMode = PER_DAY;
     pszTemp = strchr (pszString, '=');
     if (pszTemp)
     {
       if (pszTemp[1] == 'w')
       {
         DateMode = PER_WEEK;
       }
       else if (pszTemp[1] == 'm')
       {
         DateMode = PER_MONTH;
       }
     }
   }
 }

 for (i=0; i <= KEYS_NUM_MEMBERS; i++)
 {
   if (!pOptions->ShowMembers[i])
   {
     pOptions->ShowMembers[i] = ShowMembers;
     pOptions->Sort[i] = Sort;
   }
 }
 if (DateMode != NO_DATE) pOptions->DateMode = DateMode;
}



/**************************************************************************
 * SetShowKey sets the lenght of a output table as specified in the
 * command line
 *************************************************************************/
void SetShowKey (CommandLineType * pOptions, ConfigListType * Member)
{
SortType Sort;
int ShowMembers;

 if (Member->Val.pszString[0] == 'a') 
 {
   Sort =  SORT_ALPHABETICAL;
   ShowMembers = -1;
 }
 else if (isdigit(Member->Val.pszString[0]) || Member->Val.pszString[0] == '-')
 {
   ShowMembers = atoi (Member->Val.pszString);
   if (ShowMembers < 0)
   {
     Sort =  SORT_ALPHABETICAL;
     ShowMembers = abs (ShowMembers);
   }
   else
   {
     Sort =  SORT_NUMERICAL;
   }
 }
 else
 {
   Sort =  SORT_NUMERICAL;
   ShowMembers = DEFAULT_TABLE_LENGHT ;
 }

 pOptions->ShowMembers[Member->Key] = ShowMembers;
 pOptions->Sort[Member->Key] = Sort;

}


/**************************************************************************
 * adds Member to ConfigListType if it is not already there
 *************************************************************************/

void InsertConfigList (ConfigListType ** List, size_t * numelem,
                       ConfigListType * Member)
{
int Exists=FALSE;
int i;

 for (i=0;i < *numelem; i++)
 {
   if (List[i]->Key == Member->Key)
   {
     if (!strcmp(List[i]->Val.pszString,Member->Val.pszString))
     {
       Exists = TRUE;
     }
   }
 }

 if (!Exists)
 {
   i = *numelem;
   List[i] = (ConfigListType *) my_calloc (1, sizeof(ConfigListType));
   List[i]->Key = Member->Key;
   List[i]->Val.pszString = (char *) my_malloc (
                                     (strlen(Member->Val.pszString) + 1)
                                      * sizeof(char) );
   strcpy (List[i]->Val.pszString,Member->Val.pszString);
   *numelem = ++i;
 }
 return;
}



/**************************************************************************
 * adds Member to list of AliasListType if it is not already there
 *************************************************************************/

void InsertAliasList (AliasListType ** List, size_t * numelem,
     KeyType Key, char * pszToken, char * pszAlias)
{
int Exists=FALSE;
int i;


 pszToken = Truncate_String(pszToken);
 for (i=0;i < *numelem; i++)
 {
   if (List[i]->Key == Key)
   {
     if (!strcmp(List[i]->Val.pszString,pszToken))
     {
       Exists = TRUE;
     }
   }
 }

 if (!Exists)
 {
   i = *numelem;
   List[i] = (AliasListType *) my_calloc (1, sizeof(AliasListType));
   List[i]->Key = Key;
   List[i]->pszAlias = pszAlias; /* hooks to the member in AliasNameList */
   List[i]->Val.pszString = (char *) my_malloc ( (strlen(pszToken) + 1)
                                                        * sizeof(char) );
   strcpy (List[i]->Val.pszString,pszToken);
   *numelem = ++i;
 }
 return;
}



/**************************************************************************
 * InsertAliasNameList inserts an alias in the list 
 * if it is not already there.
 * Return value: pointer to the entry in the alias name list.
 *************************************************************************/

 char * InsertAliasNameList (char ** AliasNameList, char ** AliasUrlList,
                             char * pszString, char * pszUrl)
{
 char * pszAlias = NULL;
 int i = 0;

 if (pszString)
 {
  while (AliasNameList[i])
  {
    if (!strcmp(AliasNameList[i],pszString))
    {
      pszAlias = AliasNameList[i];
    }
    i++;
  }
  if (!pszAlias)
  {      /* new alias */
    AliasNameList[i] = my_malloc ((strlen(pszString) + 1) * sizeof (char));
    strcpy (AliasNameList[i],pszString);
    if (AliasUrlList && pszUrl)
    {
      AliasUrlList[i] = my_malloc ((strlen(pszUrl) + 1) * sizeof (char));
      strcpy (AliasUrlList[i],pszUrl);
    }
    pszAlias = AliasNameList[i];
  }
 }
 return pszAlias;
}


/***************************************************************************
 * SetHistoryList combines DateList and HistoryList.
 ***************************************************************************/
DateListType * SetHistoryList (DateListType * DateList, 
                               DateListType * HistoryList)
{
time_t * DateAr;     /* array of date values */
uInt * uiCountAr;    /* array of access count values */
uInt * uiByteAr;     /* array of byte values */
uInt uiCount, uiByte;
time_t FirstDate, LastDate, CurDate;
uInt NumDays = 0;
int i, j;


 if (!DateList->NumMembers)
   ;
 else if (!HistoryList->NumMembers) 
 {
   free (HistoryList);
   HistoryList = DateList;
 }
 else
 {
   FirstDate = (MinDate(HistoryList) < MinDate(DateList)) ?
                MinDate(HistoryList) : MinDate(DateList);
   LastDate  = (MaxDate(HistoryList) > MaxDate(DateList)) ?
                MaxDate(HistoryList) : MaxDate(DateList);

   NumDays = (LastDate - FirstDate) /  SECONDS_OF_A_DAY + 1;
   DateAr    =  my_calloc (NumDays, sizeof (time_t));
   uiCountAr =  my_calloc (NumDays, sizeof (uInt));
   uiByteAr  =  my_calloc (NumDays, sizeof (uInt));

   for (CurDate  = FirstDate, i = 0; CurDate <= LastDate; 
        CurDate += SECONDS_OF_A_DAY, i++)
   {
     for (j = 0, uiCount = 0, uiByte = 0; j < HistoryList->NumMembers; j++)
     {
       if (CurDate == HistoryList->Date[j])
       {
         uiCount = HistoryList->uiCountAr[j];
         uiByte  = HistoryList->uiByteAr[j];
         j       = HistoryList->NumMembers;
       }
     }
     for (j = 0; j < DateList->NumMembers; j++)
     {
       if (CurDate == DateList->Date[j])
       {
         if (DateList->uiCountAr[j] > uiCount) 
           uiCount = DateList->uiCountAr[j];
         if (DateList->uiByteAr[j] > uiByte) 
           uiByte = DateList->uiByteAr[j];
         j = DateList->NumMembers;
       }
     }
     DateAr[i]    = CurDate;
     uiCountAr[i] = uiCount;
     uiByteAr[i]  = uiByte;
   }
   
   free (HistoryList->Date);
   free (HistoryList->uiCountAr);
   free (HistoryList->uiByteAr);
   HistoryList->Date = DateAr;
   HistoryList->uiCountAr = uiCountAr;
   HistoryList->uiByteAr = uiByteAr;
   HistoryList->NumMembers = NumDays;
 }
 return HistoryList;
}


/***************************************************************************
 * CreateSpecialLists creates some additional lists specified in the
 * configuration file
 ***************************************************************************/
void CreateSpecialLists (StandardListType * pSList, 
                         CommandLineType  * pCmdLinOpts,
                         ConfigFileType   * pConfigOpts)
{
size_t NumNameListMembers = 0;   /* count of current members */
size_t MaxNameListMembers = 0;   /* current size of list */
size_t NumTokenListMembers = 0;  /* count of current members */
size_t MaxTokenListMembers = 0;  /* current size of list */
char ** NameList = NULL; /* null pointer terminated list of names */
SpecialListType ** TokenList = NULL;
StringListType * StringList;
size_t NumLines = 0;
size_t NumDays;
int iDateNum = 0;
char *pszRecord;            /* record from the configuration file */
char * pszName, * pszToken; /* string pointers for temp. use */
char * szWeekDays[] = {THURSDAY, FRIDAY, SATURDAY, SUNDAY, MONDAY, TUESDAY, WEDNESDAY};
int i, j, k, num, Ind;


 
 /*
  *  set history date list
  */
 pSList->HistoryDateList = SetHistoryList (pSList->DateList, 
                                           pSList->HistoryDateList);
 /*
  *  create weekday list
  */
 for (i = 0; i < 7; i++)
 {
   Ind = (i + (7 - pSList->DateList->Offset)) % 7;
   StringListInsert (pSList->WeekDayList, szWeekDays[Ind], iDateNum);
 }

 if (!pConfigOpts || !pConfigOpts->NumSpecialLists) return; /* basic checks */
 /*
  *  init the lists
  */
 pSList->NumSpecialLists = pConfigOpts->NumSpecialLists;
 NumDays = pSList->DateList->NumMembers;
 if (!NumDays) NumDays++;
 pSList->SpecialLists = my_malloc (pConfigOpts->NumSpecialLists * 
                                   sizeof (StringListType *));
 for (i = 0; i < pConfigOpts->NumSpecialLists; i++)
 {
   pSList->SpecialLists[i] = my_calloc (1, sizeof(StringListType));
   pSList->SpecialLists[i]->Key  = pConfigOpts->SpecialLists[i]->Key;
   pSList->SpecialLists[i]->Name = pConfigOpts->SpecialLists[i]->Name;
   pSList->SpecialLists[i]->DimCountAr = NumDays;
 }

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));

 /*
  *  now fill the lists
  */
 for (num = 0; num < pConfigOpts->NumSpecialLists; num++)
 {
   /*
    * generate NameList and TokenList (reuse them for each list)
    */
   NumNameListMembers = NumTokenListMembers = 0;
   NumLines = ConfigFileSeek (pCmdLinOpts->ConfigFile,
                              pConfigOpts->SpecialLists,
                              pConfigOpts->NumSpecialLists,
                              pConfigOpts->SpecialLists[num]->Name);
   if (MaxNameListMembers <= NumLines)       /* max. 1 name per line */
   {
     NameList = AllocCharList (NameList, &MaxNameListMembers,
                                        NumLines - MaxNameListMembers + 1);
   }
   for (i = 0; i < MaxNameListMembers; i++) /* clean up */
   {
     NameList[i] = NULL;
   }
   for (i = 0; i < NumLines; i++)
   {
     /* line has the syntax "Name = token1, token2, ..." */
     fgets (pszRecord, DEFAULT_RECORD_LENGTH, pCmdLinOpts->ConfigFile);
     pszName = Truncate_String(pszRecord);
     if (!(pszToken = ParseConfigLine (pszName))) continue;  /* skip */
     pszName = InsertAliasNameList (NameList, NULL, pszName, NULL);
     for (pszToken = strtok(pszToken,",");pszToken;
          pszToken = strtok(NULL,","))
     {
      for (j = 0; j < 2; j++)
      {
       /*
        * the aliaslist is casesensitive, so we need a copy if 
        * casesensitivity isn't wanted.
        */
       if (j)
       {
         if (!pCmdLinOpts->iCaseSensitive)
         {
           StringToLower (pszToken);
           if (!strcmp(pszToken, TokenList[NumTokenListMembers - 1]->Token)) 
             break; /* we need no copy */
         }
         else
         {
           break; /* we need no copy */
         }
       }
       pszToken = Truncate_String(pszToken);
       if (NumTokenListMembers >= MaxTokenListMembers)
       {
         TokenList = AllocSpecialList (TokenList, &MaxTokenListMembers,
                                       DEFAULT_SECTIONLIST_EXTEND);
       }
       if (!TokenList[NumTokenListMembers])   /* new entry */
       {
         TokenList[NumTokenListMembers] = my_calloc (1, sizeof(SpecialListType));
       }
       TokenList[NumTokenListMembers]->Name = pszName;
       TokenList[NumTokenListMembers]->Token = my_malloc ((strlen(pszToken)+1)
                                                           *sizeof(char));
       strcpy (TokenList[NumTokenListMembers]->Token,pszToken);
       NumTokenListMembers++;
      }
     }
   }
   /*
    * ok, now create the list itself
    */
    for (i = 0; i < MaxNameListMembers; i++)
    {
      if (NameList[i]) NumNameListMembers++;
    }
    pSList->SpecialLists[num]->DefaultExtend = NumNameListMembers;
    StringList = SelectStringList (pSList, pSList->SpecialLists[num]->Key, "");
    for (i = 0; i < NumNameListMembers; i++)
    {
      StringListInsert (pSList->SpecialLists[num], NameList[i], iDateNum);
      Ind = pSList->SpecialLists[num]->iLast;
      pSList->SpecialLists[num]->List[Ind]->uiCountAr[iDateNum]--;
    }
    for (i = 0; i < StringList->NumMembers; i++)
    {
      for (j = 0; j < NumTokenListMembers; j++)
      {
        if (!wild_strcmp (StringList->List[i]->pszStr,TokenList[j]->Token))
        {
          StringListInsert (pSList->SpecialLists[num], TokenList[j]->Name, iDateNum);
          Ind = pSList->SpecialLists[num]->iLast;
          pSList->SpecialLists[num]->List[Ind]->uiCountAr[iDateNum]--;
          for (k = 0; k < NumDays; k++)
          {
            pSList->SpecialLists[num]->List[Ind]->uiCountAr[k] += 
                                            StringList->List[i]->uiCountAr[k];
          }
          break;
        }
      }
    }


   /*
    *  clear token and name list
    */
   for (i = 0; i < NumTokenListMembers; i++)
   {
     if (TokenList[i]->Token) free (TokenList[i]->Token);
     TokenList[i]->Token = NULL;
   }
   for (i = 0; i < MaxNameListMembers; i++)
   {
     if (NameList[i]) free (NameList[i]);
   }
 }

 /*
  * free memory not longer needed
  */
 for (i = 0; i < MaxTokenListMembers; i++)
 {
   if (TokenList[i]) free (TokenList[i]);
 }
 free (pszRecord);
 free (NameList);
 free (TokenList);
#if COUNT_ALLOC_MEMORY >= 2
 eval_cpu_time ("elapsed time in creating special lists");
#endif
 return;
}


/**************************************************************************
 * fills structure ListMember from string pszToken
 * Token is one substring of the comma separated list after the
 * /include or /exclude command line options or from within the configuation 
 * file. The string pszToken should have the structure "[key=]option".
 *************************************************************************/
int FillConfigListToken (char * pszToken, ConfigListType * ListMember)
{
char * pszPosition=NULL;
KeyType Key;

 if ((pszPosition = strpbrk (pszToken,"=:")) != NULL )
 {                                               /* separate option part */
   *pszPosition = '\0';
    pszPosition++;
    if (strlen(pszPosition) == 0) 
       pszPosition = NULL;
    else
       pszPosition = Truncate_String(pszPosition);
 }
 if (pszPosition || (ListMember->Key == NO_KEY))
 {
    pszToken = Truncate_String(pszToken);
    Key = TranslateStandardListKey (pszToken);
    if (Key != NO_KEY)  /* found */
    {
       ListMember->Key = Key;  /* else same as the last time */
    }
 }
 else
 {
   Key = ListMember->Key;
 }
 /* fill structure */
 if (!pszPosition) pszPosition = pszToken;
 if ((Key != NO_KEY) && strlen(pszPosition) > 0)
 {
    pszPosition = pszPosition; /* CASE_SENSITIVE */
    ListMember->Val.pszString = pszPosition;
    return TRUE;
 }
 else
 {
    return FALSE;
 }
}



/**************************************************************************
 * TranslateStandardListKey translates a string in a standard list key.
 * If the string does not match a standard list, NO_KEY is returned.
 *************************************************************************/
KeyType TranslateStandardListKey (const char * String)
{
KeyType Key = NO_KEY;
char * pszKeyList[] = {KEYSTRINGS, NULL};
char Buf[MAXKEYLEN + 1];
char * pszKey;
int i, iAmbiguous = 0;

 strncpy (Buf, String, MAXKEYLEN);
 pszKey = StringToLower(Buf);
 for (i=0; pszKeyList[i] ;i++)
 {
   if (!strncmp(pszKeyList[i],pszKey,strlen(pszKey)))
   {
     iAmbiguous++;
     Key = i;
   }
 }
 if (iAmbiguous > 1)
 {
   fprintf (stderr,AMBIGUOUS_KEY_WARNING,pszKey);
   Key = NO_KEY;
 }

 return Key;
}


/**************************************************************************
 * CheckNumberOfTokens determines the number of members in a
 * cSeparator separated List
 *************************************************************************/
size_t CheckNumberOfTokens (char * pszString, char cSeparator)
{
 size_t Number=0;
 char * pszTemp;
 
 if (pszString)
 {
   pszTemp = my_malloc ((strlen(pszString) + 1) * sizeof(char));
   strcpy (pszTemp,pszString);
   for (pszTemp = strtok(pszTemp,&cSeparator);pszTemp;
        pszTemp = strtok(NULL,&cSeparator))
   {
      Number++;
   }
   free (pszTemp);
 }
 return Number;
}


/**************************************************************************
 * InitStandardLists does some initialisations on standard lists
 *************************************************************************/
StandardListType * InitStandardLists (void)
{
StandardListType * pSList;

 pSList = my_calloc (1, sizeof(StandardListType));

 pSList->DateList =        my_calloc (1, sizeof(DateListType));
 pSList->HistoryDateList = my_calloc (1, sizeof(DateListType));
 pSList->SiteList =        my_calloc (1, sizeof(StringListType));
 pSList->ItemList =        my_calloc (1, sizeof(StringListType));
 pSList->MethodList =      my_calloc (1, sizeof(StringListType));
 pSList->ProtocolList =    my_calloc (1, sizeof(StringListType));
 pSList->StatusList =      my_calloc (1, sizeof(StringListType));
 pSList->AgentList =       my_calloc (1, sizeof(StringListType));
 pSList->WeekDayList =     my_calloc (1, sizeof(StringListType));
 pSList->SpecialLists =    NULL;

 pSList->DateList->DefaultExtend = DEFAULT_DATE_NUM * 2;
 pSList->DateList->Offset = DAY_OFFSET;

 pSList->SiteList->DefaultExtend = DEFAULT_SITE_NUM;
 pSList->ItemList->DefaultExtend = DEFAULT_ITEM_NUM;
 pSList->MethodList->DefaultExtend = DEFAULT_METHOD_NUM;
 pSList->ProtocolList->DefaultExtend = DEFAULT_PROTOCOL_NUM;
 pSList->StatusList->DefaultExtend = DEFAULT_STATUS_NUM;
 pSList->AgentList->DefaultExtend = DEFAULT_AGENT_NUM;
 pSList->WeekDayList->DefaultExtend = 7;

 pSList->SiteList->DimCountAr = DEFAULT_DATE_NUM;
 pSList->ItemList->DimCountAr = DEFAULT_DATE_NUM;
 pSList->MethodList->DimCountAr = DEFAULT_DATE_NUM;
 pSList->ProtocolList->DimCountAr = DEFAULT_DATE_NUM;
 pSList->StatusList->DimCountAr = DEFAULT_DATE_NUM;
 pSList->AgentList->DimCountAr = DEFAULT_DATE_NUM;

 pSList->SiteList->Key = SITE;
 pSList->ItemList->Key = ITEM;
 pSList->MethodList->Key = METHOD;
 pSList->ProtocolList->Key = PROTOCOL;
 pSList->StatusList->Key = STATUS;
 pSList->AgentList->Key = AGENT;
 pSList->WeekDayList->Key = NO_KEY;

 pSList->SiteList->Name = "Sites";
 pSList->ItemList->Name = "Items";
 pSList->MethodList->Name = "Methods";
 pSList->ProtocolList->Name = "Protocols";
 pSList->StatusList->Name = "Status Codes";
 pSList->AgentList->Name = "User Agents";
 pSList->WeekDayList->Name = "Weekdays";

 pSList->WeekDayList->Sort = SORT_NOSORT;
 pSList->WeekDayList->ShowMembers = 7;
 return pSList;
}


/***************************************************************************
 * AdjustFromHistory uses values read from history file to adjust some 
 * start values.
 ***************************************************************************/
 void AdjustFromHistory (StandardListType * pSList, 
                         CommandLineType  * pCmdLinOpts,
                         ConfigFileType   * pConfigOpts)
{
time_t FirstDate, StartDate;
time_t StartPeriod, EndPeriod;
enum DATE_MODE DateMode = pConfigOpts->DateMode;
uInt NumDays;

 if (!pCmdLinOpts->since && pSList->HistoryDateList->NumMembers)
 {
   StartDate = FirstDate = MaxDate (pSList->HistoryDateList);
   if (DateMode == NO_DATE) DateMode = PER_DAY;
   do
   {
     GetStartEndPeriod (StartDate + SECONDS_OF_A_DAY, 0, 
                        DateMode, pSList->DateList->Offset,
                        &StartPeriod, &EndPeriod);
     if ((StartPeriod < FirstDate) && (EndPeriod <= pCmdLinOpts->before)) 
       FirstDate = StartPeriod;
   } while (DateMode-- > PER_DAY);
   if (FirstDate < pCmdLinOpts->before) pCmdLinOpts->since = FirstDate;
#if COUNT_ALLOC_MEMORY >= 2
   fprintf (stderr,"starting with %s",ctime(&FirstDate));
#endif
 }
 NumDays = (pCmdLinOpts->before - pCmdLinOpts->since) / SECONDS_OF_A_DAY;
 if (!pSList->DateList->NumMembers && (NumDays < DEFAULT_DATE_NUM))
 {
   pSList->SiteList->DimCountAr = NumDays;
   pSList->ItemList->DimCountAr = NumDays;
   pSList->MethodList->DimCountAr = NumDays;
   pSList->ProtocolList->DimCountAr = NumDays;
   pSList->StatusList->DimCountAr = NumDays;
 }
 return;
}


