/*
   (C) Copyright 1996  Michael Stenns
   
   Permission to use, copy, modify, and distribute this program without fee 
   is hereby granted, provided that this copyright and permission notice 
   appear on all copies and supporting documentation.

   All sourcefiles starting with "gd" are part or Thomas Boutell's graphic
   library gd1.2. gd 1.2 is copyright 1994, 1995, Quest Protein Database 
   Center, Cold Spring Harbor Labs.

   If you want use the gd1.2 library, remark that the files provided with
   this package are slightly modified. You should get the original set from
   http://www.boutell.com/gd/ .

*/
/***************************************************************************/
#define GRAPHICS_MODUL


#include <math.h>
#include "accstat.h"
#include "fonts.h"
#include "gd.h"

/* string constants */

#ifndef GRAPHIC_INVALID_COLOR
#define GRAPHIC_INVALID_COLOR "WARNING, invalid color \"%6s\" specified\n"
#endif

#ifndef GRAPHIC_INVALID_TYPE
#define GRAPHIC_INVALID_TYPE "WARNING, unsupported source \"%s\" for image\n"
#endif

#ifndef IMAGE_YAXIS_DESCRIPTION
#define IMAGE_YAXIS_DESCRIPTION "Accesses"
#endif

#define IMAGE_MIN_WIDTH 20      /* pixel */
#define IMAGE_MAX_WIDTH 10000   /* pixel */
#define IMAGE_MIN_HEIGH 20      /* pixel */
#define IMAGE_MAX_HEIGH 10000   /* pixel */

/* declarations */

enum CHART_MODE {LINE_CH, PIE_CH, BAR_CH};
typedef enum CHART_MODE ChartType;

typedef struct 
        { uInt      width;          /* width of the image in pixel */
          uInt      heigh;          /* heigh of the image in pixel */
          uInt      Ymin;           /* for axis scaling */
          uInt      Ymax;           /* for axis scaling */
          ChartType chart;          /* chart type */
          char     *source;         /* datelist, stringlist etc. */
          uInt      max;            /* max. number of items */
          char      col_back[7];    /* background color */
          char      col_trans[7];   /* transparent color */
          char      col_string[7];  /* string color */
          char      col_line[7];    /* line color */
          char      col_border[7];  /* border color */
          char      XAxisText[48];  /* x axis description */
          char      YAxisText[48];  /* y axis description */
          uInt      numValues;      /* number of items */
          uInt     *ValueAr;        /* array with num elements */
          time_t   *DateAr;         /* filled only for date lists */
          char    **StringAr;       /* same for string lists */
          char     *DateFormat;     /* format to use in strftime */
          int       WeekOffset;     /* defines 1st day of the week */
          DateType  Period;         /* daily, weekly ... */
          SortType  Sort;           /* sorting of stringlists */
          time_t    StartPeriod;    /* first day in period */
          time_t    EndPeriod;      /* last day in period */
          char    **Strings;        /* NULL pointer terminated list */
        } ImageInfoType;

/* prototypes */



ImageInfoType * FillImageInfo (FILE * ConfigFile, size_t NumLines);
int IsColor (char * pszRGBColor);
int SetGraphicsValues (StandardListType * pSList, 
                       CommandLineType  * pCmdLinOpts,
                       ConfigFileType   * pConfigOpts,
                       ImageInfoType    * ImageInfo);

/* graphic functions */
void PrintPolyLine (gdImagePtr ImagePtr, char * pszColor,
                    uInt *ValueAr, uInt NumValues,
                    uInt YMin, uInt YMax, int YStart, int YEnd,
                    int XStart, int XEnd);

void PrintBars (gdImagePtr ImagePtr, gdFontPtr LabelFont,
                char * pszLineColor, char * pszFontColor,
                uInt *ValueAr, uInt NumValues,
                char **StringAr, time_t *DateAr, 
                char *DateFormat, int WeekOffset,
                uInt YMin, uInt YMax, int YStart, int YEnd,
                int XStart, int XEnd);

void PrintYAxisLabels (gdImagePtr ImagePtr, char * pszColor, gdFontPtr LabelFont,
                       uInt YMin, uInt YMax, int YStart, int YEnd, int XStart);
void PrintXAxisLabels (gdImagePtr ImagePtr, char * pszColor, gdFontPtr LabelFont,
                       int YStart, int XStart, int XEnd, int NumValues);
void PrintAxisDescriptions (gdImagePtr ImagePtr, char * pszColor, gdFontPtr AxisFont,
                            char * XAxisText, char * YAxisText, int YHeigh,
                            int YStart, int YEnd, int XStart, int XEnd);

int AllocateColor (gdImagePtr ImagePtr, char * pszColor);
int GetNextColor (gdImagePtr ImagePtr);
void DrawImage (FILE * ImageFile, ImageInfoType * ImageInfo);

/***************************************************************************
 * CreateImages creates all images specified in the configuration file
 ***************************************************************************/
void CreateImages   (StandardListType * pSList, 
                     CommandLineType  * pCmdLinOpts,
                     ConfigFileType   * pConfigOpts)
{
FILE * ImageFile;
char ImageFilename[FILENAME_MAX + 1];
ImageInfoType * ImageInfo;
uInt NumPeriod = 0, NumFiles = 0;
time_t StartPeriod, EndPeriod;
int Status = FALSE;
int i;


 if (!pConfigOpts || !pConfigOpts->NumImages) return; /* basic checks */


 for (i = 0; pConfigOpts->ImageList[i]; i++, NumPeriod = 0, NumFiles = 0)
 {
   if ((strlen(pConfigOpts->OutPath) + 
        strlen(pConfigOpts->ImageList[i]->FileName)) <= FILENAME_MAX)
   {
     sprintf (ImageFilename,"%s%s",pConfigOpts->OutPath,
                                pConfigOpts->ImageList[i]->FileName);
   }
   else
   {
     fprintf (stderr, OUTFIL_MAXLEN_ERROR, pConfigOpts->OutPath, 
                          pConfigOpts->ImageList[i]->FileName,FILENAME_MAX);
     continue;   /* wrong filename */
   }
#if COUNT_ALLOC_MEMORY >= 2
     fprintf (stderr,"%s\n",ImageFilename);
#endif
   /* positions the config file*/
   Status = fseek (pCmdLinOpts->ConfigFile, 
                   pConfigOpts->ImageList[i]->FileSection->Offset, SEEK_SET);
   if (Status != 0)  continue;   /* skip this section */
   ImageInfo = FillImageInfo (pCmdLinOpts->ConfigFile, 
                              pConfigOpts->ImageList[i]->FileSection->Lenght);
   ImageInfo->Period = pConfigOpts->ImageList[i]->Period;
   while (SetOutputFilename (&NumPeriod, &NumFiles, ImageFilename, &StartPeriod, 
                             &EndPeriod, pConfigOpts->ImageList[i] ,pSList))
   {
     ImageInfo->StartPeriod = StartPeriod;
     ImageInfo->EndPeriod   = EndPeriod;
     ImageInfo->WeekOffset  = pSList->DateList->Offset;
     if (!(ImageFile = fopen (ImageFilename,"wb")))
     {
       fprintf (stderr, OUTFIL_OPEN_ERROR, ImageFilename);
       continue;
     }
     if (SetGraphicsValues (pSList, pCmdLinOpts, pConfigOpts, ImageInfo))
     {
       DrawImage (ImageFile, ImageInfo);
     }
     if (ImageInfo->source)     free (ImageInfo->source);
     if (ImageInfo->ValueAr)    free (ImageInfo->ValueAr);
     if (ImageInfo->DateAr)     free (ImageInfo->DateAr);
     if (ImageInfo->DateFormat) free (ImageInfo->DateFormat);
     if (ImageInfo->StringAr)   free (ImageInfo->StringAr);
     free (ImageInfo);
     ImageInfo = NULL;
     fclose (ImageFile);
   }
 }
 return;
}


/***************************************************************************
 *
 ***************************************************************************/
void DrawImage (FILE * ImageFile, ImageInfoType * ImageInfo)
{
gdImagePtr ImagePtr;
/* define draw area */
gdFontPtr XLabelFont, YLabelFont, AxisFont;
int iColor;
int iDrawX1 = 2;
int iDrawX2 = ImageInfo->width - 2;
int iDrawY1 = ImageInfo->heigh - 2;
int iDrawY2 = 1;

 /* initialize start values */
 if ((ImageInfo->width > 100) && (ImageInfo->heigh > 100))
 {
   YLabelFont = getgdFontLarge ();
   AxisFont = getgdFontGiant ();
   if (ImageInfo->chart == LINE_CH)
     XLabelFont = getgdFontGiant ();
   else
     XLabelFont = getgdFontSmall ();
 }
 else
 {
   XLabelFont = YLabelFont = getgdFontSmall ();
   AxisFont = getgdFontMediumBold ();
 }
 iDrawX1 += (AxisFont->h > YLabelFont->h) ? AxisFont->h : YLabelFont->h;
 iDrawY1 -= (AxisFont->h > XLabelFont->h) ? AxisFont->h : XLabelFont->h;
 if (((ImageInfo->width * 0.2) < iDrawX1) || ((ImageInfo->heigh * 0.2) > iDrawY1))
 {  /* on very small pictures avoid axis text */
   iDrawX1 = 1;
   iDrawY1 = ImageInfo->heigh - 2;
   AxisFont = XLabelFont = YLabelFont = NULL;
 }
 /* allocate image memory */
 ImagePtr = gdImageCreate(ImageInfo->width, ImageInfo->heigh);
 AllocateColor (ImagePtr, ImageInfo->col_back);  /* background color */
 gdImageColorTransparent (ImagePtr, AllocateColor (ImagePtr, ImageInfo->col_trans));

#if 0
 iColor = AllocateColor (ImagePtr, ImageInfo->col_line);
 gdImageRectangle (ImagePtr, 0, 0, ImageInfo->width -1, ImageInfo->heigh -1, iColor);
#endif
 if (ImageInfo->chart == LINE_CH)
 {
   PrintYAxisLabels (ImagePtr, ImageInfo->col_string, YLabelFont,
                     ImageInfo->Ymin, ImageInfo->Ymax, 
                     iDrawY1, iDrawY2, iDrawX1);
   PrintXAxisLabels (ImagePtr, ImageInfo->col_string, XLabelFont,
                     iDrawY1, iDrawX1, iDrawX2, ImageInfo->numValues);
   PrintAxisDescriptions (ImagePtr, ImageInfo->col_string, AxisFont,
                          ImageInfo->XAxisText, ImageInfo->YAxisText, ImageInfo->heigh,
                          iDrawY1, iDrawY2, iDrawX1, iDrawX2);
   PrintPolyLine (ImagePtr, ImageInfo->col_line, ImageInfo->ValueAr, 
                ImageInfo->numValues, ImageInfo->Ymin, ImageInfo->Ymax,
                iDrawY1, iDrawY2+1, iDrawX1, iDrawX2);
   if (XLabelFont) /* picture not too small */
   {
     PrintPolyLine (ImagePtr, ImageInfo->col_line, ImageInfo->ValueAr, 
                  ImageInfo->numValues, ImageInfo->Ymin, ImageInfo->Ymax,
                  iDrawY1-1, iDrawY2, iDrawX1, iDrawX2);
   }
   iColor = AllocateColor (ImagePtr, ImageInfo->col_border);
   gdImageRectangle (ImagePtr, iDrawX1, iDrawY1, iDrawX2, iDrawY2, iColor);
 }
 else if (ImageInfo->chart == BAR_CH)
 {
   if (AxisFont && XLabelFont)
   { /* substract both AxisFont->h and YLabelFont->h */
     iDrawY1 -= (AxisFont->h < XLabelFont->h) ? AxisFont->h : XLabelFont->h;
   }
   PrintYAxisLabels (ImagePtr, ImageInfo->col_string, YLabelFont,
                     ImageInfo->Ymin, ImageInfo->Ymax, 
                     iDrawY1, iDrawY2, iDrawX1);
   PrintAxisDescriptions (ImagePtr, ImageInfo->col_string, AxisFont,
                          ImageInfo->XAxisText, ImageInfo->YAxisText, ImageInfo->heigh,
                          iDrawY1, iDrawY2, iDrawX1, iDrawX2);
   PrintBars   (ImagePtr, XLabelFont,
                ImageInfo->col_line, ImageInfo->col_string, 
                ImageInfo->ValueAr, ImageInfo->numValues,
                ImageInfo->StringAr, ImageInfo->DateAr, 
                ImageInfo->DateFormat, ImageInfo->WeekOffset,
                ImageInfo->Ymin, ImageInfo->Ymax,
                iDrawY1, iDrawY2+1, iDrawX1, iDrawX2);
   iColor = AllocateColor (ImagePtr, ImageInfo->col_border);
   gdImageRectangle (ImagePtr, iDrawX1, iDrawY1, iDrawX2, iDrawY2, iColor);
 }
 /* print image  */
 gdImageGif (ImagePtr, ImageFile);
 /* free memory */
 gdImageDestroy(ImagePtr);
 return;
}


#define BAR_MIN_WIDTH 8     /* pixel */
#define BAR_DEF_DISTANCE 6  /* pixel */
/***************************************************************************
 * 
 ***************************************************************************/
void PrintBars (gdImagePtr ImagePtr, gdFontPtr LabelFont,
                char * pszLineColor, char * pszFontColor,
                uInt *ValueAr, uInt NumValues,
                char **StringAr, time_t *DateAr, 
                char *DateFormat, int WeekOffset,
                uInt YMin, uInt YMax, int YStart, int YEnd,
                int XStart, int XEnd)
{
double PixelDiff = (abs(YEnd - YStart) - 1);
double ValueDiff = fabs((YMax - YMin));
double XDiff = (XEnd - XStart) - BAR_DEF_DISTANCE;
char Buffer[16];
int LineColor = AllocateColor (ImagePtr, pszLineColor);
int FontColor = AllocateColor (ImagePtr, pszFontColor);
int BarWidth, MaxChar = 0;
uInt Y;
int i, X1, X2, Y1, Y2;

 
 GetNextColor (NULL); /* initialize */
 NumValues = (NumValues < (XDiff / (BAR_MIN_WIDTH + BAR_DEF_DISTANCE))) ? 
              NumValues : (XDiff / (BAR_MIN_WIDTH + BAR_DEF_DISTANCE));
 if (NumValues && (ValueDiff > 1e-6))
 {
  XDiff -= NumValues * BAR_DEF_DISTANCE;
  BarWidth = (XDiff / NumValues);
  if (LabelFont) 
  {
    MaxChar = BarWidth / LabelFont->w;
    if (MaxChar < 3)
    {
      LabelFont = getgdFontTiny();
      MaxChar = BarWidth / LabelFont->w;
    }
  }
  if (MaxChar > 15) MaxChar = 15;
  Y1 = YStart;
  XStart += BAR_DEF_DISTANCE;
  for (i = 0; i < NumValues; i++)
  {
    Y = (ValueAr[i] < YMin) ? YMin : (ValueAr[i] > YMax) ? YMax : ValueAr[i];
    X1 = XStart + i * (XDiff / NumValues + BAR_DEF_DISTANCE);
    Y2 = YEnd + (PixelDiff - (PixelDiff / ValueDiff) * Y);
    X2 = X1 + BarWidth;
    gdImageFilledRectangle (ImagePtr, X1, Y2, X2, Y1, GetNextColor(ImagePtr));
    gdImageRectangle (ImagePtr, X1, Y2, X2, Y1, LineColor);
    if (LabelFont && MaxChar)
    {
      if (StringAr)
      {
        strncpy (Buffer, StringAr[i], MaxChar);
      }
      else if (DateAr && DateFormat)
      {
        PrintDate (Buffer, MaxChar, DateAr[i], DateFormat, WeekOffset);
        Buffer[MaxChar] = '\0';
      }
      else
      {
        Buffer[0] = '\0';
      }
      X1 = X1 + (double)BarWidth/2.0 - (double)(strlen(Buffer) * LabelFont->w) / 2.0;
      gdImageString (ImagePtr, LabelFont, X1, YStart + 1, Buffer, FontColor);
    }
  }
 }
 return;
}


/***************************************************************************
 *
 ***************************************************************************/
int GetNextColor (gdImagePtr ImagePtr)
{
static int ind = 0;
static char * ColorMap[] = {
"FF0000","FFFF00","0000FF","FF8000","00FF00","00FFFF","FF00FF",
"FF8080","FFFF80","8080FF","FFC080","80FF80","80FFFF","FF80FF",
"C08080","C0C000","0000C0","C0A000","00C000","00C0C0","C000C0",NULL};
#if 0
static char * ColorMap[] = {"FF0000","FFFF00","00FF00","00FFFF",
                            "0080FF","FF80C0","FF8040","808040",
                            "0000FF","FF00FF","008040","804000",NULL};
#endif
int color = -1;

 if (!ImagePtr)
 {
   ind = 0;
 }
 else
 {
   if (!ColorMap[ind]) ind = 0;
   color = AllocateColor (ImagePtr, ColorMap[ind]);
   if (color < 0) color = ind;  /* color palette overflow */
   ind++;
 }

 return color;
}

/***************************************************************************
 *
 ***************************************************************************/
void PrintAxisDescriptions (gdImagePtr ImagePtr, char * pszColor, gdFontPtr AxisFont,
                            char * XAxisText, char * YAxisText, int YHeigh,
                            int YStart, int YEnd, int XStart, int XEnd)
{
int Color = AllocateColor (ImagePtr, pszColor);
int X1, Y1;

 if (AxisFont)
 {
   if (YAxisText)
   {
     X1 = 1;
     Y1 = (YStart - YEnd) / 2 + (strlen (YAxisText) / 2) *  AxisFont->w;
     if ((X1 >= 0) && (Y1 < (YStart * 0.8)))
     {
       gdImageStringUp (ImagePtr, AxisFont, X1, Y1, YAxisText, Color);
     }
   }
   if (XAxisText)
   {
     Y1 = YHeigh - AxisFont->h - 1;
     X1 = (XEnd - XStart) / 2 - (strlen (XAxisText) / 2) *  AxisFont->w;
     if (X1 > XStart)
     {
       gdImageString (ImagePtr, AxisFont, X1, Y1, XAxisText, Color);
     }
   }
 }
 return;
}

/***************************************************************************
 *
 ***************************************************************************/
void PrintXAxisLabels (gdImagePtr ImagePtr, char * pszColor, gdFontPtr LabelFont,
                       int YStart, int XStart, int XEnd, int NumValues)
{
int Color = AllocateColor (ImagePtr, pszColor);
char Buf[16];
int X1, Y1;

 if (LabelFont)
 {
   sprintf (Buf,"%u", 1);
   X1 = XStart;
   Y1 = YStart + 1;
   gdImageString (ImagePtr, LabelFont, X1, Y1, Buf, Color);
   sprintf (Buf,"%u", NumValues);
   X1 = XEnd - strlen(Buf) * LabelFont->w;
   gdImageString (ImagePtr, LabelFont, X1, Y1, Buf, Color);
 }
 return;
}

/***************************************************************************
 *
 ***************************************************************************/
void PrintYAxisLabels (gdImagePtr ImagePtr, char * pszColor, gdFontPtr LabelFont,
                       uInt YMin, uInt YMax, int YStart, int YEnd, int XStart)
{
int Color = AllocateColor (ImagePtr, pszColor);
char Buf[16];
int X1, Y1;

 if (LabelFont)
 {
   sprintf (Buf,"%u",YMin);
   X1 = XStart - LabelFont->h;
   Y1 = YStart;
   if ((X1 >= 0) && (Y1 >= 0))
   {
     gdImageStringUp (ImagePtr, LabelFont, X1, Y1, Buf, Color);
   }
   sprintf (Buf,"%u",YMax);
   Y1 = YEnd + strlen(Buf) * LabelFont->w;
   if ((X1 >= 0) && (Y1 >= 0))
   {
     gdImageStringUp (ImagePtr, LabelFont, X1, Y1, Buf, Color);
   }
 }
 return;
}

/***************************************************************************
 * PrintPolyLine prints a line described by the array ValueAr into the image
 ***************************************************************************/
void PrintPolyLine (gdImagePtr ImagePtr,  char * pszColor,
                    uInt *ValueAr, uInt NumValues,
                    uInt YMin, uInt YMax, int YStart, int YEnd,
                    int XStart, int XEnd)
{
double PixelDiff = (abs(YEnd - YStart) - 1);
double ValueDiff = fabs((YMax - YMin));
double dStep;
int Color = AllocateColor (ImagePtr, pszColor);
uInt Y;
int i, X1, X2, Y1, Y2;

 if (ValueDiff > 1e-6)
 {
  dStep = (double)(XEnd - XStart - 1);
  if (NumValues > 1) dStep /= (NumValues - 1);
  for (i = 0, Y1 = 0; i < NumValues; i++, Y1 = Y2)
  {
    Y = (ValueAr[i] < YMin) ? YMin : (ValueAr[i] > YMax) ? YMax : ValueAr[i];
    Y2 = YEnd + (PixelDiff - (PixelDiff / ValueDiff) * Y);
    if (i)
    {
      X1 = XStart + (i - 1) * dStep + 1;
      X2 = XStart + i * dStep + 1;
      gdImageLine(ImagePtr, X1, Y1, X2, Y2, Color);
    }
    else if (NumValues == 1)
    {
      gdImageLine(ImagePtr, XStart, Y2, XEnd, Y2, Color);
    }
  }
 }
 return;
}


/***************************************************************************
 * AllocateColor allocates a color described in a 6 octed hex. string
 * and return it's index. Allocating the same color more than once 
 * is avoided.
 ***************************************************************************/
int AllocateColor (gdImagePtr ImagePtr, char * pszRGB)
{
int red = 0, green = 0, blue = 0;
int ColorIndex = -1;

 if (*pszRGB)  /* color has been specified */
 {
   sscanf (pszRGB, "%2x%2x%2x", &red, &green, &blue);
   ColorIndex = gdImageColorExact(ImagePtr, red, green, blue);
   if (ColorIndex < 0)
   {
     ColorIndex = gdImageColorAllocate (ImagePtr, red, green, blue);
   }
 }
 
 return ColorIndex;
}


/***************************************************************************
 *
 ***************************************************************************/
int SetGraphicsValues (StandardListType * pSList, 
                       CommandLineType  * pCmdLinOpts,
                       ConfigFileType   * pConfigOpts,
                       ImageInfoType    * ImageInfo)
{
int iReturn = TRUE;
TableHeadInfoType tblHead;
uInt   CurrentCount = 0;
ByteCountType CurrentBytes = {0,0};
time_t CurrentPeriod = 0;
int NumTableMembers;
uInt Ymax = 0;
char * pszPeriod;
StringListType * StringList;
int i;

 /*
  * =================================
  *            TimeList
  * =================================
  */
 if (!strcmp(ImageInfo->source,"time"))
 {
   int Resolution = TIMLIST_DIM / ImageInfo->max;

   if (Resolution < 1) Resolution = 1;
   ImageInfo->numValues = TIMLIST_DIM / Resolution;
   ImageInfo->ValueAr = my_calloc (ImageInfo->numValues, sizeof (uInt));
   for (i=0; i < TIMLIST_DIM ; i++)
   {
     ImageInfo->ValueAr [i/Resolution] += pSList->TimeList[i];
   }
   for (i=0; i < ImageInfo->numValues; i++)
   {
      if (Ymax < ImageInfo->ValueAr[i]) Ymax = ImageInfo->ValueAr[i];
   }
 }
 /*
  * =================================
  *            DateList
  * =================================
  */
 else if (!strcmp(ImageInfo->source,"date"))
 {
   DateListType * DateList;
   int Indices[32], MaxIndices;

   DateList = InitTableInfo (&tblHead, pSList, pCmdLinOpts->HideList, 
                             ImageInfo->source, NO_KEY,
                             ImageInfo->max, ImageInfo->Period);

   if (ImageInfo->Period == PER_MONTH)
   {
     MaxIndices = 32;
     pszPeriod = "Month";
   }
   else if (ImageInfo->Period == PER_WEEK)
   {
     MaxIndices = 8;
     pszPeriod = "Weeks";
   }
   else
   {
     MaxIndices = 2;
     pszPeriod = "Days";
   }

   ImageInfo->ValueAr = my_calloc (ImageInfo->max, sizeof (uInt));
   ImageInfo->DateAr = my_calloc (ImageInfo->max, sizeof (time_t));

   /* adjust date table for the current period */
   NumTableMembers = tblHead.NumMembers;
   CurrentPeriod = MaxDate (DateList);
   /* only values from history file used: */
   if (!ImageInfo->EndPeriod) ImageInfo->EndPeriod = CurrentPeriod;
   for (i = tblHead.NumMembers; i && (CurrentPeriod > ImageInfo->EndPeriod); i--)
   {
     CurrentPeriod =  GetPeriodIndices (Indices, 1, i, 
                                DateList, ImageInfo->Period);
     if (CurrentPeriod > ImageInfo->EndPeriod) NumTableMembers--;
   }
   for (i = (tblHead.ShowMembers && (NumTableMembers > tblHead.ShowMembers)) ?
             NumTableMembers - tblHead.ShowMembers : 0 ; 
        i < NumTableMembers; i++)
   {             /* evaluate startup values */
     CurrentPeriod =  GetPeriodIndices (Indices, MaxIndices, i + 1, 
                                DateList, ImageInfo->Period);
     EvalDateListPeriod (&Indices[1], Indices[0], DateList,
                         &CurrentCount, &CurrentBytes);
     ImageInfo->ValueAr[ImageInfo->numValues] = CurrentCount;
     ImageInfo->DateAr[ImageInfo->numValues] = CurrentPeriod;
     ImageInfo->numValues++;
     if (Ymax < CurrentCount) Ymax = CurrentCount;
   }
   if (!*ImageInfo->XAxisText)
   {
     strncpy (ImageInfo->XAxisText, pszPeriod, sizeof ImageInfo->XAxisText - 1);
   }
   if (!*ImageInfo->YAxisText)
   {
     strncpy (ImageInfo->YAxisText, IMAGE_YAXIS_DESCRIPTION, sizeof ImageInfo->YAxisText - 1);
   }
 }
 /*
  * =================================
  *           StringList
  * =================================
  */
 else if ((StringList = SelectStringList (pSList, NO_KEY, ImageInfo->source)))
 {
   InitTableInfo (&tblHead, pSList, pCmdLinOpts->HideList, 
                  ImageInfo->source, StringList->Key,
                  ImageInfo->max, NO_DATE);

   ImageInfo->ValueAr  = my_calloc (ImageInfo->max, sizeof (uInt));
   ImageInfo->StringAr = my_calloc (ImageInfo->max, sizeof (char *));

   if (ImageInfo->Sort == SORT_ALPHABETICAL)
   {
     qsort (StringList->List , StringList->NumMembers, 
            sizeof (void *), ItemListStrComp);
   }
   else if (ImageInfo->Sort == SORT_NUMERICAL)
   {
     qsort (StringList->List , StringList->NumMembers, 
            sizeof (void *), StringListCountComp);
   }

   for (i = 0; (ImageInfo->numValues < tblHead.ShowMembers) &&
                                  (i < StringList->NumMembers); i++)
   {
     if (CheckAgainstHideList (pCmdLinOpts->HideList, StringList->Key, 
                               StringList->List[i]->pszStr)  
         && StringList->List[i]->Count)
     {
     ImageInfo->ValueAr[ImageInfo->numValues]  = StringList->List[i]->Count;
     ImageInfo->StringAr[ImageInfo->numValues] = StringList->List[i]->pszStr;
     ImageInfo->numValues++;
     if (Ymax < StringList->List[i]->Count) Ymax = StringList->List[i]->Count;
     }
   }
 }
 else
 {
   fprintf (stderr, GRAPHIC_INVALID_TYPE, ImageInfo->source);
   iReturn = FALSE;
 }
 /* scale y values */
 if (Ymax && !ImageInfo->Ymax) 
 {                         /* scale max value */
   int Pow  = nDigits(Ymax) - 2;
   uInt Scale;
   int Adjust;
   for (i = 0, Adjust = 1; i < Pow; i++) Adjust *= 10;
   if (Adjust && (Ymax % Adjust))
   {
     Scale = Adjust * (1 + Ymax / Adjust);
   }
   else
   {
     Scale = Ymax;
   }
   ImageInfo->Ymax = Scale;
 }
 return iReturn;
}



/***************************************************************************
 * ImageInfoType reads the infos from the section CreateImage in the
 * configuration file and fills the structure ImageInfo with the informations
 * found there.
 ***************************************************************************/
ImageInfoType * FillImageInfo (FILE * ConfigFile, size_t NumLines)
{
ImageInfoType * pImageInfo = my_calloc (1, sizeof(ImageInfoType));
char *pszRecord;            /* record from the configuration file */
char * pszCondition, * pszValue;
int i;


 /* default values */
 pImageInfo->source   = pImageInfo->DateFormat = NULL;
 pImageInfo->ValueAr  = NULL;
 pImageInfo->DateAr   = NULL;
 pImageInfo->StringAr = NULL;
 pImageInfo->chart = LINE_CH;
 pImageInfo->Sort = SORT_NOSORT;
 strcpy (pImageInfo->col_back,"FFFFFF");    /* white */
 strcpy (pImageInfo->col_string,"0000FF"); /* blue */
 strcpy (pImageInfo->col_line, "FF0000");   /* red */
 strcpy (pImageInfo->col_border, "FF0000");   /* red */

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));

 for (i = 0; i < NumLines; i++)
 {
   fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
   pszCondition = Truncate_String (pszRecord);
   pszValue     = ParseConfigLine (pszCondition);
   pszCondition = StringToLower (pszCondition);
   if (pszValue)
   {
     if (!strcmp (pszCondition,"width"))                      /* Width */
     {
       pImageInfo->width = abs(atoi(pszValue));
     }
     else if (!strcmp (pszCondition,"heigh"))                 /* Heigh */
     {
       pImageInfo->heigh = abs(atoi(pszValue));
     }
     else if (!strcmp (pszCondition,"source"))                /* source */
     {
       if (!pImageInfo->source)
       {
         char * pszDateFormat;

         pszValue = Truncate_Config_File_Record (pszValue);
         pszDateFormat = ParseConfigLine (pszValue);
         pImageInfo->source = my_malloc ((strlen(pszValue)+1) * sizeof(char));
         strcpy (pImageInfo->source, pszValue);
         if (pszDateFormat)
         {
           pImageInfo->DateFormat = my_malloc ((strlen(pszDateFormat)+1)
                                                * sizeof(char));
           strcpy (pImageInfo->DateFormat, pszDateFormat);
         }
       }
     }
     else if (!strcmp (pszCondition,"sorting"))             /* Sorting */
     {
       pszValue = StringToLower (pszValue);
       if (pszValue[0] == 'a')
       {
         pImageInfo->Sort = SORT_ALPHABETICAL;
       }
       else if (pszValue[0] == 'n')
       {
         pImageInfo->Sort = SORT_NUMERICAL;
       }
       else if (pszValue[0] == 'u')
       {
         pImageInfo->Sort = SORT_NOSORT;
       }
    }
    else if (!strcmp (pszCondition,"showmember"))            /* ShowMember */
     {
       pImageInfo->max = abs(atoi(pszValue));
     }
     else if (!strcmp (pszCondition,"xaxistitle"))            /* XAxisTitle */
     {
       strncpy (pImageInfo->XAxisText, pszValue, sizeof pImageInfo->XAxisText - 1);
     }
     else if (!strcmp (pszCondition,"yaxistitle"))            /* YAxisTitle */
     {
       strncpy (pImageInfo->YAxisText, pszValue, sizeof pImageInfo->YAxisText - 1);
     }
     else if (!strcmp (pszCondition,"charttype"))             /* ChartType */
     {
       if (!strncmp (pszValue, "line", sizeof "line" - 1))
         pImageInfo->chart = LINE_CH;
       else if (!strncmp (pszValue, "bar", sizeof "bar" - 1))
         pImageInfo->chart = BAR_CH;
       else if (!strncmp (pszValue, "pie", sizeof "pie" - 1))
         pImageInfo->chart = PIE_CH;
     }
     else if (strstr (pszCondition, "color"))
     {
       pszValue = Truncate_Config_File_Record (pszValue);
       if (IsColor(pszValue))
       {
         if (!strcmp (pszCondition, "backgroundcolor"))       /* BackgroundColor */
           strcpy (pImageInfo->col_back, pszValue);
         else if (!strcmp (pszCondition, "transparentcolor")) /* TransparentColor */
           strcpy (pImageInfo->col_trans, pszValue);
         else if (!strcmp (pszCondition, "textcolor"))      /* TextColor */
           strcpy (pImageInfo->col_string, pszValue);
         else if (!strcmp (pszCondition, "linecolor"))        /* LineColor */
           strcpy (pImageInfo->col_line, pszValue);
         else if (!strcmp (pszCondition, "bordercolor"))      /* BorderColor */
           strcpy (pImageInfo->col_border, pszValue);
       }
       else
       {
          fprintf (stderr, GRAPHIC_INVALID_COLOR , pszValue);
       }
     }
   }
 }
 /* consistency checks */
 if (pImageInfo->width < IMAGE_MIN_WIDTH) pImageInfo->width = IMAGE_MIN_WIDTH;
 if (pImageInfo->heigh < IMAGE_MIN_HEIGH) pImageInfo->heigh = IMAGE_MIN_HEIGH;
 if (pImageInfo->width > IMAGE_MAX_WIDTH) pImageInfo->width = IMAGE_MAX_WIDTH;
 if (pImageInfo->heigh > IMAGE_MAX_HEIGH) pImageInfo->heigh = IMAGE_MAX_HEIGH;


 free (pszRecord);
 return pImageInfo;
}


/***************************************************************************
 * IsColor looks if pszRGBColor is a valid RGB color in hex. form
 ***************************************************************************/
int IsColor (char * pszRGBColor)
{
int isColor = TRUE;
int i;

 if (strlen (pszRGBColor) >= 6)
 {
  for (i = 0; i < 6; i++)
  {
    if (isxdigit(pszRGBColor[i]))  pszRGBColor[i] = toupper (pszRGBColor[i]);
    else                           isColor = FALSE;
  }
  pszRGBColor[6] = '\0';
 }
 else
 {
  isColor = FALSE;
 }
 return isColor;
}
