/*
   (C) Copyright 1996  Michael Stenns
   
   Permission to use, copy, modify, and distribute this program without fee 
   is hereby granted, provided that this copyright and permission notice 
   appear on all copies and supporting documentation.

   All sourcefiles starting with "gd" are part or Thomas Boutell's graphic
   library gd1.2. gd 1.2 is copyright 1994, 1995, Quest Protein Database 
   Center, Cold Spring Harbor Labs.

   If you want use the gd1.2 library, remark that the files provided with
   this package are slightly modified. You should get the original set from
   http://www.boutell.com/gd/ .

*/
/***************************************************************************/
/* 
 * some modules needed by accstat
*/

#define MODULES_C


#include "accstat.h"

typedef struct { void * ptr;    /* address of the pointer to be freed */
                 size_t size;   /* and it's allocated size */
               } FreeInfoType;


#ifdef MEMPAGE_ACCOUNT
uInt MemPageAccesses = 0;
uInt MemPageBytes = 0;
#endif

#ifndef ALLOC_MEMORY_FAIL
#define ALLOC_MEMORY_FAIL "Error getting Memory after allocating %u bytes in %u blocks\n"
#endif



/***************************************************************************
 * mkDay converts a time structure into the number of seconds 
 * since 00:00:00 GMT 1/1/1970.
 * The value is shortend to a full date (midnight) value.
 ***************************************************************************/
time_t mkDay (struct tm *tptr)
{
return (time_t) DaysSince1970 (tptr->tm_mday,  tptr->tm_mon + 1, 
                tptr->tm_year + 1900) * SECONDS_OF_A_DAY;
}

/***************************************************************************
 * DateListInsert insets memebers in the date list
 ***************************************************************************/
int DateListInsert (DateListType *pDatL, time_t Date, unsigned int uiBytes)
{
int iDateNum;        /* current date number, return value */
int i, iNotFound;

 for (i=0, iNotFound=TRUE;i<pDatL->NumMembers && iNotFound;i++)
 {
    iNotFound = pDatL->Date[i] - Date;
 }
 if (iNotFound)
 {   /* new element */
   iDateNum = pDatL->NumMembers;
   if (pDatL->NumMembers >= pDatL->MaxMembers)
   {        /* extend list */
     pDatL->MaxMembers += pDatL->DefaultExtend;
     pDatL->Date = (time_t *) my_recalloc (pDatL->Date, 
                    pDatL->NumMembers * sizeof (time_t),
                    pDatL->MaxMembers * sizeof (time_t));  
     pDatL->uiCountAr = (unsigned int *) my_recalloc (pDatL->uiCountAr, 
                         pDatL->NumMembers * sizeof (unsigned int),
                         pDatL->MaxMembers * sizeof (unsigned int));  
     pDatL->uiByteAr = (unsigned int *) my_recalloc (pDatL->uiByteAr, 
                         pDatL->NumMembers * sizeof (unsigned int),
                         pDatL->MaxMembers * sizeof (unsigned int));  
   }
   pDatL->Date[iDateNum] = Date;
   pDatL->uiByteAr[iDateNum] += uiBytes;
   pDatL->uiCountAr[iDateNum]++;
   pDatL->NumMembers++;
 }
 else
 {
   iDateNum = --i;
   pDatL->uiByteAr[iDateNum] += uiBytes;
   pDatL->uiCountAr[iDateNum]++;
 }
 return iDateNum;
}


/**************************************************************************
 * TimeListInsert parses a time string (hh:mm:ss) and determines the
 * server hits as a function of the time.
 *************************************************************************/
void TimeListInsert (uInt TimeList[TIMLIST_DIM], char * pszTime)
{
uInt uiSeconds = 0; 
char * pszHour, * pszMinute, * pszSecond;
int Index;

 pszHour   = strtok (pszTime,":");
 pszMinute = strtok (NULL,":");
 pszSecond = strtok (NULL,":");

 if (pszHour && pszMinute && pszSecond)
 {
   uiSeconds = abs(atoi (pszHour) * 3600 + atoi (pszMinute) * 60 + atoi (pszSecond));
   Index = uiSeconds / (SECONDS_OF_A_DAY / TIMLIST_DIM);
   if (Index < TIMLIST_DIM) TimeList[Index]++;
 }
 return;
}


/***************************************************************************
 * StringListInsert checks pszStr against the existing members in pL. If
 * found, the appropriate counters are incremented, else a new member  
 * will be added to the list.
 ***************************************************************************/
void StringListInsert (StringListType *pL, const char *pszStr, int iDateNum)
{
int i;

if (pszStr)
{
 /* check dimension of array's */
 while ((iDateNum >= pL->DimCountAr) && (pL->Key != NO_KEY))
 {
    for (i=0;i < pL->NumMembers;i++)
    {
       pL->List[i]->uiCountAr = CountArrayAlloc (pL->List[i]->uiCountAr,
                                pL->DimCountAr * sizeof (uInt),
                                DEFAULT_DATE_EXTEND * sizeof (uInt));
    }
    pL->DimCountAr += DEFAULT_DATE_EXTEND;
#if 0
printf ("StringListInsert: reallocating uiCountAr to %u\n",(uInt)pL->DimCountAr);
#endif
 }


 if (pL->NumMembers && strcmp (pL->List[pL->iLast]->pszStr, pszStr))
 {
   size_t NumSortValues = ((pL->NumMembers-1)/STRINGLIST_SORT_INTERVAL)*STRINGLIST_SORT_INTERVAL;
   StringCountType ** pMember = bsearch (pszStr, pL->List, NumSortValues, 
                                sizeof (void *), StringListBsearchComp);

   if (pMember)
   {
     i = (pMember -  pL->List);
   }
   else
   {
     for (i = NumSortValues; i < pL->NumMembers && 
                 strcmp (pL->List[i]->pszStr, pszStr) ; i++);
   }
   pL->iLast = i;
 }
 else  /* cache hit, do not search the hole array */
 {
   i = pL->iLast;
 }
 if (i < pL->NumMembers)                                          /* found */
 {
   pL->List[i]->uiCountAr[iDateNum]++; 
 }
 else                                            /* not found, new element */
 {
   if (pL->NumMembers >= pL->MaxMembers)
   {        /* extend list */
     pL->List = AllocStringCountList (pL->List, &pL->MaxMembers, pL->DefaultExtend);
     for (i = pL->NumMembers; i < pL->MaxMembers; i++)
     {
       pL->List[i] = StringCountAlloc (pL->List[i]);
     }
   }
   pL->List[pL->NumMembers]->pszStr = string_alloc (pszStr);
   if (pL->DimCountAr)
   {
     pL->List[pL->NumMembers]->uiCountAr = CountArrayAlloc (NULL,  0,
                                              pL->DimCountAr * sizeof (uInt));
                                         
     pL->List[pL->NumMembers]->uiCountAr[iDateNum]++;
   }
   pL->iLast = pL->NumMembers;
   if (!(pL->NumMembers % STRINGLIST_SORT_INTERVAL) && pL->NumMembers)
   {
     /*
      * Do not sort the last element in the list! pL->iLast becomes undefined, 
      * which might cause some side effects.
      */
     qsort (pL->List , pL->NumMembers, sizeof (void *), ItemListStrComp);
   }
#if 0
if (!pL->NumMembers) printf ("starting with %u days\n",(uInt) pL->DimCountAr);
#endif
   pL->NumMembers++;
 }
}  /* pszStr */

}


/***************************************************************************
 *
 ***************************************************************************/
StringCountType * StringCountAlloc (StringCountType * pStringCount)
{
static MemPageType * MemPage = NULL;
                                    /* align to (StringCountType *) bounds */
static size_t AllocSize = sizeof (StringCountType) + ( 
                  (sizeof (StringCountType) % sizeof (StringCountType *)) ?
                   sizeof (StringCountType *) - 
                  (sizeof (StringCountType) % sizeof (StringCountType *)) : 0);

 if (!MemPage || 
    (((char *)MemPage->ptr - (char *)MemPage->base) < AllocSize))
 {
   if (MemPage) MemPage->ptr = MemPage->base;
   MemPage = GetMemoryPage (NULL, 0);
   memset (MemPage->base, 0, MEMPAGE_PAGESIZE);
 }
 MemPage->ptr   = (char *)MemPage->ptr - AllocSize;
 MemPage->size += AllocSize;
 pStringCount = MemPage->ptr;
#ifdef MEMPAGE_ACCOUNT
 MemPageAccesses++;
 MemPageBytes += AllocSize;
#endif

 return pStringCount;
}


/***************************************************************************
 *
 ***************************************************************************/
uInt * CountArrayAlloc (uInt * uiCountArray, size_t OldSize, size_t AddSize)
{
static MemPageType * MemPage = NULL;
size_t AllocSize = (OldSize + AddSize);

if (AllocSize)
{
 if (!MemPage || 
    (((char *)MemPage->ptr - (char *)MemPage->base) < AllocSize))
 {
   if (MemPage) MemPage->ptr = MemPage->base;
   MemPage = GetMemoryPage (NULL, 0);
   memset (MemPage->base, 0, MEMPAGE_PAGESIZE);
 }
 MemPage->ptr   = (char *) MemPage->ptr - AllocSize;
 MemPage->size += AllocSize;
 if (uiCountArray && OldSize) /* copy old values and free block */
 {
   int i = OldSize / sizeof (uInt);
   uInt * pTarget = MemPage->ptr;

   /* memcpy (MemPage->ptr, uiCountArray, OldSize); */
   while (i--) *pTarget++ = *uiCountArray++;
   GetMemoryPage (uiCountArray, OldSize); /* free alloc. memory */
 }
 uiCountArray = MemPage->ptr;
#ifdef MEMPAGE_ACCOUNT
 MemPageAccesses++;
 MemPageBytes += AllocSize;
#endif
}
else
{
 uiCountArray = NULL;
}
 return uiCountArray;
}


/***************************************************************************
 *
 ***************************************************************************/
char * string_alloc (const char * pszString)
{
static MemPageType * MemPage = NULL;
size_t StringLen;
char * pszStr = NULL;

 if (pszString)
 {
   StringLen = strlen (pszString) + 1;
   if (StringLen > MEMPAGE_PAGESIZE)
   {
     pszStr = my_malloc (StringLen * sizeof (char));
   }
   else
   {
     if (!MemPage || 
        (((char *)MemPage->ptr - (char *)MemPage->base) < StringLen))
     {
       if (MemPage) MemPage->ptr = MemPage->base;
       MemPage = GetMemoryPage (NULL, 0);
     }
     MemPage->ptr   = (char *) MemPage->ptr - StringLen;
     MemPage->size += StringLen;
     pszStr = MemPage->ptr;
   }
   strcpy (pszStr, pszString);
#ifdef MEMPAGE_ACCOUNT
MemPageAccesses++;
MemPageBytes += StringLen;
#endif
 }
 return pszStr;
}


/***************************************************************************
 *
 ***************************************************************************/
MemPageType * GetMemoryPage (void * ptr, size_t SizeAlloc)
{
static MemPageType ** PageList = NULL;
static uInt NumPages = 0;
static uInt MaxPages = 0;
static MemPageType * freeptr = NULL;
static int NumFreeptr;
MemPageType * memptr = NULL;
int i;

 if (ptr) /* free block which contains ptr */
 {
   char * tmpptr;

   for (i = 0; ptr && (i < NumPages); i++)
   {
     if (i)
       tmpptr = PageList[i]->base;
     else   /* first block is for internal use only */
       tmpptr = (freeptr) ? freeptr->base : NULL;

     if (tmpptr)
     {
/*
       if ((tmpptr <=  (char *) ptr) && 
           (tmpptr + MEMPAGE_PAGESIZE) >= (char *) ptr)
*/
       if ( (uLong)((char *)ptr - tmpptr) <= MEMPAGE_PAGESIZE )
       {
         if (i)
         {
           freeptr = PageList[i];
           NumFreeptr = i;
         }
         freeptr->size -= SizeAlloc;
         ptr = NULL;
         if (!freeptr->size && (freeptr->base == freeptr->ptr)) /* release block */
         {
           freeptr = NULL;
           memptr = PageList[NumFreeptr];
           PageList[NumFreeptr] = PageList[--NumPages];
           PageList[NumPages] = memptr;
         }
       }
     }
   }
 }
 else /* get new block */
 {
   if (NumPages >= MaxPages)
   {
     MemPageType * MemPage = NULL;
     uInt Size = sizeof (MemPageType) + ( /* align to (MemPageType *) bounds */
                 (sizeof (MemPageType) % sizeof (MemPageType *)) ?
                  sizeof (MemPageType *) - 
                 (sizeof (MemPageType) % sizeof (MemPageType *)) : 0);


     PageList = my_recalloc (PageList, NumPages * sizeof (MemPageType *), 
                                MEMPAGE_EXTEND * sizeof (MemPageType *));
     MaxPages += MEMPAGE_EXTEND;
     PageList[NumPages] = my_calloc (1, sizeof (MemPageType));
     MemPage = PageList[NumPages++];
     MemPage->base = my_calloc (MEMPAGE_EXTEND, Size);
     MemPage->ptr = (char *)MemPage->base + MEMPAGE_EXTEND * Size;
     for (i = NumPages; i < MaxPages; i++)
     {
       MemPage->ptr   = (char *)MemPage->ptr - Size;
       MemPage->size += Size;
       PageList[i] = MemPage->ptr;
       PageList[i]->base = PageList[i]->ptr = NULL;
     }
   }
   memptr = PageList[NumPages++];
   if (!memptr->base) 
   {
     memptr->base = my_malloc (MEMPAGE_PAGESIZE);
   }
   memptr->ptr = (char *)memptr->base + MEMPAGE_PAGESIZE;
 }

 return memptr;
}


/***************************************************************************
 *  like malloc, stops the program on fail
 ***************************************************************************/
void *my_malloc (size_t size)
{
void * ptr;

 ptr =  malloc (size);
 if (ptr == NULL)
 {
#if COUNT_ALLOC_MEMORY
    fprintf (stderr, ALLOC_MEMORY_FAIL, AllocMemoryCount, AllocMemoryBlocks);
#endif
    fputs (ALLOCFAIL,stderr);
    exit (EXIT_FAILURE);
 }
#if COUNT_ALLOC_MEMORY
 AllocMemoryCount += size;
 AllocMemoryBlocks++;
#endif
 return ptr;
}


/***************************************************************************
 *  like realloc, stops the program on fail
 ***************************************************************************/
void *my_realloc (void *oldptr, size_t size)
{

 oldptr =  realloc (oldptr, size);
 if (oldptr == NULL)
 {
#if COUNT_ALLOC_MEMORY
    fprintf (stderr, ALLOC_MEMORY_FAIL, AllocMemoryCount, AllocMemoryBlocks);
#endif
    fputs (ALLOCFAIL,stderr);
    exit (EXIT_FAILURE);
 }
 return oldptr;
}


/***************************************************************************
 * like calloc, stops the program on fail
 ***************************************************************************/
void *my_calloc (size_t nmemb, size_t size)
{
void * ptr;

 ptr =  calloc (nmemb, size);
 if (ptr == NULL)
 {
#if COUNT_ALLOC_MEMORY
    fprintf (stderr, ALLOC_MEMORY_FAIL, AllocMemoryCount, AllocMemoryBlocks);
#endif
    fputs (ALLOCFAIL,stderr);
    exit (EXIT_FAILURE);
 }
#if COUNT_ALLOC_MEMORY
 AllocMemoryCount += nmemb * size;
 AllocMemoryBlocks++;
#endif
 return ptr;
}


/***************************************************************************
 * like "recalloc", stops the program on fail, 
 *                  the added memory is initialized with zero's.
 ***************************************************************************/
void *my_recalloc (void *oldptr, size_t old_size, size_t add_size)
{
void * ptr;

 if (!old_size) oldptr = NULL; /* just for security */

 if (oldptr)
   ptr =  realloc (oldptr, old_size + add_size);
 else
   ptr =  malloc  (old_size + add_size);

 if (ptr == NULL)
 {
#if COUNT_ALLOC_MEMORY
    fprintf (stderr, ALLOC_MEMORY_FAIL, AllocMemoryCount, AllocMemoryBlocks);
#endif
    fputs (ALLOCFAIL,stderr);
    exit (EXIT_FAILURE);
 }
 memset (&((char *)ptr)[old_size], 0, add_size);
#if COUNT_ALLOC_MEMORY
 AllocMemoryCount += add_size;
 AllocMemoryBlocks++;
#endif
 return ptr;
}


/**************************************************************************
 * type ** Alloc...List sizes or resizes an array of pointer.
 * all (new) pointers are initialized with the null pointer.
 * num_elem is set to the new value.
 * a pointer to the list is returned.
 *************************************************************************/

/**************************************************************************/
char ** AllocCharList (void * oldptr, size_t *num_elem, size_t add_elem)
{
char **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
AliasListType ** AllocAliasList (void * oldptr, size_t *num_elem, size_t add_elem)
{
AliasListType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
ConfigListType ** AllocConfigList (void * oldptr, size_t *num_elem, size_t add_elem)
{
ConfigListType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
SectionDescriptType ** AllocSectionDescriptList 
                       (void * oldptr, size_t *num_elem, size_t add_elem)
{
SectionDescriptType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
TableDescriptType ** AllocTableDescriptList 
                       (void * oldptr, size_t *num_elem, size_t add_elem)
{
TableDescriptType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
OutputFileDescriptType ** AllocOutputFileDescriptList 
                       (void * oldptr, size_t *num_elem, size_t add_elem)
{
OutputFileDescriptType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
SpecialListType ** AllocSpecialList (void * oldptr, size_t *num_elem, size_t add_elem)
{
SpecialListType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}

/**************************************************************************/
StringCountType ** AllocStringCountList (void * oldptr, size_t *num_elem, size_t add_elem)
{
StringCountType **ptr;

 if (add_elem)
 { 
   ptr = my_recalloc (oldptr, *num_elem * sizeof *ptr, add_elem * sizeof *ptr);
   if (ptr[*num_elem]) /* only if NULL - pointers aren't binary zero's */
   {
     int i;
     for (i = *num_elem; i < (*num_elem +  add_elem); i++) ptr[i] = NULL;
   }
 }
 else
 {
   ptr = oldptr;
 }
 *num_elem += add_elem;
 return ptr;
}


/**************************************************************************
 * determines the number of days since 1 january 1970
 *************************************************************************/

int DaysSince1970 ( int Day, int Month, int Year)
{
int isLeapYear = FALSE;
int nDays;
int i;
static  int DaysOfMonth[]={31,28,31,30,31,30,31,31,30,31,30,31};

 Month--;  /* january = 0 */

/* #define LeapYear(Y) (! (Y % 4) && (Y % 100) || !(Y % 400) ) */

 if (!(Year % 4) && !(!(Year % 100) && (Year % 400)) ) isLeapYear = TRUE;

 if (Year < 1970 || Month < 0 || Month > 11 || 
     Day < 1    || Day > DaysOfMonth[Month])
 {
  if (!isLeapYear || Month != 2 || Day > 29) return 0; /* invalid date */
 }

 Year = (Year - 1970) ; /* Years since 1970 */
 nDays =  Year * 365 +
         (Year + 1) / 4   -
         (Year + 69) / 100 +
         (Year + 369) / 400 ;

 for (i=0;i<Month;i++)
 {
   nDays += DaysOfMonth[i];
 }
 nDays += Day - 1;
 if (isLeapYear && (Month > 1)) nDays++;
 return nDays;
}



/**************************************************************************
 * for use as compare function in qsort
 *************************************************************************/
int StringListCountComp (const void *e1, const void *e2)
{
  return ( (*(StringCountType **) e2)->Count -
           (*(StringCountType **) e1)->Count );
}


/**************************************************************************
 * for use as compare function in qsort
 *************************************************************************/
int ItemListStrComp (const void *e1, const void *e2)
{
  return (strcmp ( (*(StringCountType **) e1)->pszStr ,
                   (*(StringCountType **) e2)->pszStr));
}



/**************************************************************************
 * for use as compare function in qsort
 *************************************************************************/
int DateListComp (const void *e1, const void *e2)
{
  return ( ( *(time_t *)e1 < *(time_t *)e2 ) ?  -1 : 
             *(time_t *)e1 - *(time_t *)e2 );
}


/**************************************************************************
 * for use as compare function in bsearch on lists of type StringCountType
 *************************************************************************/
int StringListBsearchComp (const void *ckey, const void *celem)
{
  return (strcmp ( (char *) ckey ,
                   (*(StringCountType **) celem)->pszStr));
}


/**************************************************************************
 * CheckQualifier checks pszArg for a valid qualifier.
 * If pszArg is no qualifier, PARAMETER is returned.
 * On fail a NO_QUALIFIER is returned, else the found QUALIFIER.
 * If a option subqualifier is found, a pointer to it's position in pszArg
 * is returned as pszOption.
 *************************************************************************/
enum QUALIFIER CheckQualifier(char * pszArg, char **  pszOption)
{
enum QUALIFIER TestArg;
 /* CONFIG,EXTRAKT,OUTPUT,INCLUDE,EXCLUDE,SINCE,BEFORE,SHOW,HELP */
char * Qualifier[]={QUALIFIER_LIST,NULL};
char * pszQualifier;
char * pszPosition=NULL;
int iAmbiguous=0;
int i;


 if (pszArg[0] != '-' 
#ifndef UNIX
     && pszArg[0] != '/'
#endif
    ) return PARAMETER; 
 pszQualifier = pszArg; pszQualifier++;
 if ((pszPosition = strpbrk (pszQualifier,"=:")) != NULL )
 {                                               /* separate option part */
   *pszPosition = '\0';
    pszPosition++;
    if (strlen(pszPosition) < 1) pszPosition = NULL;
 }
 StringToLower (pszQualifier);
 if (!strcmp(pszQualifier,"?")) pszQualifier = "help";
 for (i=0,TestArg=CONFIG;Qualifier[i] != NULL;i++)
 {
   if (strncmp(Qualifier[i],pszQualifier,strlen(pszQualifier)))
   {
      if (!iAmbiguous) TestArg++;       /* not equal */
   }
   else
   {
      iAmbiguous++;
      if (iAmbiguous > 1)
      {
         fprintf (stderr,AMBIGUOUS_WARNING,pszArg);
         TestArg = NO_QUALIFIER;
         break;
      }
   }
 }
 if (!iAmbiguous)  /* not found */
 {
    TestArg = NO_QUALIFIER;
    fprintf (stderr,UNKNOWN_WARNING,pszArg);
 }
 pszOption[0] = pszPosition;
 return  TestArg;
}

/**************************************************************************
 * StringToLower changes strings to lower case
 *************************************************************************/
char * StringToLower( char * pszString)
{
char * pszTemp;

 /* convert to lower */
 if (pszString)
 {
    pszTemp = pszString;
    while (*pszTemp)
    {
       *pszTemp = tolower ( *pszTemp);
        pszTemp++;
    }
 }
 return pszString;
}


/**************************************************************************
 * Truncate_String removes leading or trailing whitespace characters
 * from a string.
 *************************************************************************/
char * Truncate_String (char * pszString)
{
int StringLenght, i;

 if (pszString)
 {
   if ((StringLenght = strlen (pszString)))
   {
     for (i = StringLenght-1; i && isspace(pszString[i]); i--) 
     {
       pszString[i] = '\0';
     }
   }
   if ((StringLenght = strlen (pszString)))
   {
     for (i = 0; (i < StringLenght) && isspace(pszString[i]); i++); 
     pszString = &pszString[i];
   }
 }
 return pszString;
}


/***************************************************************************
 * ParseConfigLine parses a line with the syntax "Option = something".
 * There should be no whitespace characters at the beginning of Option!
 * If the syntax is ok the line will be separated in Option and something 
 * part and a pointer to "something" is returned. If the syntax check 
 * failed, a NULL pointer is returned.
 ***************************************************************************/
char * ParseConfigLine (char * pszLine)
{
char * pszSomething = NULL;
char * pszRight;

/* if (pszLine && (pszLine[0] != '#') && (!isspace(pszLine[0]))) */
 if (pszLine && (pszLine[0] != '#'))
 {                                                /* exists and no comment */
   pszRight = strchr (pszLine, '='); /* look for the separator */
   if (pszRight)    /* separator found */
   {
     *pszRight = '\0'; pszRight++;
     pszRight = Truncate_String(pszRight);
     pszLine  = Truncate_String(pszLine);
     if (strlen(pszRight) && strlen(pszLine)) /* now seem's all ok */
     {
       pszSomething = pszRight;
     }
   }
 }
 return pszSomething;
}


/***************************************************************************
 * SelectStringList returns a pointer to the stringlist of type Key out of
 * the standard lists. If Key is NO_KEY, pszSpecialList is matched against
 * the special lists.
 ***************************************************************************/
StringListType * SelectStringList (StandardListType * pSList, KeyType Key, 
                                   char * pszSpecialList)
{
StringListType * StringList = NULL;
int i;

 if ((Key == NO_KEY) && pszSpecialList)
 {
   Key = TranslateStandardListKey (pszSpecialList);
 }
 switch (Key)
 {
   case SITE:
      StringList = pSList->SiteList;
      break;
   case METHOD:
      StringList = pSList->MethodList;
      break;
   case ITEM:
      StringList = pSList->ItemList;
      break;
   case PROTOCOL:
      StringList = pSList->ProtocolList;
      break;
   case STATUS:
      StringList = pSList->StatusList;
      break;
   case AGENT:
      StringList = pSList->AgentList;
      break;
   case NO_KEY:
      if (pszSpecialList)
      {
        if (!strcmp(pSList->WeekDayList->Name,pszSpecialList))
        {
            StringList = pSList->WeekDayList;
        }
        for (i = 0; (i < pSList->NumSpecialLists) && !StringList; i++)
        {
          if (!strcmp(pSList->SpecialLists[i]->Name,pszSpecialList))
          {
            StringList = pSList->SpecialLists[i];
          }
        }
      }
      break;
   default:
      break;
 }
 return StringList;
}



/**************************************************************************
 * InsertThousandsSeparator converts two unsigned long integer to a 
 * string including a grouping separator for non-monetary values.
 * return value:
 * a pointer to the new string
 *************************************************************************/

char * InsertThousandsSeparator (unsigned long ulLow, unsigned long ulHigh)
{
static char szValue[32];
size_t NumDigits, NumSeparator;
int i,j;

 if (ulHigh)
   NumDigits = 9 + nDigits (ulHigh);
 else
   NumDigits = nDigits (ulLow); 
 NumSeparator = (NumDigits - 1) / 3;

 if ((NumDigits + NumSeparator + 1) > 32)
 {      /* that's bad! */
    strcpy (szValue, "**error**");
    return szValue;
 }

 szValue[NumDigits + NumSeparator] = '\0';
 if (ulHigh) 
 {
   sprintf (szValue,"%lu%9.9lu",ulHigh,ulLow);
 }
 else
 {
   sprintf (szValue,"%lu",ulLow);
 }

 for (i = 0; i < NumSeparator; i++)
 {
   for (j = 0; j < 3; j++)
   {
     szValue[NumDigits - (i * 3) - j - 1 + NumSeparator - i] =
                                  szValue[NumDigits - (i * 3) - j - 1];
   }
   szValue[NumDigits - (i * 3) - j - 1 + NumSeparator - i] = THOUSANDS_SEP;
 }

 return szValue;
}


/**************************************************************************
 * nDigits evaluates the number of digits of ulValue
 *************************************************************************/
unsigned int nDigits (unsigned long ulValue)
{
unsigned int Digits = 1;

 while (ulValue /= 10) Digits++;
 return Digits;
}


/**************************************************************************
 * DateFromString_t converts a date string into the time_t format (seconds
 * since 1-Jan-1970).
 * supported input formats:
 *   dd.mm.jjjj, dd-mm-jjjj, dd-mmm-jjjj, dd/mmm/jjjj, mm/dd/jjjj
 *   today, tomorrow, yesterday, NULL == today
 *   nn  (number of days before today)
 *************************************************************************/


time_t DateFromString_t (const char * pszString)
{
char * pszMonth = "JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC";
char szMonth[4];
char * pszPos1, * pszPos2;
int iDay, iMonth, iYear;
int iNumParameter, i;
time_t t_Date = 0;

 if (!pszString || !strcmp(pszString,"today"))
 {
   t_Date = ( time(NULL) / SECONDS_OF_A_DAY ) * SECONDS_OF_A_DAY;
 }
 else if (!strcmp(pszString,"yesterday"))
 {
   t_Date = (( time(NULL) / SECONDS_OF_A_DAY ) - 1) * SECONDS_OF_A_DAY;
 }
 else if (!strcmp(pszString,"tomorrow"))
 {
   t_Date = (( time(NULL) / SECONDS_OF_A_DAY ) + 1) * SECONDS_OF_A_DAY;
 }
 else if (strlen(pszString) < 6) /* possibly a delta time */
 {
   iDay = atoi (pszString);
   i    = time(NULL) / SECONDS_OF_A_DAY ;   /* Days since 1-Jan-1970 */
   iDay = (iDay < 0) ? 0 : (iDay > i) ? i : iDay;
   t_Date = (i - iDay) * SECONDS_OF_A_DAY;
 }
 else
 {
   pszPos1 = pszPos2 = strpbrk (pszString,".-/");
   if (!pszPos1 || strlen(pszString) < 6 ||  strlen(pszString) > 11)
   {
     t_Date = 0;
   }
   else
   {
     pszPos2++;
     pszPos2 = strpbrk (pszPos2,".-/");
     if ((pszPos2 - pszPos1) == 4)
     {
       iNumParameter = sscanf (pszString,"%d%*c%3s%*c%d",&iDay,szMonth,&iYear);
       szMonth[3] = '\0';
       for (i=0,pszPos1=szMonth;i<3;i++,pszPos1++)
       {
          *pszPos1 = toupper ( *pszPos1);
       }
       pszPos1 = strstr(pszMonth,szMonth);
       if (pszPos1 != NULL)
       {
          iMonth = (pszPos1 - pszMonth)/3 + 1;
       }
     }
     else
     {
       if (*pszPos1 == '/')     /* mm/dd/jjjj */
       {
         iNumParameter = sscanf (pszString,"%d%*c%d%*c%d",&iMonth,&iDay,&iYear);
       }
       else     /* dd.mm.jjjj */
       {
         iNumParameter = sscanf (pszString,"%d%*c%d%*c%d",&iDay,&iMonth,&iYear);
       }
     }
     if (iNumParameter == 3)
     {
       if (iYear < 100) iYear += 1900;
       t_Date = DaysSince1970 ( iDay,  iMonth, iYear) * SECONDS_OF_A_DAY;
     }
   }
 }
 return t_Date;
}


/**************************************************************************
 * allows wildcards in s2
 * valid combinations:
 *
 * case 1: s2 = sub
 * case 2: s2 = *subn
 * case 3: s2 = sub0*
 * case 4: s2 = sub0*sub1*...*subn
 *
 * return values:
 * -1   lower or NULL Pointer
 *  0   equal 
 *  1   graeter
 *************************************************************************/
int wild_strcmp (const char *s1, const char *s2)
{
size_t s1_len, s2_len, Ind_s1;
size_t token_len;
int Index, MaxIndex;
char * pszSub1, * pszSub2, * pszSubn;
int iStatus;


 if (!s1 || !s2)   /* check for NULL Pointers */
 {
   iStatus = -1;
 }
 else if (!(pszSub1 = strchr (s2, '*'))) /* no wildcards found */
 {
   iStatus = strcmp (s1, s2);
 }
 else
 {
   s1_len = strlen (s1);
   s2_len = strlen (s2);
   iStatus = 0;
   pszSubn = strrchr (s2, '*');
   Ind_s1 = 0;

   if (s2[0] != '*')  /* sub*... */
   {
     Ind_s1 = token_len = (size_t) (pszSub1 - s2);
     iStatus = strncmp (s1, s2, token_len);
   }

   if (!iStatus && s2[s2_len-1] != '*') /* ...*sub */
   {
     token_len = s2_len - (size_t) (pszSubn - s2) - 1;
     if (token_len > s1_len)
     {
       iStatus = 1;
     }
     else
     {
       iStatus = strncmp (&s1[s1_len-token_len],&s2[s2_len-token_len], token_len);
       s1_len -= token_len ;
     }
   }

   while (!iStatus && (pszSubn - pszSub1)) /* ...*sub*... */
   {
     pszSub2 = strchr (++pszSub1, '*');
     token_len = (size_t) (pszSub2 - pszSub1);
     if (token_len > (s1_len - Ind_s1))
     {
       iStatus = 1;
     }
     else
     {
       for (Index = Ind_s1, MaxIndex = (s1_len - token_len - 1);
            (iStatus = strncmp (&s1[Index], pszSub1, token_len)) &&
            (Index <= MaxIndex);
            Index++);
       Ind_s1 = Index + token_len;
       pszSub1 = pszSub2;
     }
   }
 }

 if (iStatus < 0) iStatus = -1;
 else if (iStatus > 0) iStatus = 1; 

 return iStatus;
}


/***************************************************************************
 * MakeDateString creates a 6 byte string from a given time. The format 
 * is yymmdd for daily, yyyyww for weekly and yyyymm for monthly periods.
 ***************************************************************************/
char * MakeDateString (time_t Date, enum DATE_MODE DateMode, int WeekOffset)
{
static char DateString[7];

 if (Date)
 {
   char * pszFormat;
   int iDay, iMonthOrWeek, iYear;
   struct tm *p_tm;

   p_tm = localtime (&Date);
   iDay         = p_tm->tm_mday;
   iMonthOrWeek = p_tm->tm_mon + 1;
   iYear        = p_tm->tm_year + 1900;
   switch (DateMode)
   {
     default:
     case PER_DAY:
       pszFormat = "%2.2d%2.2d%2.2d";
       iYear = iYear % 100;
       break;
     case PER_WEEK:
       pszFormat = "%4.4d%2.2d";
       iMonthOrWeek = (p_tm->tm_yday + WeekOffset) / 7;
#if 0
       iYear += iMonthOrWeek / 52;
       iMonthOrWeek = iMonthOrWeek % 52 + 1;
#else
       iMonthOrWeek++;
#endif
       break;
     case PER_MONTH:
       pszFormat = "%4.4d%2.2d";
       break;
   }
   sprintf (DateString, pszFormat, iYear, iMonthOrWeek, iDay);
   if (strchr(DateString, ' ')) /* should not occur with ansi c */
   {
     int i;
     for (i = 0; i < 6; i++)
     {
       if (DateString[i] == ' ') DateString[i] = '0';
     }
   }
 }
 else
 {
   strcpy (DateString, PERIOD_REPLACEMENT);
 }

 return DateString;
}


/***************************************************************************
 * PrintDate outputs a date 
 * legal part of the format string are
 * dd      - %d - day as two digits
 * ddd     - %a - abbreviated weekday name 
 * dddd    - %A - full weekday name 
 * mm      - %m - month as two digits
 * mmm     - %b - abbreviated month name
 * mmmm    - %B - full month name
 * ww      -    - week number as two digits
 * yy      - %y - year without century
 * yyyy    - %Y - year as four digits
 ***************************************************************************/
int PrintDate (char * Buffer, int BufLen, time_t Date, char * pszFormat, int WeekOffset)
{
char * pszToken;
int iLenght;

 if (Buffer) *Buffer = '\0';
 if (pszFormat)
 {
   while ((pszToken = strstr(pszFormat, "dddd")))
   {
     *pszToken++ = '%'; *pszToken++ = 'A';
     memmove (pszToken, pszToken + 2, strlen(pszToken) - 1); 
   }
   while ((pszToken = strstr(pszFormat, "mmmm")))
   {
     *pszToken++ = '%'; *pszToken++ = 'B';
     memmove (pszToken, pszToken + 2, strlen(pszToken) - 1); 
   }
   while ((pszToken = strstr(pszFormat, "yyyy")))
   {
     *pszToken++ = '%'; *pszToken++ = 'Y';
     memmove (pszToken, pszToken + 2, strlen(pszToken) - 1); 
   }
   while ((pszToken = strstr(pszFormat, "ddd")))
   {
     *pszToken++ = '%'; *pszToken++ = 'a';
     memmove (pszToken, pszToken + 1, strlen(pszToken)); 
   }
   while ((pszToken = strstr(pszFormat, "mmm")))
   {
     *pszToken++ = '%'; *pszToken++ = 'b';
     memmove (pszToken, pszToken + 1, strlen(pszToken)); 
   }
   while ((pszToken = strstr(pszFormat, "dd")))
   {
     *pszToken++ = '%'; *pszToken = 'd';
   }
   while ((pszToken = strstr(pszFormat, "mm")))
   {
     *pszToken++ = '%'; *pszToken = 'm';
   }
   while ((pszToken = strstr(pszFormat, "ww")))
   {
     char * pszDate = MakeDateString (Date, PER_WEEK, WeekOffset);
     *pszToken++ = pszDate[4]; *pszToken = pszDate[5]; 
   }
   while ((pszToken = strstr(pszFormat, "yy")))
   {
     *pszToken++ = '%'; *pszToken = 'y';
   }
 }
 else
 {
   pszFormat = "%d-%h-%Y";
 }
 iLenght = strftime (Buffer, BufLen - 1, pszFormat, localtime(&Date));
 if (iLenght) Buffer[iLenght+1] = '\0';
 return iLenght;
}


/***************************************************************************
 * eval_cpu_time returns the estimated cpu time in clock_t units. It had to
 * be initialized at program start with a NULL pointer argument. If called
 * with a string argument, the cpu time since last call is printed to
 * stdout.
 ***************************************************************************/
clock_t eval_cpu_time (char * OutString)
{
static clock_t LastTime = 0;
static clock_t StartTime = 0;
clock_t CurrentTime = clock();

 if (!OutString)
   LastTime = StartTime = clock();
 else if (strlen(OutString))
 {
   fprintf (stderr,"%s: %.1f seconds\n", OutString,
           ((double)(CurrentTime - LastTime)) / CLOCKS_PER_SEC);
   LastTime = CurrentTime;
 }

 return (CurrentTime - StartTime);
}


/***************************************************************************
 * MinDate returns the oldest date found in DateList
 ***************************************************************************/
time_t MinDate (DateListType * DateList)
{
time_t FirstTime = 0;
int i;

 if (DateList->NumMembers)
 {
   FirstTime = DateList->Date[0];
   for (i = 0; i < DateList->NumMembers; i++)
   {
     if (DateList->Date[i] < FirstTime) FirstTime = DateList->Date[i];
   }
 }
 return FirstTime;
}

/***************************************************************************
 * MaxDate returns the newest date found in DateList
 ***************************************************************************/
time_t MaxDate (DateListType * DateList)
{
time_t LastTime = 0;
int i;

 if (DateList->NumMembers)

 {
   LastTime = DateList->Date[0];
   for (i = 0; i < DateList->NumMembers; i++)
   {
     if (DateList->Date[i] > LastTime)  LastTime  = DateList->Date[i];
   }
 }
 return LastTime;
}



/***************************************************************************
 * ByteCountAdd adds to members of the struct ByteCountType.
 * Full billions are counted in NumBytes->high.
 ***************************************************************************/
void  ByteCountAdd (ByteCountType * NumBytes, uLong low, uLong high)
{
 NumBytes->low += low;
 NumBytes->high += high;
 while (NumBytes->low >= 1000000000)
 {
   NumBytes->high++;
   NumBytes->low -= 1000000000;
 }
 return;
}

/***************************************************************************
 * ByteCountMax returns in MaxVal the higher of the two structs
 ***************************************************************************/
void  ByteCountMax (ByteCountType * MaxVal, ByteCountType * Val)
{

 if ( (MaxVal->high  < Val->high)  || 
    ( (MaxVal->high == Val->high)  && (MaxVal->low < Val->low)))
 {
   MaxVal->low  = Val->low;
   MaxVal->high = Val->high;
 }
 return;
}

/***************************************************************************
 ***************************************************************************/
double ByteCountProz (ByteCountType * Val, ByteCountType * All)
{
double dWert, dTotal, dProz = 0.0;

 dWert  = Val->low + 1000000000.0 * Val->high;
 dTotal = All->low + 1000000000.0 * All->high;
 if (dTotal) dProz  = (100.0 * (dWert / dTotal));
 return dProz;
}


/***************************************************************************
 * StringListMaxValue determines the highest value in the current list
 ***************************************************************************/
uLong StringListMaxValue (StringListType *pSL)
{
int i;
uLong ulMaxValue = 0;


  for (i = 0; i < pSL->NumMembers; i++)
  {
    if (pSL->List[i]->Count > ulMaxValue) 
    {
      ulMaxValue = pSL->List[i]->Count;
    }
  }
  return ulMaxValue;
}

/***************************************************************************
 * StringListTotalCount evaluates the total count for StringList pSL.
 * The stringlist must be initialized for the given period.
 ***************************************************************************/
uLong StringListTotalCount (StringListType *pSL)
{
uLong ulTotalCount = 0;
int i;

 for (i = 0; i < pSL->NumMembers; i++) 
 {
   ulTotalCount += pSL->List[i]->Count;
 }
 return ulTotalCount;
} 


/***************************************************************************
 * StringListAdd adds the counters for a specified period into Count
 ***************************************************************************/
void StringListAdd (StringListType *pSL, int * Indices, int NumIndices,
                                                       uInt MaxIndices)
{
int i, j;

 if (pSL)
 {
  if (NumIndices < 0)  /* all Members */
  {
    for (i = 0; i < pSL->NumMembers; i++)
    {
      for (j = 0, pSL->List[i]->Count = 0; j < MaxIndices; j++) 
      {
        pSL->List[i]->Count += pSL->List[i]->uiCountAr[j];
      }
    }
  }
  else                 /* only specified period */
  {
    for (i = 0; i < pSL->NumMembers; i++)
    {
      for (j = 0, pSL->List[i]->Count = 0; j < NumIndices; j++) 
      {
        pSL->List[i]->Count += pSL->List[i]->uiCountAr[Indices[j]];
      }
    }
  }
 }
 return;
}


/***************************************************************************
 * CountStringListMembers counts the hits to the in pszMask specified 
 * members. pszMask is a comma separated list which can contain wildcard
 * characters.
 ***************************************************************************/
uLong CountStringListMembers (StringListType * StringList, char * pszMask)
{
uLong Count = 0;
char * pszWild;
int i;

 if (StringList && pszMask)
 {
   for (pszWild = strtok(pszMask,","); pszWild; 
        pszWild = strtok(NULL,","))
   {
     pszWild = Truncate_String (pszWild);
     if (*pszWild)
     {
       for (i = 0; i < StringList->NumMembers; i++)
       {
         if (!wild_strcmp(StringList->List[i]->pszStr,pszWild))
         {
           Count += StringList->List[i]->Count;
         }
       }
     }
   }
 }
 return Count;
}

/***************************************************************************
 * EvalDateListPeriod evaluates the total bytes and counts from DateList
 * within a given Period.
 ***************************************************************************/
void EvalDateListPeriod (int * Indices, uInt NumIndices, DateListType * DateList,
                         unsigned int * TotalCount, ByteCountType * pTotalBytes)
                         
{
int i;

 *TotalCount = pTotalBytes->low = pTotalBytes->high = 0;
 if (Indices && DateList)
 {
   for (i = 0; i < NumIndices; i++)
   {
     *TotalCount  += DateList->uiCountAr[Indices[i]];
     ByteCountAdd (pTotalBytes, DateList->uiByteAr[Indices[i]], 0);
   }
 }
 return;
}

