/*
   (C) Copyright 1996  Michael Stenns
   
   Permission to use, copy, modify, and distribute this program without fee 
   is hereby granted, provided that this copyright and permission notice 
   appear on all copies and supporting documentation.

   All sourcefiles starting with "gd" are part or Thomas Boutell's graphic
   library gd1.2. gd 1.2 is copyright 1994, 1995, Quest Protein Database 
   Center, Cold Spring Harbor Labs.

   If you want use the gd1.2 library, remark that the files provided with
   this package are slightly modified. You should get the original set from
   http://www.boutell.com/gd/ .

*/
/***************************************************************************/
#define OUTPUT_C

#include "accstat.h"

/* string constants */


/* declarations */


/* prototypes */


/**************************************************************************
 * SummeryReport shows a short summery statistic of the read logfiles
 *************************************************************************/
void SummeryReport (StandardListType * pStandardList, 
                     CommandLineType * pCmdLinOptions)
{
KeyType Key;
int i;

 i = -1; StandardListsSetPeriod (pStandardList, &i);
 for (i=0,Key=SITE; i < (KEYS_NUM_MEMBERS - 1); i++,Key++)
 {
   StringListType * pStringList = SelectStringList (pStandardList, Key, "");

   pStringList->Sort = pCmdLinOptions->Sort[Key];
   pStringList->ShowMembers = pCmdLinOptions->ShowMembers[Key];
   StringListPrint (stdout, pStringList, 76, pCmdLinOptions->HideList);
   if (pStringList->ShowMembers && pStringList->NumMembers) puts ("\n");
 }

 for (i = 0; i < pStandardList->NumSpecialLists; i++)
 {
   pStandardList->SpecialLists[i]->Sort = 
                            pCmdLinOptions->Sort[KEYS_NUM_MEMBERS];
   pStandardList->SpecialLists[i]->ShowMembers = 
                            pCmdLinOptions->ShowMembers[KEYS_NUM_MEMBERS];
   StringListPrint (stdout, pStandardList->SpecialLists[i], 
                    76, pCmdLinOptions->HideList);
   if (pStandardList->SpecialLists[i]->ShowMembers) puts ("\n");
 }

DateListPrint   (stdout, pStandardList->DateList, 0, 76, pCmdLinOptions->DateMode);
#if COUNT_ALLOC_MEMORY >= 2
 eval_cpu_time ("elapsed time in producing output");
#endif
#if COUNT_ALLOC_MEMORY >= 1
 fprintf (stderr,COUNT_ALLOC_MEMORY_FORMAT,AllocMemoryCount,AllocMemoryBlocks);
#endif
 SummeryReportPrint (stdout, pStandardList, pCmdLinOptions, 
        MinDate (pStandardList->DateList), MaxDate (pStandardList->DateList));

 return;
}

/**************************************************************************
 * SummeryReportPrint shows a short summery statistic of the read logfiles
 *************************************************************************/
void SummeryReportPrint (FILE * outfile, StandardListType * pStandardList, 
                         CommandLineType * pCmdLinOptions, 
                         time_t StartTime, time_t EndTime)
{
size_t NumClientError = 0;      /* status codes 400 - 500 */
size_t NumServerError = 0;      /* status codes > 500  */
size_t NumSuccessTransfers = 0; /* rest */
size_t NumRecordMatched = 0;
ByteCountType NumBytes = {0,0};
uInt NumSites = 0, NumItems = 0, NumMethods = 0, NumProtocols = 0;
uInt NumStatus = 0, NumDays = 0;
uInt uiStatusCode;

char * pszBytesTransferred;
char sStartTime[16], sEndTime[16];
int i;


 for (i = 0; i < pStandardList->DateList->NumMembers; i++)
 {

   if ((pStandardList->DateList->Date[i] >= StartTime) &&
       (pStandardList->DateList->Date[i] <= EndTime))
   {
     NumDays++;
     NumRecordMatched += pStandardList->DateList->uiCountAr[i];
     ByteCountAdd (&NumBytes,pStandardList->DateList->uiByteAr[i],0);
   }
 }
 pszBytesTransferred = InsertThousandsSeparator 
                       (NumBytes.low, NumBytes.high);

 for (i = 0; i < pStandardList->StatusList->NumMembers; i++)
 {
  if (pStandardList->StatusList->List[i]->Count)
  {
    NumStatus++;
    uiStatusCode = atoi (pStandardList->StatusList->List[i]->pszStr);
    if (uiStatusCode >= 400)
    {
      NumClientError += pStandardList->StatusList->List[i]->Count;
    }
    if (uiStatusCode >= 500)
    {
      NumServerError += pStandardList->StatusList->List[i]->Count;
    }
  }
 }
 NumClientError -= NumServerError;
 NumSuccessTransfers = NumRecordMatched 
                       - NumClientError - NumServerError;

 for (i = 0; i < pStandardList->SiteList->NumMembers; i++)
 {
   if (pStandardList->SiteList->List[i]->Count) NumSites++;
 }
 for (i = 0; i < pStandardList->ItemList->NumMembers; i++)
 {
   if (pStandardList->ItemList->List[i]->Count) NumItems++;
 }
 for (i = 0; i < pStandardList->MethodList->NumMembers; i++)
 {
   if (pStandardList->MethodList->List[i]->Count) NumMethods++;
 }
 for (i = 0; i < pStandardList->ProtocolList->NumMembers; i++)
 {
   if (pStandardList->ProtocolList->List[i]->Count) NumProtocols++;
 }
   

 if (!StartTime) StartTime = time(&StartTime) - (StartTime % SECONDS_OF_A_DAY);
 strftime(sStartTime, 16, SUMMARY_TIME_FORMAT, localtime(&StartTime));
 strftime(sEndTime, 16, SUMMARY_TIME_FORMAT, localtime(&EndTime));


#ifndef SUMMARY_REPORT_FORMAT
#define SUMMARY_REPORT_FORMAT " \
============================================================\n\
 %u record%s out of %u record%s matched, %u site%s\n\
 fetched %u different document%s using %u method%s and\n\
 %u protocol%s on %u day%s transfering %s byte%s\n\
 of data; %u status code%s indicates %u successful\n\
 transfer%s, %u client and %u server errors.\n"
#endif

 if (NumDays > 1)
 {
   fprintf (outfile, SUMMARY_REPORT_TITLE_FORMAT, sStartTime, sEndTime);
 }
 else
 {
   fprintf (outfile, SUMMARY_REPORT_TITLE_1DAY_FORMAT, sStartTime);
 }

 fprintf (outfile, SUMMARY_REPORT_FORMAT,
          (uInt) NumRecordMatched, 
          (NumRecordMatched != 1) ? PLURAL_SIGN : "", 
          pStandardList->NumRecord,  
          (pStandardList->NumRecord != 1) ? PLURAL_SIGN : "", 
          NumSites, (NumSites != 1) ? PLURAL_SIGN : "",
          NumItems, (NumItems != 1) ? PLURAL_SIGN : "",
          NumMethods, (NumMethods != 1) ? PLURAL_SIGN : "",
          NumProtocols, (NumProtocols != 1) ? PLURAL_SIGN : "",
          NumDays, (NumDays != 1) ? PLURAL_SIGN : "",
          pszBytesTransferred,
          (NumBytes.low != 1) ? PLURAL_SIGN : "",
          NumStatus, (NumStatus != 1) ? PLURAL_SIGN : "",
          (uInt) NumSuccessTransfers,
          (NumSuccessTransfers != 1) ? PLURAL_SIGN : "",
          (uInt) NumClientError,
          (uInt) NumServerError);

 return;
} 



/***************************************************************************
 * DateListPrint displays the datelist in the appropriate format 
 ***************************************************************************/
void DateListPrint (FILE * outfile, DateListType *pDateList,
                    uInt MaxRows, uInt MaxWidth, enum DATE_MODE DateMode)
{
time_t CurrentPeriod;
size_t NumRows;
char szFormat[32], szDataBuf[32], sRow[80];
char * pszBytesTransferred;
char * pszPeriod, * pszDateFormat;
unsigned int NumCountOfPeriod;
unsigned int CountMax = 0;
ByteCountType NumBytes;
size_t NumCountDigits, NumByteDigits = 0;
int i;
int Indices[32], MaxIndices;

 if (pDateList->NumMembers && DateMode != NO_DATE)
 {
   switch (DateMode)
   {
     default:
     case PER_DAY:
        MaxIndices = 2;
        pszPeriod = "Daily Report";
        pszDateFormat = SUMMARY_TIME_FORMAT;
        break;
     case PER_WEEK:
        MaxIndices = 8;
        pszPeriod = "Weekly Report";
        pszDateFormat = SUMMARY_TIME_FORMAT;
        break;
     case PER_MONTH:
        MaxIndices = 32;
        pszPeriod = "Monthly Report";
        pszDateFormat = "%B %Y";
        break;
   }
   /*
    * evaluate start values
    */
   CountMax = NumCountDigits = NumByteDigits = 0;
   NumRows = NumberOfPeriods (pDateList, DateMode);
   for (i = (MaxRows && NumRows > MaxRows) ? NumRows - MaxRows : 0 ; 
        i < NumRows; i++)
   {             /* evaluate startup values */
     GetPeriodIndices (Indices, MaxIndices, i + 1, pDateList, DateMode);
     EvalDateListPeriod (&Indices[1], Indices[0], pDateList,
                         &NumCountOfPeriod, &NumBytes);
     if (CountMax < NumCountOfPeriod)  CountMax = NumCountOfPeriod;
     pszBytesTransferred = InsertThousandsSeparator
                           (NumBytes.low, NumBytes.high);
     if (NumByteDigits < strlen (pszBytesTransferred))
     {
       NumByteDigits = strlen (pszBytesTransferred);
     }
   }
   /*
    * print head of table
    */
   NumCountDigits = nDigits (CountMax);
   if (NumCountDigits < 7) NumCountDigits = 7;
   if (NumByteDigits < 7) NumByteDigits = 7;
   sprintf (szFormat,"%%%uu | %%%us | %%.%us\n",
                     (uInt) NumCountDigits,(uInt) NumByteDigits,
                     (uInt) (MaxWidth - 6 - NumCountDigits - NumByteDigits));
   memset (sRow, ' ', 80);
   memcpy (&sRow[(NumCountDigits/2)-1],"Count",5);
   memcpy (&sRow[(NumByteDigits/2)-1+NumCountDigits+2],"Bytes",5);
   memcpy (&sRow[NumCountDigits+NumByteDigits+6],pszPeriod,strlen(pszPeriod));
   sRow[NumCountDigits+1] = '|';
   sRow[NumCountDigits+NumByteDigits+4] = '|';
   sRow[MaxWidth] = '\0';
   fprintf (outfile,"%.72s\n",sRow);
   memset (sRow, '-', 80); sRow[0] = ' ';
   sRow[NumCountDigits+1] = '|';
   sRow[NumCountDigits+NumByteDigits+4] = '|';
   sRow[MaxWidth] = '\0';
   fprintf (outfile,"%.72s\n",sRow);
   /*
    * print rows
    */
   for (i = (MaxRows && NumRows > MaxRows) ? NumRows - MaxRows : 0 ; 
        i < NumRows; i++)
   {
     CurrentPeriod = GetPeriodIndices (Indices, MaxIndices, i + 1, pDateList, DateMode);
     EvalDateListPeriod (&Indices[1], Indices[0], pDateList,
                         &NumCountOfPeriod, &NumBytes);
     pszBytesTransferred = InsertThousandsSeparator
                           (NumBytes.low, NumBytes.high);
     strftime (szDataBuf, 30, pszDateFormat, localtime(&CurrentPeriod));
     if (NumCountOfPeriod || NumBytes.low || NumBytes.high)
     {
       fprintf (outfile,szFormat, NumCountOfPeriod , pszBytesTransferred, szDataBuf);
     }
   }
 }
 return;
}



/***************************************************************************
 * StringListPrint display a StringList in table form
 ***************************************************************************/
void StringListPrint (FILE * outfile, StringListType *pSL, 
                      unsigned int MaxWidth, ConfigListType ** HideList)
{
unsigned int Num;
unsigned int NumRows;
unsigned long ulMaxValue;
size_t MaxValidValue = 0;
int i;
char szFormat[40];
char * String;

 ulMaxValue = StringListMaxValue (pSL);
 if (pSL->NumMembers && pSL->ShowMembers)
 {
  if (MaxWidth < 15) MaxWidth = 15;
  NumRows = NumValidStringListMembers (pSL, HideList, &MaxValidValue);
  if (pSL->Sort == SORT_ALPHABETICAL)
  {
    qsort (pSL->List , pSL->NumMembers, sizeof (void *), ItemListStrComp);
  }
  else  if (pSL->Sort == SORT_NUMERICAL) 
  {
    qsort (pSL->List , pSL->NumMembers, sizeof (void *), StringListCountComp);
  }
  NumRows = (pSL->ShowMembers < NumRows) ? pSL->ShowMembers : NumRows;
  if (MaxValidValue < 100000)
  {
    sprintf (szFormat,"  Count | Top %%u of %%u %%s\n%%.%us\n",MaxWidth);
    fprintf (outfile,szFormat,NumRows,(uInt) pSL->NumMembers,pSL->Name,
" -------|------------------------------------------------------------");
    sprintf (szFormat,"%%%du | %%.%us \n",7,MaxWidth-10);
  }
  else
  {
    sprintf (szFormat,"    Count  | Top %%u of %%u %%s\n%%.%us\n",MaxWidth);
    fprintf (outfile,szFormat,NumRows,(uInt) pSL->NumMembers,pSL->Name,
" ----------|------------------------------------------------------------");
    sprintf (szFormat,"%%%du | %%.%us \n",10,MaxWidth-13);
  }
  for (i = 0, Num = 0;(Num < NumRows) && (i < pSL->NumMembers); i++)
  {
   if (CheckAgainstHideList (HideList, pSL->Key, pSL->List[i]->pszStr) &&  
                                       pSL->List[i]->Count)
   {
     if (pSL->Key == STATUS)
     {
       String =  GetStatusCode (pSL->List[i]->pszStr);
     }
     else
     {
       String = pSL->List[i]->pszStr;
     }
     fprintf (outfile, szFormat, pSL->List[i]->Count,String);
     Num++;
   }
  }
 }
 return;
}



/***************************************************************************
 * CheckAgainstHideList checks the Value of keytype Key against the list
 * Hidelist.
 * It returns TRUE, if Value is not found, else FALSE.
 ***************************************************************************/
int CheckAgainstHideList (ConfigListType ** HideList, KeyType Key,
                          char * pszString)
{
int Status = TRUE;
int i = 0;

 if (!HideList) return TRUE;
 while (HideList[i])  /* list is null pointer terminated */
 {
   if (HideList[i]->Key == Key)
   {
     if (!wild_strcmp (pszString,
                       HideList[i]->Val.pszString)) Status = FALSE;
   }
   i++;
 }
 return Status;
}

/***************************************************************************
 * GetStatusCode expands the status code from the three digits
 ***************************************************************************/
char * GetStatusCode (char * pszCode)
{
static char szUnknown[30];
static char * StatusCodes[] = {
      "100  Continue",
      "101  Switching Protocols",
      "200  OK",
      "201  Created",
      "202  Accepted",
      "203  Non-Authorative Information",
      "204  No Content",
      "205  Reset Contend",
      "206  Partial Contend",
      "300  Multiple Choices",
      "301  Moved Permanently",
      "302  Moved Temporarily",
      "303  See Other",
      "304  Not Modified",
      "305  Use Proxy",
      "400  Bad Request",
      "401  Unauthorized",
      "402  Payment Required",
      "403  Forbidden",
      "404  Not Found",
      "405  Method Not Allowed",
      "406  None Acceptable",
      "407  Proxy Authentication Required",
      "408  Request Timeout",
      "409  Conflict",
      "410  Gone",
      "411  Lenght Required",
      "412  Unless True",
      "500  Internal Server Error",
      "501  Not Implemented",
      "502  Bad Gateway",
      "503  Service Unavailable",
      "504  Gateway Timeout", NULL   };
int i = 0;

 if (pszCode)
 {
   while (StatusCodes[i] && strncmp(StatusCodes[i],pszCode,3)) i++;
   if (StatusCodes[i]) 
     pszCode = StatusCodes[i];
   else
   {
     sprintf (szUnknown,"%3.3s  Unknown Status Code",pszCode);
     pszCode = szUnknown;
   }
 }
 else
 {
   pszCode = "Error";
 }
 return pszCode;
}


/***************************************************************************
 * WriteHistoryFile writes the history file to disk.
 ***************************************************************************/
void WriteHistoryFile (ConfigFileType * pConfigOpts, StandardListType * pSList)
{
FILE * HistFile = NULL;
char HistoryFileName[FILENAME_MAX + 1];
DateListType * HistoryDateList = pSList->HistoryDateList;
DateListType * DateList = pSList->DateList;
time_t StartDate, EndDate, CurrentDate, TempDate;
char szStartDate[12], szEndDate[12];
uInt uiVal1;
int i;

 if (!pConfigOpts->HistoryFile || !DateList->NumMembers) return;

 if (((strlen(pConfigOpts->OutPath) + 
     strlen(pConfigOpts->HistoryFile)) <= FILENAME_MAX))
 {
   sprintf (HistoryFileName,"%s%s",pConfigOpts->OutPath,pConfigOpts->HistoryFile);
   HistFile = fopen (HistoryFileName,"w");
   if (!HistFile)
   {
     fprintf (stderr, HISTFILE_OPENWRITE_ERROR, HistoryFileName);
     return;
   }
 }
 else
 {
   return;
 }
 StartDate = MinDate(HistoryDateList);
 EndDate   = MaxDate(HistoryDateList);
 CurrentDate  = time(NULL); 
 CurrentDate -= (CurrentDate % SECONDS_OF_A_DAY);
 /* never store the current day */
 if (EndDate >= CurrentDate) EndDate = CurrentDate - SECONDS_OF_A_DAY;

 /* adjust StartDate to the beginning of a week */
 StartDate = (7  - pSList->DateList->Offset) * SECONDS_OF_A_DAY + SECONDS_OF_A_WEEK * 
              (-1 + (StartDate + pSList->DateList->Offset  * SECONDS_OF_A_DAY)
                        / SECONDS_OF_A_WEEK);

 fputs (HISTORY_FILE_STARTLINE, HistFile);
 strftime (szStartDate, sizeof szStartDate,  HISTORY_FILE_TIMEFORMAT, 
           localtime (&StartDate));
 strftime (szEndDate, sizeof szEndDate,  HISTORY_FILE_TIMEFORMAT, localtime (&EndDate));
 fprintf (HistFile, " Time: %s - %s\n#", szStartDate, szEndDate);
 
 for (CurrentDate = StartDate; CurrentDate <= EndDate; CurrentDate += SECONDS_OF_A_WEEK)
 {
   strftime (szStartDate, sizeof szStartDate,  HISTORY_FILE_TIMEFORMAT, 
             localtime (&CurrentDate));
   fprintf (HistFile,"\n Week of: %s\n Count: ",szStartDate);
   for (TempDate  = CurrentDate; 
        TempDate < (CurrentDate + SECONDS_OF_A_WEEK); 
        TempDate += SECONDS_OF_A_DAY)
   {
     for (i = 0, uiVal1 = 0; !uiVal1 && (i < HistoryDateList->NumMembers); i++)
     {
       if (TempDate == HistoryDateList->Date[i])
       {
         uiVal1 = HistoryDateList->uiCountAr[i];
       }
     }
     if (TempDate <= EndDate) fprintf (HistFile," %u ",uiVal1);
   }
   fprintf (HistFile,"\n Bytes: ");
   for (TempDate  = CurrentDate; 
        TempDate < (CurrentDate + SECONDS_OF_A_WEEK); 
        TempDate += SECONDS_OF_A_DAY)
   {
     for (i = 0, uiVal1 = 0; !uiVal1 && (i < HistoryDateList->NumMembers); i++)
     {
       if (TempDate == HistoryDateList->Date[i])
       {
         uiVal1 = HistoryDateList->uiByteAr[i];
       }
     }
     if (TempDate <= EndDate) fprintf (HistFile," %u ",uiVal1);
   }
 }
 fprintf (HistFile,"\n");
 fclose (HistFile);
 return;
}


/***************************************************************************
 * ReadHistoryFile reads the history file from disk
 ***************************************************************************/
void ReadHistoryFile (ConfigFileType * pConfigOpts, StandardListType * pSList)
{
FILE * HistFile = NULL;
char HistoryFileName[FILENAME_MAX + 1];
time_t StartDate = 0, EndDate = 0, CurrentDate = 0, TempDate;
char *pszRecord, *pszTruncate, *pszToken;
int i, iDateNum;

 if (pConfigOpts->HistoryFile && ((strlen(pConfigOpts->OutPath) + 
     strlen(pConfigOpts->HistoryFile)) <= FILENAME_MAX))
 {
   sprintf (HistoryFileName,"%s%s",pConfigOpts->OutPath,pConfigOpts->HistoryFile);
   HistFile = fopen (HistoryFileName,"r");
 }
 if (HistFile)
 {
   pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));
   while (fgets(pszRecord, DEFAULT_RECORD_LENGTH, HistFile))
   {
     pszTruncate = StringToLower (Truncate_Config_File_Record (pszRecord));
     if (!*pszTruncate)
       ;
     else if (!StartDate)
     {
       if (!strncmp (pszTruncate, "time:", sizeof "time:" - 1))
       { /* Time: dd-mmm-yyyy - dd-mmm-yyyy */
         pszTruncate += sizeof "time:";
         pszToken = strstr (pszTruncate," - ");
         if (pszToken)
         {
           pszToken++; *pszToken = '\0'; pszToken++;
           pszTruncate = Truncate_String (pszTruncate);
           pszToken    = Truncate_String (pszToken);
         }
         if (strlen(pszToken) && strlen(pszTruncate))
         {
           StartDate = DateFromString_t (pszTruncate);
           EndDate   = DateFromString_t (pszToken);
           if (StartDate < EndDate)
           {
             pSList->HistoryDateList->DefaultExtend = 1 + (EndDate - StartDate)
                                                        /  SECONDS_OF_A_DAY;
             pSList->HistoryDateList->Offset = pSList->DateList->Offset;
           }
           else /* syntax error! */
           {
             StartDate = EndDate = 0;
           }
         }
       }
     }
     else if (!strncmp (pszTruncate, "week of:", sizeof "week of:" - 1))
     {
       pszToken = pszTruncate + sizeof "week of:";
       pszToken = Truncate_String (pszToken);
       CurrentDate = DateFromString_t (pszToken);
       if (CurrentDate > EndDate) CurrentDate = 0;
     }
     else if (CurrentDate && !strncmp (pszTruncate, "count:", sizeof "count:" - 1))
     {

       pszToken = pszTruncate + sizeof "count:";
       for (pszToken = strtok(pszToken, " "), i = 0; 
            pszToken && (i < 7); pszToken=strtok(NULL, " "), i++)
       {
         TempDate = CurrentDate + i * SECONDS_OF_A_DAY;
         iDateNum = DateListInsert (pSList->HistoryDateList, TempDate, 0);
         pSList->HistoryDateList->uiCountAr[iDateNum] = atoi (pszToken);
       }
     }
     else if (CurrentDate && !strncmp (pszTruncate, "bytes:", sizeof "bytes:" - 1))
     {
       pszToken = pszTruncate + sizeof "bytes:";
       for (pszToken = strtok(pszToken, " "), i = 0; 
            pszToken && (i < 7); pszToken=strtok(NULL, " "), i++)
       {
         TempDate = CurrentDate + i * SECONDS_OF_A_DAY;
         iDateNum = DateListInsert (pSList->HistoryDateList, TempDate, atoi(pszToken));
         pSList->HistoryDateList->uiCountAr[iDateNum]--;
       }
     }
   }
   fclose (HistFile);
 }
 return;
}

/***************************************************************************
 * StandardListsSetPeriod defines Count for all standard and
 * special lists.
 ***************************************************************************/
void StandardListsSetPeriod (StandardListType * pSList, int * Indices)
{
KeyType Key;
int i;


 /*
  * standard lists
  */
 for (i=0,Key=SITE; i < (KEYS_NUM_MEMBERS - 1); i++,Key++)
 {
   StringListType * pStringList = SelectStringList (pSList, Key, "");

   StringListAdd (pStringList, &Indices[1], Indices[0], 
                  pSList->DateList->NumMembers);
 }
 /*
  * special lists
  */
 for (i = 0; i < pSList->NumSpecialLists; i++)
 {
   StringListAdd (pSList->SpecialLists[i], &Indices[1], Indices[0],
                  pSList->DateList->NumMembers);
 }
 /*
  * weekday list
  */
 for (i = 0; i < 7; i++)
 {
   pSList->WeekDayList->List[i]->Count = 0;
 }
 if (pSList->HistoryDateList)
 {
   int Ind;

   if (Indices[0] < 0) /* all Members */
   {
     for (i = 0; i < pSList->HistoryDateList->NumMembers; i++)
     {
       Ind = (pSList->HistoryDateList->Date[i]/SECONDS_OF_A_DAY) % 7;
       Ind = (Ind + (7 - pSList->DateList->Offset)) % 7;
       pSList->WeekDayList->List[Ind]->Count += pSList->HistoryDateList->uiCountAr[i];
     }
   }
   else  /* just a period */
   {
     for (i = 1; i <= Indices[0]; i++)
     {
       Ind = (pSList->HistoryDateList->Date[Indices[i]]/SECONDS_OF_A_DAY) % 7;
       Ind = (Ind + (7 - pSList->DateList->Offset)) % 7;
       pSList->WeekDayList->List[Ind]->Count += 
               pSList->HistoryDateList->uiCountAr[Indices[i]];
     }
   }
 }
}



/***************************************************************************
 * NumValidStringListMembers returns the number of members of StringList
 * not matched in HideList.
 ***************************************************************************/
size_t NumValidStringListMembers (StringListType * pStringList,
                                  ConfigListType ** HideList,
                                  size_t * MaxValidValue)
{
size_t NumValidMembers = 0;
int i;

 *MaxValidValue = 0;
 for (i = 0; i < pStringList->NumMembers; i++)
 {
   if (CheckAgainstHideList (HideList, pStringList->Key, 
                     pStringList->List[i]->pszStr) &&
                     pStringList->List[i]->Count )
   {
     NumValidMembers++;
     if ( pStringList->List[i]->Count > *MaxValidValue) 
                          *MaxValidValue = pStringList->List[i]->Count;
   }
 }

 return NumValidMembers;
}


/***************************************************************************
 * NumberOfPeriods evaluates the number of days, weeks or month read
 * from the logfiles.
 ***************************************************************************/
int NumberOfPeriods (DateListType * DateList, enum DATE_MODE DateMode)
{
int NumPeriods = 0;
time_t FirstTime, LastTime;
struct tm *p_tm;
size_t StartPeriod, EndPeriod;

 if (DateList->NumMembers)
 {
   FirstTime = MinDate (DateList);
   LastTime  = MaxDate (DateList); 
   switch (DateMode)
   {
     case PER_DAY:
        NumPeriods = (LastTime - FirstTime) / SECONDS_OF_A_DAY + 1;
        break;
     case PER_WEEK:
        NumPeriods = (LastTime + DateList->Offset * SECONDS_OF_A_DAY)/SECONDS_OF_A_WEEK -
                    (FirstTime + DateList->Offset * SECONDS_OF_A_DAY)/SECONDS_OF_A_WEEK + 1;
        break;
     case PER_MONTH:
        p_tm = localtime (&FirstTime);
        StartPeriod = p_tm->tm_year * 12 + p_tm->tm_mon;
        p_tm = localtime (&LastTime);
        EndPeriod = p_tm->tm_year * 12 + p_tm->tm_mon;
        NumPeriods = EndPeriod - StartPeriod + 1;
        break;
     case NO_DATE:
        break;
   }
 }
 return NumPeriods;
}


/***************************************************************************
 *
 ***************************************************************************/
void GetStartEndPeriod (time_t CurrentDate, int NumPeriod, 
                       enum DATE_MODE DateMode, int WeekOffset,
                       time_t * StartPeriod, time_t * EndPeriod)
{
struct tm *p_tm;

 switch (DateMode)
 {
   case PER_DAY:
      *StartPeriod = *EndPeriod = CurrentDate + NumPeriod * SECONDS_OF_A_DAY;
      break;
   case PER_WEEK:
      *StartPeriod = (7  - WeekOffset) * SECONDS_OF_A_DAY + SECONDS_OF_A_WEEK * 
                    (NumPeriod -1 + (CurrentDate + WeekOffset  * 
                     SECONDS_OF_A_DAY) / SECONDS_OF_A_WEEK);
      *EndPeriod   = *StartPeriod + 6 * SECONDS_OF_A_DAY;
      break;
   case PER_MONTH:
      p_tm = localtime (&CurrentDate);
      p_tm->tm_mday = 1;
      p_tm->tm_sec = p_tm->tm_min = p_tm->tm_hour = 0;
      p_tm->tm_mon += NumPeriod;
      while (p_tm->tm_mon > 11)
      {
        p_tm->tm_mon -= 12;
        p_tm->tm_year++;
      }
      while (p_tm->tm_mon < 0)
      {
        p_tm->tm_mon += 12;
        p_tm->tm_year--;
      }
      *StartPeriod = mkDay (p_tm);
      if (p_tm->tm_mon == 11)
      {
        p_tm->tm_mon = 0;
        p_tm->tm_year++;
      }
      else
      {
        p_tm->tm_mon++;
      }
      *EndPeriod = mkDay (p_tm) - SECONDS_OF_A_DAY;
      break;
   case NO_DATE:
      break;
   }

 return;
}

/***************************************************************************
 * GetPeriodIndices stores the indices of the specified period in the 
 * array Indices. 
 * Remember: the dates in DateList aren't sorted, but the order is the
 * same as in counter arrays in the stringlists.
 * The number of matched indices is returned in Indices[0].
 * Return value is the start of the current period if MaxIndices is graeter
 * than null, else the end of the period.
 ***************************************************************************/
time_t GetPeriodIndices (int * Indices, size_t MaxIndices, size_t NumPeriod,
                         DateListType * DateList, enum DATE_MODE DateMode)
{
int NumIndices = 0;
int MaxPeriod = 0;
time_t FirstTime;
time_t StartPeriod = 0, EndPeriod = 0, ReturnTime = 0;
int i;

 for (i = 0; i < MaxIndices; i++) Indices[i] = 0;
 if (MaxIndices) MaxIndices--;
 FirstTime = MinDate (DateList);
 MaxPeriod = NumberOfPeriods (DateList, DateMode);
 if (NumPeriod && MaxPeriod >= NumPeriod)
 {
   NumPeriod--;
   GetStartEndPeriod (FirstTime, NumPeriod, 
                      DateMode, DateList->Offset,
                      &StartPeriod, &EndPeriod);
   if (MaxIndices)
   {
     for (i = 0; i < DateList->NumMembers && MaxIndices; i++)
     {
       if (DateList->Date[i] >= StartPeriod &&
           DateList->Date[i] <= EndPeriod)
       {
         Indices[++NumIndices] = i; MaxIndices--;
       }
     }
     Indices[0] = NumIndices;
     ReturnTime = StartPeriod;
   }
   else
   {
     ReturnTime = EndPeriod;
   }
 }
 return ReturnTime;
}


/***************************************************************************
 * ParseOutFileRecord search an string for insertions of the special syntax
 * "<!--Accstat Tag = Value -->". 
 * If found, the structure pTagInfo is filled with the appropriate values.
 * Return values:
 * TRUE, if the insertion is found, else FALSE.
 *
 * If "something" is somewhat meaningful it 
 * takes care for the specified actions.
 * Currently known are
 *   <!--Accstat Table = Section -->       !inserts a table
 ***************************************************************************/
int ParseOutFileRecord (char * pszString, TagInfoType * pTagInfo)
{
char * pszTag, * pszRest;
char Buffer[8];
int RetVal = FALSE;

 while ((pszTag = strstr(pszString, "<!--")))
 {
   strncpy (Buffer, &pszTag[4], 7);
   if (!strncmp(StringToLower (Buffer), "accstat", 7) && 
       (pszRest = strstr(pszTag, "-->")))
   {
     *pszTag = '\0';
     pszTag += 11;     /* <!--Accstat */
     *pszRest = '\0';
     pszRest += 3;     /* --> */
     pTagInfo->Tag = Truncate_String (pszTag);
     pTagInfo->Value = ParseConfigLine (pTagInfo->Tag);
     pTagInfo->Tag = StringToLower (pTagInfo->Tag);
     pTagInfo->RestOfRecord = pszRest;
     RetVal = TRUE;
     break;
   }
   else
   {
     pszString = pszTag + 4;  /* look for the next comment */
   }
 }
 return RetVal;
}

/***************************************************************************
 * TagPosition search for %...% or %...#n% tags
 ***************************************************************************/
char * TagPosition (char * String, size_t * PosLen, size_t * TagLen, 
                    unsigned int * FieldWidth)
{
char * pszTag = NULL, * pszNextTag = NULL, * pszValPos = NULL;

 *PosLen = *TagLen = *FieldWidth = 0;
 pszTag = strchr (String,'%');
 if (pszTag)
 {
   pszNextTag = strchr (pszTag+1,'%');
   if (pszNextTag)
   {
     *PosLen = pszNextTag - pszTag + 1;
     pszValPos = strchr (pszTag+1,'#');
     if (pszValPos && (pszValPos < pszNextTag))
     {
       *TagLen = pszValPos - pszTag;
       *FieldWidth = atoi(++pszValPos);
       if (*FieldWidth > (DEFAULT_RECORD_LENGTH / 10)) *FieldWidth = 0;
     }
     else
     {
       *TagLen = *PosLen - 1;
     }
   }
 }
  return pszTag;
}


/***************************************************************************
 * CheckTableTag searches a table list for the specified table and returns
 * a pointer to it if found.
 ***************************************************************************/
TableDescriptType * CheckTableTag (size_t NumTables, 
                                   TableDescriptType ** TableList, 
                                   char * pszTableName)
{
TableDescriptType * pTable = NULL;
int i;

 if (NumTables && TableList && pszTableName) /* what else can go wrong? */
 {
   for (i = 0; (i < NumTables) && !pTable; i++)
   {
     if (TableList[i] && !strcmp (TableList[i]->FileSection->Name, pszTableName))
     {
       pTable = TableList[i];
     }
   }
 }
 return pTable;
}


/***************************************************************************
 * CheckUrlFile checks, if pszFilename exists. If pszFilename contains
 * the PERIOD_REPLACEMENT placeholder, it is previously replaced with
 * the date string, which is 6 byte long. 
 ***************************************************************************/
char * CheckUrlFile (char * pszFilename, char * pszDate, char * pszOutPath)
{
 if (pszFilename)
 {
   FILE * DateFile;
   char * pszPos;
   char DateFileName[FILENAME_MAX + 1];

   if (pszDate && (pszPos = strstr(pszFilename, PERIOD_REPLACEMENT)))
   {
     int i = sizeof PERIOD_REPLACEMENT - 1;
     while (i--) *pszPos++ = *pszDate++;
   }
   if (pszOutPath && ((strlen(pszOutPath) + strlen(pszFilename)) <= FILENAME_MAX))
   {
     sprintf(DateFileName,"%s%s",pszOutPath,pszFilename);
     DateFile = fopen (DateFileName, "r");
     if (DateFile)
       fclose (DateFile);
     else
       pszFilename = NULL;
   }
 }
 return pszFilename;
}


/***************************************************************************
 * CheckUrlAlias checks, if pszString is a Member of AliasNameList. 
 * If true, the appropriate alias url is returned.
 * pszUrlPattern is a chain of NumUrlPattern zero terminated strings.
 ***************************************************************************/
char * CheckUrlAlias(char * pszString, char * pszUrlPattern, int NumUrlPattern,
                     char ** AliasNameList, char ** AliasUrlList)
{
char * pszUrl = NULL;
int i = 0;

 if (pszUrlPattern)
 {
   if (AliasNameList)
   {
     while (AliasNameList[i] && !pszUrl)
     {
       if (!strcmp(AliasNameList[i],pszString))
       {
         pszUrl = AliasUrlList[i];
       }
       i++;
     }
   }
   while (!pszUrl && NumUrlPattern--)
   {
     if (!wild_strcmp(pszString,pszUrlPattern)) pszUrl = pszString;
     if (NumUrlPattern)
     {
       pszUrlPattern += strlen (pszUrlPattern); while (!*pszUrlPattern)pszUrlPattern++;
     }
   }
 }
 return pszUrl;
}

/***************************************************************************
 * SetOutputFilename sets the outfile name and initializes a given period
 ***************************************************************************/
int SetOutputFilename (uInt * pNumPeriod, uInt * pNumFiles, char * FileName,
                       time_t * StartPeriod, time_t * EndPeriod,
                       OutputFileDescriptType * FileDescript,
                       StandardListType * pSList)
{
int Retval = TRUE;
int Indices[32];

 if (FileDescript->Period == NO_DATE)
 {
   *StartPeriod = MinDate (pSList->DateList);
   *EndPeriod   = MaxDate (pSList->DateList);
   Indices[0] = -1;
   if (*pNumPeriod)
     Retval = FALSE;
   else
     StandardListsSetPeriod (pSList, Indices);
   (*pNumPeriod)++; (*pNumFiles)++;
 }
 else
 {
   time_t FirstDate, LastDate;
   time_t LastPeriod;
   uInt uiPeriod;
   char * pszPos, * pszDate;

   uiPeriod = NumberOfPeriods (pSList->DateList, FileDescript->Period) - *pNumPeriod;
   FirstDate = MinDate (pSList->DateList);
   LastDate  = MaxDate (pSList->DateList);
   LastPeriod = GetPeriodIndices (Indices, 2, uiPeriod + 1, 
                                   pSList->DateList, FileDescript->Period);
   pszPos = strstr (FileName, MakeDateString(LastPeriod,
                    FileDescript->Period,pSList->DateList->Offset));
   *EndPeriod  = GetPeriodIndices (Indices, 0, uiPeriod, 
                                  pSList->DateList, FileDescript->Period);
   *StartPeriod = GetPeriodIndices (Indices, 32, uiPeriod, 
                                   pSList->DateList, FileDescript->Period);
   pszDate = MakeDateString(*StartPeriod,FileDescript->Period,pSList->DateList->Offset);
   if (pszPos)
   {
     int i = 6;
     while (i--) *pszPos++ = *pszDate++;
     if (!*StartPeriod || (*StartPeriod < FirstDate) || 
         (*pNumFiles >= FileDescript->MaxFiles)) Retval = FALSE;
   }
   else if (*pNumPeriod > 1)
   {
     Retval = FALSE;
   }

   if (Retval)
   {
     if (*EndPeriod > LastDate) /* check next period */
     {
       (*pNumPeriod)++;  
       Retval = SetOutputFilename (pNumPeriod, pNumFiles, FileName, StartPeriod, 
                                   EndPeriod, FileDescript, pSList);
     }
     else
     {
       FILE * outFile;

       (*pNumPeriod)++; (*pNumFiles)++;
       if (pszPos && (outFile = fopen (FileName, "r")))
       {               /* never overwrite periodical files (pszPos != NULL) */
         fclose (outFile);
         Retval = SetOutputFilename (pNumPeriod, pNumFiles, FileName,  StartPeriod,
                                     EndPeriod, FileDescript, pSList);
       }
       else
       {
         StandardListsSetPeriod (pSList, Indices);
         if (!pszPos)  (*pNumPeriod)++; /* write only one version */
       }
     }
   }
 }

 return Retval;
}


/***************************************************************************
 * CreateOutputFiles creates the output files specified in the 
 * configuration file.
 *
 ***************************************************************************/
void CreateOutputFiles  (StandardListType * pSList, 
                         CommandLineType  * pCmdLinOpts,
                         ConfigFileType   * pConfigOpts)
{
FILE * outfil;
char outfilname[FILENAME_MAX + 1];
char *pszRecord;            /* record from the configuration file */
char *pszDateReplacement;
uInt NumPeriod = 0, NumFiles = 0;
time_t StartPeriod, EndPeriod;
int Status = FALSE;
int i, j, ErrNum;


 if (!pConfigOpts || !pConfigOpts->NumOutFiles) return; /* basic checks */

 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));

 for (i = 0; pConfigOpts->OutFileList[i]; i++, NumPeriod = 0, NumFiles = 0)
 {
   if ((strlen(pConfigOpts->OutPath) + 
        strlen(pConfigOpts->OutFileList[i]->FileName)) <= FILENAME_MAX)
   {
     sprintf (outfilname,"%s%s",pConfigOpts->OutPath,
                                pConfigOpts->OutFileList[i]->FileName);
   }
   else
   {
     fprintf (stderr, OUTFIL_MAXLEN_ERROR, pConfigOpts->OutPath, 
                          pConfigOpts->OutFileList[i]->FileName,FILENAME_MAX);
     continue;   /* wrong filename */
   }
   pszDateReplacement = strstr (outfilname,PERIOD_REPLACEMENT);
   while (SetOutputFilename (&NumPeriod, &NumFiles, outfilname, &StartPeriod, 
                             &EndPeriod, pConfigOpts->OutFileList[i] ,pSList))
   {
     /* positions the config file*/
     Status = fseek (pCmdLinOpts->ConfigFile, 
                pConfigOpts->OutFileList[i]->FileSection->Offset, SEEK_SET);
     if (Status != 0)  continue;   /* skip this section */
     if (!(outfil = fopen (outfilname,"w")))
     {
       fprintf (stderr, OUTFIL_OPEN_ERROR, outfilname);
       continue;
     }
#if COUNT_ALLOC_MEMORY >= 2
     fprintf (stderr,"%s\n",outfilname);
#endif
     for (j = 0; j < pConfigOpts->OutFileList[i]->FileSection->Lenght; j++)
     {
       fgets (pszRecord, DEFAULT_RECORD_LENGTH, pCmdLinOpts->ConfigFile);
       ErrNum = PrintOutfilRecord (outfil, pszRecord, StartPeriod, EndPeriod,
                                        pSList, pCmdLinOpts, pConfigOpts);
       if (ErrNum != SUCCESS)
       {
         if (ErrNum != ERR_UNKNOWNTAG)
         {
           fprintf (stderr, OUTFIL_SYNTAX_ERROR, j+1 , 
                  pConfigOpts->OutFileList[i]->FileName);
         }
         if (ErrNum == ERR_NOSECTION)
           fprintf (stderr, CONFIG_ERROR_NOSECTION);
         else if (ErrNum == ERR_NOTABLE)
           fprintf (stderr, OUTFIL_TABLE_NOSPEC);
       }
     }
     fclose (outfil);
   }
#define MAX_FILE_ERR 5
   if (pszDateReplacement && pConfigOpts->OutFileList[i]->MaxFiles) /* remove old files */
   {
     time_t StartDate, EndDate, CurrentDate = MaxDate (pSList->DateList);
     int NumFileErr;
     char * pszDate;

     for (j = -pConfigOpts->OutFileList[i]->MaxFiles, NumFileErr = 0;
          NumFileErr < MAX_FILE_ERR;j--)
     {
       GetStartEndPeriod (CurrentDate, j, pConfigOpts->OutFileList[i]->Period,
                          pSList->DateList->Offset,
                          &StartDate, &EndDate);
       pszDate = MakeDateString(StartDate,pConfigOpts->OutFileList[i]->Period,
                                  pSList->DateList->Offset);
       memcpy (pszDateReplacement, pszDate, sizeof PERIOD_REPLACEMENT - 1);
       if (remove(outfilname))
       {
         NumFileErr++;
       }
       else
       {
         fprintf (stderr, "removing %s\n", outfilname);
       }
       if ((NumFileErr == MAX_FILE_ERR) && (StartDate > MinDate (pSList->DateList)))
       {
         CurrentDate =  MinDate (pSList->DateList);
         NumFileErr = 0;
         j = 0;
       }
     }
   }
 }

 free (pszRecord);
 if (pCmdLinOpts->ConfigFile) 
 {
   fclose (pCmdLinOpts->ConfigFile);
   pCmdLinOpts->ConfigFile = NULL;
 }
 return;
}


/***************************************************************************
 *
 * Supported tags are:
 * table           - inserts a table
 * summeryreport   - writes a summery report without html formatting
 * version         - prints the version number of accstat
 * date            - prints the current date
 * beginperiod     - prints the first date of currend period
 * endperiod       - prints the last date of currend period
 * totalrecords    - total number of logfile records
 * matchingrecords - valid number of logfile records
 * nummembers      - total number of list members
 * validmembers    - valid number of list members
 * countmembers    - number of matching members in stringlist
 ***************************************************************************/
int PrintOutfilRecord (FILE * outfil, char * pszRecord,
                        time_t StartPeriod,  time_t EndPeriod,
                        StandardListType * pSList, 
                        CommandLineType  * pCmdLinOpts,
                        ConfigFileType   * pConfigOpts)
{
char * pszLine = pszRecord;
TagInfoType TagInfo;
int ErrorNumber = SUCCESS;
char buf[80];
int i;

 while (ParseOutFileRecord (pszLine, &TagInfo))
 {
   fputs (pszLine, outfil);
   if (!strcmp(TagInfo.Tag,"summeryreport"))           /* SummeryReport */
   {
     SummeryReportPrint (outfil, pSList, pCmdLinOpts, StartPeriod, EndPeriod);
   }
   else if (!strcmp(TagInfo.Tag,"version"))            /* Version number */
   {
     fputs (ACCSTAT_VERSION, outfil);
   }
   else if (!strcmp(TagInfo.Tag,"date"))               /* current date */
   {
     PrintDate (buf, 80, time(NULL), TagInfo.Value, pSList->DateList->Offset);
     fputs (buf, outfil);
   }
   else if (!strcmp(TagInfo.Tag,"beginperiod"))        /* BeginPeriod */
   {
     PrintDate (buf, 80, StartPeriod, TagInfo.Value, pSList->DateList->Offset);
     fputs (buf, outfil);
   }
   else if (!strcmp(TagInfo.Tag,"endperiod"))          /* EndPeriod */
   {
     PrintDate (buf, 80, EndPeriod, TagInfo.Value, pSList->DateList->Offset);
     fputs (buf, outfil);
   }
   else if (!strcmp(TagInfo.Tag,"table"))              /* Table */
   {
     if (TagInfo.Value)
     {
       TableDescriptType * pTable = CheckTableTag (pConfigOpts->NumTables,
                                    pConfigOpts->TableList, TagInfo.Value);
       if (pTable)
       {
         long int Offset = ftell (pCmdLinOpts->ConfigFile);
         InsertTable (outfil, StartPeriod, EndPeriod, &TagInfo, pTable,
                      pSList, pCmdLinOpts, pConfigOpts);
         fseek (pCmdLinOpts->ConfigFile, Offset, SEEK_SET);
       }
       else
         ErrorNumber = ERR_NOSECTION;
     }
     else
       ErrorNumber = ERR_NOTABLE;
   }
   else if (!strcmp(TagInfo.Tag,"totalrecords"))          /* TotalRecords */
   {
     fprintf (outfil,"%u",pSList->NumRecord);
   }
   else if (!strcmp(TagInfo.Tag,"matchingrecords"))       /* MatchingRecords */
   {
     fprintf (outfil,"%lu",StringListTotalCount (pSList->StatusList));
   }
   else if (!strcmp(TagInfo.Tag,"nummembers"))       /* NumMembers */
   {
     uInt NumMembers = 0;
     StringListType * StringList;

     if (TagInfo.Value)
     {
       if (!strcmp (TagInfo.Value,"date"))
       {
         NumMembers = (EndPeriod - StartPeriod) / SECONDS_OF_A_DAY + 1;
       }
       else if ((StringList = SelectStringList (pSList, NO_KEY, TagInfo.Value)))
       {
         for (i = 0; i < StringList->NumMembers; i++) 
         {
           if (StringList->List[i]->Count) NumMembers++;
         }
       }
       fprintf (outfil,"%u",NumMembers);
     }
     else
       ErrorNumber = ERR_INVALIDLIST;
   }
   else if (!strcmp(TagInfo.Tag,"validmembers"))       /* ValidMembers */
   {
     uInt ValidMembers = 0;
     StringListType * StringList;

     if (TagInfo.Value)
     {
       if (!strcmp (TagInfo.Value,"date"))
       {
         ValidMembers = (EndPeriod - StartPeriod) / SECONDS_OF_A_DAY + 1;
       }
       else if ((StringList = SelectStringList (pSList, NO_KEY, TagInfo.Value)))
       {
         size_t MaxCount;
         ValidMembers = NumValidStringListMembers 
                               (StringList, pCmdLinOpts->HideList, &MaxCount);
       }
       fprintf (outfil,"%u",ValidMembers);
     }
     else
     {
       ErrorNumber = ERR_INVALIDLIST;
     }
   }
   else if (!strcmp(TagInfo.Tag,"countmembers"))   /* CountMembers */
   {
     char * pszMask = ParseConfigLine (TagInfo.Value);
     StringListType * StringList = SelectStringList (pSList, NO_KEY, TagInfo.Value);
     uLong Count = CountStringListMembers (StringList, pszMask);

     fprintf (outfil,"%lu",Count);
   }
   else if (!strcmp(TagInfo.Tag,"transferredbytes"))   /* TransferredBytes */
   {
     ByteCountType NumBytes = {0,0};
     for (i = 0; i < pSList->HistoryDateList->NumMembers; i++)
     {
       if ((pSList->HistoryDateList->Date[i] >= StartPeriod) &&
           (pSList->HistoryDateList->Date[i] <= EndPeriod))
       {
         ByteCountAdd (&NumBytes,pSList->DateList->uiByteAr[i],0);
       }
     }
     fputs (InsertThousandsSeparator(NumBytes.low, NumBytes.high), outfil);
   }
   else
   {
     ErrorNumber = ERR_UNKNOWNTAG;
     fprintf (stderr, OUTFILE_UNKNOWN_TAG, TagInfo.Tag, 
                      (TagInfo.Value) ? TagInfo.Value : "");
   }
   pszLine = TagInfo.RestOfRecord;
 }
 fputs (pszLine, outfil);
 return ErrorNumber;
}

/***************************************************************************
 * InsertTable inserts a table into an output file
 * It reads the appropriate section from the configuration file and fills
 * the structure TableInfo with the informations found there.
 * supported tags are:
 *  Sorting = Numerical or Alphabetical   (only first char is significant)
 *  ShowMembers = n                       (max. number of table rows)
 *  Layout = simple                       (simple or named section)
 *  Width  = 44                           (max. number of table columns)
 *  Period = daily    (daily, weekly or monthly, only first char is significant)
 *
 ***************************************************************************/
void InsertTable (FILE * outfil, time_t StartPeriod, time_t EndPeriod,
                  TagInfoType       * pTagInfo,
                  TableDescriptType * pTable,
                  StandardListType  * pSList, 
                  CommandLineType   * pCmdLinOpts,
                  ConfigFileType    * pConfigOpts)

{
static int isrecursion = FALSE;
FILE * ConfigFile = pCmdLinOpts->ConfigFile;
SectionDescriptType ** LayoutList = pConfigOpts->LayoutList;
ConfigListType ** HideList = pCmdLinOpts->HideList;
size_t NumLayoutLists = pConfigOpts->NumLayoutLists;

struct { char * Layout; 
         char * UrlPattern;
         int    NumUrlPattern;
         SortType SortType; 
         int    iShowMembers;        /* max. number to show, -1 = all */
         uInt   Width;               /* max. width of the table */
         enum DATE_MODE DateMode;    /* period*/
         enum TABLE_TYPE TableType;  /* kind of table */
       } TableInfo = {NULL, NULL, 0, SORT_NUMERICAL, 5, 76, NO_DATE, UNKNOWN_TABLE};

char *pszRecord;            /* record from the configuration file */
char * pszCondition, * pszValue;
int i, j, Index; 


 /* recursive calls to InsertTable may cause a endless loop */
 if (isrecursion) return;
 isrecursion = TRUE;
 fseek (ConfigFile, pTable->FileSection->Offset, SEEK_SET);
 pszRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));

 /*
  * determine type of table
  */
  if (pTable->SpecialList && !strcmp (pTable->SpecialList,"date"))
  {
    TableInfo.TableType = DATELIST_TABLE;
  }
  else
  {
    TableInfo.TableType = STRINGLIST_TABLE;
  }

/*
 * each line in the table definition section should have the syntax
 * Condition = Value
 */
 for (i = 0; i < pTable->FileSection->Lenght; i++)
 {
   fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
   pszCondition = Truncate_String (pszRecord);
   pszValue     = ParseConfigLine (pszCondition);
   pszCondition = StringToLower (pszCondition);
   if (!strcmp (pszCondition,"clickable") && !TableInfo.UrlPattern) /* Clickable */
   {
     char * pszTemp;
     TableInfo.NumUrlPattern = 1;
     if (!pszValue) pszValue = DEFAULT_CLICKABLE_LIST;
     /* allocate additional bytes for datelist placeholder  and \0 */
     j = 1 + sizeof PERIOD_REPLACEMENT - sizeof PERIOD_PLACEHOLDER;
     TableInfo.UrlPattern = my_malloc ((strlen(pszValue) + j) * sizeof(char));
     strcpy (TableInfo.UrlPattern, pszValue);
     pszTemp = TableInfo.UrlPattern;
     if (!pCmdLinOpts->iCaseSensitive) pszTemp = StringToLower (pszTemp);
     while ((pszTemp = strchr(pszTemp,',')))
     {
       while ((*pszTemp == ',') || (*pszTemp == ' ')) 
       {
         *pszTemp = '\0';
         pszTemp++;
       }
       if (*pszTemp)
       {
        TableInfo.NumUrlPattern++;
        pszTemp = Truncate_String (pszTemp);
       }
     }
   }
   else if (pszValue)
   {
     if (!strcmp (pszCondition,"layout"))                      /* Layout */
     {
       if (!strcmp (pszValue,"simple")) /* internal layout format */
       {
         TableInfo.Layout = "simple";
       }
       else if ((Index = ConfigFileSeek(NULL, LayoutList, NumLayoutLists, pszValue)))
       {
         Index--;
         TableInfo.Layout = LayoutList[Index]->Name;
       }
       else
       {                            /* section not found */
         fprintf (stderr, DEFINETABLE_SYNTAX_ERROR, pTagInfo->Value);
         fprintf (stderr, DEFINETABLE_VALUE_CHECK, pszCondition, pszValue);
         fprintf (stderr, CONFIG_ERROR_NOSECTION);
       }
     }
     else if (!strcmp (pszCondition,"sorting"))             /* Sorting */
     {
       pszValue = StringToLower (pszValue);
       if (pszValue[0] == 'a')
       {
         TableInfo.SortType = SORT_ALPHABETICAL;
       }
       else if (pszValue[0] == 'n')
       {
         TableInfo.SortType = SORT_NUMERICAL;
       }
       else if (pszValue[0] == 'u')
       {
         TableInfo.SortType = SORT_NOSORT;
       }
     }
     else if (!strcmp (pszCondition,"period"))              /* Period */
     {
       pszValue = StringToLower (pszValue);
       if (pszValue[0] == 'm')
       {
         TableInfo.DateMode = PER_MONTH;
       }
       else if (pszValue[0] == 'w')
       {
         TableInfo.DateMode = PER_WEEK;
       }
       else
       {
         TableInfo.DateMode = PER_DAY;
       }
     }
     else if (!strcmp (pszCondition,"width"))             /* Width */
     {
         TableInfo.Width = atoi (pszValue);
     }
     else if (!strcmp (pszCondition,"showmembers"))         /* ShowMembers */
     {
       if ((pszValue[0] == 'A') || (pszValue[0] == 'a'))
       {
         TableInfo.iShowMembers = -1;
       }
       else
       {
         TableInfo.iShowMembers = abs (atoi (pszValue));
       }
     }
     else
     {
       printf ("unknown condition \"%s = %s\"\n",pszCondition,pszValue);
     }
   }
 }
 /*
  *  ===================================
  *        now print the table
  *  ===================================
  */
 if (!TableInfo.Layout)
 {
   fprintf (stderr, DEFINETABLE_SYNTAX_ERROR, pTagInfo->Value);
   fprintf (stderr, DEFINETABLE_NOLAYOUT);
 }
 else if (!strcmp(TableInfo.Layout,"simple"))
 {
   if (TableInfo.TableType == DATELIST_TABLE)
   {
     uInt NumRows;
     if (TableInfo.iShowMembers == -1)  NumRows = 0;
     else  NumRows = TableInfo.iShowMembers;
     DateListPrint   (outfil, pSList->DateList, NumRows, TableInfo.Width, 
                              TableInfo.DateMode);
   }
   else
   {
     StringListType * StringList;
     
     StringList = SelectStringList (pSList, pTable->FileSection->Key, 
                                            pTable->SpecialList);
     StringList->Sort = TableInfo.SortType;
     StringList->ShowMembers = TableInfo.iShowMembers;
     StringListPrint (outfil, StringList, TableInfo.Width, HideList);
   }
 }
 else
 {
   /*
    * =============================
    *    user defined layout
    * =============================
    */
   size_t NumLines;
   char * pszOutRecord = my_malloc ( DEFAULT_RECORD_LENGTH * sizeof(char));

   if ((NumLines = ConfigFileSeek(ConfigFile, LayoutList, NumLayoutLists,
                                  TableInfo.Layout)))
                                
   {
     size_t NumSections = 0;
     StringListType * StringList;
     SectionDescriptType ** pzSections = NULL;
     char ** RowLines = NULL;  /* lines in section row */
     TableHeadInfoType tbl;
     TableRowInfoType  tbl_row = {"", NULL, 0, 0, {0, 0}};

     /* search for head, row and foot section */
     pzSections = CheckConfigFileSyntax (ConfigFile, &NumSections, NumLines);

     /*
      * =============================
      *          StringList
      * =============================
      */
     if (TableInfo.TableType == STRINGLIST_TABLE)
     {
       StringList = SelectStringList (pSList, pTable->FileSection->Key, 
                                            pTable->SpecialList);
       StringList = InitTableInfo (&tbl, pSList, HideList, 
                                   pTable->SpecialList, pTable->FileSection->Key,
                                   TableInfo.iShowMembers, NO_DATE);
       if (TableInfo.SortType == SORT_ALPHABETICAL)
       {
         qsort (StringList->List , StringList->NumMembers, 
                sizeof (void *), ItemListStrComp);
       }
       else if (TableInfo.SortType == SORT_NUMERICAL)
       {
         qsort (StringList->List , StringList->NumMembers, 
                sizeof (void *), StringListCountComp);
       }
       /*
        * head
        */
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,"Head");
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         PrintTableLine (&tbl, &tbl_row, pszOutRecord, pszRecord); 
         PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                            pSList, pCmdLinOpts, pConfigOpts);
       }
       /*
        * row
        */
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,
                                  "Row");
       if (NumLines)
       {
         RowLines = my_malloc (NumLines * sizeof (char *));
       }
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         RowLines[i] = my_malloc ((strlen(pszRecord) + 1) * sizeof (char));
         strcpy (RowLines[i],pszRecord);
       }
       tbl_row.Current = 0;
       for (i = 0; (tbl_row.Current < tbl.ShowMembers) &&
                   (i < StringList->NumMembers); i++)
       {
        if (CheckAgainstHideList (HideList, StringList->Key, 
                                            StringList->List[i]->pszStr) &&  
                                            StringList->List[i]->Count)
        {
         tbl_row.Current++;
         if (StringList->Key == STATUS)
         {
           tbl_row.Item = GetStatusCode (StringList->List[i]->pszStr);
           tbl_row.Url  = NULL;
         }
         else
         {
           tbl_row.Item = StringList->List[i]->pszStr;
           tbl_row.Url  = CheckUrlAlias(tbl_row.Item, TableInfo.UrlPattern,
                                        TableInfo.NumUrlPattern,
                                        pCmdLinOpts->AliasNameList, 
                                        pCmdLinOpts->AliasUrlList);
         }
         tbl_row.CurrentCount = StringList->List[i]->Count;
         for (j = 0; j < NumLines; j++)
         {
           PrintTableLine (&tbl, &tbl_row, pszOutRecord, RowLines[j]); 
           PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                              pSList, pCmdLinOpts, pConfigOpts);
         }
        }
       }

       for (i = 0; i < NumLines; i++) free (RowLines[i]);
       free (RowLines);
       /*
        * foot
        */
       tbl_row.CurrentCount = 0;
       tbl_row.Item = "";
       tbl_row.Url = NULL;
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,"Foot");
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         PrintTableLine (&tbl, &tbl_row, pszOutRecord, pszRecord); 
         PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                            pSList, pCmdLinOpts, pConfigOpts);
       }
     }
     /*
      * =============================
      *          DateList
      * =============================
      */
     else if (TableInfo.TableType == DATELIST_TABLE && TableInfo.DateMode != NO_DATE)
     {
       DateListType * DateList;
       int Indices[32], MaxIndices;
       char szFormat[32];
       time_t CurrentPeriod = 0;
       char * pszPos, * pszCurrentPeriod = NULL;
       int NumTableMembers;

       /* change file_!nn!.ext to file_!nnnn!.ext */
       if (TableInfo.UrlPattern && 
          (pszPos = strstr (TableInfo.UrlPattern, PERIOD_PLACEHOLDER)))
       {                              /* restore TableInfo.UrlPattern */
         memmove (pszPos + (sizeof PERIOD_REPLACEMENT - sizeof PERIOD_PLACEHOLDER),
                  pszPos, strlen(pszPos) + 1);
         memcpy  (pszPos, PERIOD_REPLACEMENT, sizeof PERIOD_REPLACEMENT - 1);
       }

       DateList = InitTableInfo (&tbl, pSList, HideList, 
                                 pTable->SpecialList, pTable->FileSection->Key,
                                 TableInfo.iShowMembers, TableInfo.DateMode);

       if (TableInfo.DateMode == PER_MONTH)     MaxIndices = 32;
       else if (TableInfo.DateMode == PER_WEEK) MaxIndices = 8;
       else                                     MaxIndices = 2;
       /*
        * head
        */
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,"Head");
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         PrintTableLine (&tbl, &tbl_row, pszOutRecord, pszRecord); 
         PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                            pSList, pCmdLinOpts, pConfigOpts);
       }
       /*
        * row
        */
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,
                                  "Row");
       if (NumLines)
       {
         RowLines = my_malloc (NumLines * sizeof (char *));
       }
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         RowLines[i] = my_malloc ((strlen(pszRecord) + 1) * sizeof (char));
         strcpy (RowLines[i],pszRecord);
       }

       tbl_row.Current = 0;
       tbl_row.Item = szFormat;
       /* adjust date table for the current period */
       NumTableMembers = tbl.NumMembers;
       CurrentPeriod = MaxDate (DateList);
       for (i = tbl.NumMembers; i && (CurrentPeriod > EndPeriod); i--)
       {
         CurrentPeriod =  GetPeriodIndices (Indices, 1, i, 
                                    DateList, TableInfo.DateMode);
         if (CurrentPeriod > EndPeriod) NumTableMembers--;
       }
       for (i = (tbl.ShowMembers && (NumTableMembers > tbl.ShowMembers)) ?
                 NumTableMembers - tbl.ShowMembers : 0 ; 
            i < NumTableMembers; i++)
       {             /* evaluate startup values */
         CurrentPeriod =  GetPeriodIndices (Indices, MaxIndices, i + 1, 
                                    DateList, TableInfo.DateMode);
         EvalDateListPeriod (&Indices[1], Indices[0], DateList,
                             &tbl_row.CurrentCount, &tbl_row.CurrentBytes);
         tbl_row.Current++;
         strftime (szFormat, 30, tbl.DateFormat, localtime(&CurrentPeriod));
         pszCurrentPeriod = MakeDateString (CurrentPeriod, TableInfo.DateMode, 
                                            DateList->Offset);
         tbl_row.Url  = CheckUrlFile (TableInfo.UrlPattern, pszCurrentPeriod, 
                                      pConfigOpts->OutPath);
         for (j = 0; j < NumLines; j++)
         {
           PrintTableLine (&tbl, &tbl_row, pszOutRecord, RowLines[j]); 
           PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                              pSList, pCmdLinOpts, pConfigOpts);
         }
         if (TableInfo.UrlPattern && 
            (pszPos = strstr (TableInfo.UrlPattern, pszCurrentPeriod)))
         {                              /* restore TableInfo.UrlPattern */
           memcpy (pszPos, PERIOD_REPLACEMENT, sizeof PERIOD_REPLACEMENT - 1);
         }
       }

       for (i = 0; i < NumLines; i++) free (RowLines[i]);
       free (RowLines);
       /*
        * foot
        */
       tbl_row.CurrentCount = 0;
       tbl_row.Item = "";
       tbl_row.Url = NULL;
       NumLines = ConfigFileSeek(ConfigFile, pzSections, NumSections,"Foot");
       for (i = 0; i < NumLines; i++)
       {
         fgets (pszRecord, DEFAULT_RECORD_LENGTH, ConfigFile);
         PrintTableLine (&tbl, &tbl_row, pszOutRecord, pszRecord); 
         PrintOutfilRecord (outfil, pszOutRecord, StartPeriod, EndPeriod,
                            pSList, pCmdLinOpts, pConfigOpts);
       }
     }
     free (pszOutRecord);
     FreeSectionDescriptTypeList (pzSections, NumSections);
   }
 }
 free (pszRecord);
 if (TableInfo.UrlPattern) free (TableInfo.UrlPattern);
 isrecursion = FALSE;
 return;
}


/***************************************************************************
 * InitTableInfo fills the structure Table of type TableHeadInfoType.
 * It contains some static infos about a datelist or a stringlist.
 ***************************************************************************/
void * InitTableInfo (TableHeadInfoType * Table,
                      StandardListType  * pSList,
                      ConfigListType   ** HideList,
                      char * pszSpecialList, KeyType Key,
                      int ShowMembers,
                      enum DATE_MODE DateMode)
{
void * ReturnList = NULL;
int i;

 Table->TotalCount   = 0;
 Table->MaxCount     = 0;
 Table->MaxBytes.low   = Table->MaxBytes.high   = 0;
 Table->TotalBytes.low = Table->TotalBytes.high = 0;
 Table->NumMembers   = 0;
 Table->TotalCount   = 0;
 /*
  * =============================
  *   determine type of table
  * =============================
  */
 if (pszSpecialList && !strcmp (pszSpecialList,"date"))
 {
   Table->Type = DATELIST_TABLE;
 }
 else
 {
   Table->Type = STRINGLIST_TABLE;
 }

 if (Table->Type ==  STRINGLIST_TABLE)
 {
   StringListType * StringList = NULL;

   StringList = SelectStringList (pSList, Key, 
                                          pszSpecialList);
   Table->Name         = StringList->Name;
   Table->DateFormat   = NULL;
   for (i = 0; i < StringList->NumMembers; i++) 
   {
     Table->TotalCount += StringList->List[i]->Count;
     if (StringList->List[i]->Count) Table->NumMembers++;
   }
   Table->ValidMembers = NumValidStringListMembers (StringList, HideList, 
                                               (size_t *) &Table->MaxCount);
   if (!ShowMembers) ShowMembers = Table->ValidMembers;
   Table->ShowMembers = (ShowMembers < Table->ValidMembers) ? 
                           ShowMembers : Table->ValidMembers;
   ReturnList = StringList;
 }
 else if (Table->Type ==  DATELIST_TABLE)
 {
   DateListType * DateList = pSList->HistoryDateList;
   int Indices[32], MaxIndices;
   uInt   CurrentCount  = 0;
   ByteCountType CurrentBytes = {0, 0};
   int i;

   switch (DateMode)
   {
     default:
     case PER_DAY:
        MaxIndices = 2;
        Table->Name = "Daily Report";
        Table->DateFormat = SUMMARY_TIME_FORMAT;
        break;
     case PER_WEEK:
        MaxIndices = 8;
        Table->Name = "Weekly Report";
        Table->DateFormat = SUMMARY_TIME_FORMAT;
        break;
     case PER_MONTH:
        MaxIndices = 32;
        Table->Name = "Monthly Report";
        Table->DateFormat = "%B %Y";
        break;
   }
   Table->ValidMembers = Table->NumMembers = NumberOfPeriods (DateList, DateMode);
   if (!ShowMembers) ShowMembers = Table->ValidMembers;
   Table->ShowMembers = (ShowMembers < Table->ValidMembers) ? 
                         ShowMembers : Table->ValidMembers;

   for (i = (ShowMembers && Table->NumMembers > ShowMembers) ? 
             Table->NumMembers - ShowMembers : 0 ; 
        i < Table->NumMembers; i++)
   {             /* evaluate startup values */
     GetPeriodIndices (Indices, MaxIndices, i + 1, DateList, DateMode);
     EvalDateListPeriod (&Indices[1], Indices[0], DateList,
                         &CurrentCount, &CurrentBytes);
     Table->TotalCount  += CurrentCount;
     if (Table->MaxCount < CurrentCount)  
         Table->MaxCount = CurrentCount;
     ByteCountAdd (&Table->TotalBytes, CurrentBytes.low, CurrentBytes.high);
     ByteCountMax (&Table->MaxBytes, &CurrentBytes);
   }

   ReturnList = DateList;
 }
 return ReturnList;
}

/***************************************************************************
 *  valid Tags are %num%, %num_members%, %tablename%, %count%, %item%
 *  %tablename%, %show_members%, %num_members%
 *  in a table format section the following variable are known:
 *  %tablename%         name of the table
 *  %show_members%      number of table rows
 *  %num_members%       number of list members
 *  %valid_members%     number of list members not matched by hidelist
 *  %hidden_members%    number of list members matched by hidelist
 *  %num%               row number
 *  %count%             hit count
 *  %max_count%         max. hit count
 *  %proz_count%        percental hit count
 *  %total_count%       total hit count
 *  %bytes%             transfered bytes
 *  %proz_bytes%        percental transfered bytes
 *  %item%              current item
 ***************************************************************************/
void PrintTableLine (TableHeadInfoType * tblHead, TableRowInfoType * tblRow, 
                     char * outLine, char * inLine)
{
 char * pszTag;
 size_t PosLen, TagLen;
 unsigned int FieldWidth;
 double dTemp;
 char szFormat[32];
 int iLen;

 pszTag = inLine;
 outLine[0] = '\0';
 while ((pszTag = TagPosition (inLine, &PosLen, &TagLen, &FieldWidth)) && 
         (strlen(outLine) < (DEFAULT_RECORD_LENGTH * 0.8)))
 {
   strncat (outLine,inLine,pszTag - inLine);
   if (TagLen == 1)
   {
     sprintf (outLine,"%s%s",outLine, "%");
   }
   else if (!strncmp (pszTag, "%tablename%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%.%us",FieldWidth);
     else
       strcpy (szFormat,"%s%s");
     sprintf (outLine,szFormat,outLine, tblHead->Name);
   }
   else if (!strncmp (pszTag, "%item%", TagLen))
   {
     if (!FieldWidth) FieldWidth = strlen (tblRow->Item);
     if (tblRow->Url)
     {
       sprintf (szFormat,"%%s<a href=\"%%s\">%%.%us</a>",FieldWidth);
       sprintf (outLine,szFormat,outLine, tblRow->Url, tblRow->Item);
     }
     else
     {
       sprintf (szFormat,"%%s%%.%us",FieldWidth);
       sprintf (outLine,szFormat,outLine, tblRow->Item);
     }
   }
   else if (!strncmp (pszTag, "%show_members%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%.%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, (uInt)tblHead->ShowMembers);
   }
   else if (!strncmp (pszTag, "%num%", TagLen))
   {
     iLen = nDigits(tblHead->ShowMembers);
     FieldWidth = (FieldWidth > iLen) ? FieldWidth : iLen;
     sprintf (szFormat,"%%s%%%uu", FieldWidth);
     sprintf (outLine,szFormat,outLine, (uInt)tblRow->Current);
   }
   else if (!strncmp (pszTag, "%num_members%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%.%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, (uInt)tblHead->NumMembers);
   }
   else if (!strncmp (pszTag, "%valid_members%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%.%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, (uInt)tblHead->ValidMembers);
   }
   else if (!strncmp (pszTag, "%hidden_members%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%.%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, (uInt)(tblHead->NumMembers - 
                                             tblHead->ValidMembers));
   }
   else if (!strncmp (pszTag, "%count%", TagLen))
   {
     iLen = nDigits(tblHead->MaxCount);
     FieldWidth = (FieldWidth > iLen) ? FieldWidth : iLen;
     sprintf (szFormat,"%%s%%%uu", FieldWidth);
     sprintf (outLine,szFormat,outLine, tblRow->CurrentCount);
   }
   else if (!strncmp (pszTag, "%max_count%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, tblHead->MaxCount);
   }
   else if (!strncmp (pszTag, "%total_count%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%%uu",FieldWidth);
     else
       strcpy (szFormat,"%s%u");
     sprintf (outLine,szFormat,outLine, tblHead->TotalCount);
   }
   else if (!strncmp (pszTag, "%bytes%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%%u.%us",FieldWidth,FieldWidth);
     else
       strcpy (szFormat,"%s%s");
     sprintf (outLine,szFormat,outLine, InsertThousandsSeparator
             (tblRow->CurrentBytes.low,tblRow->CurrentBytes.high));
   }
   else if (!strncmp (pszTag, "%max_bytes%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%%u.%us",FieldWidth,FieldWidth);
     else
       strcpy (szFormat,"%s%s");
     sprintf (outLine,szFormat,outLine, InsertThousandsSeparator
             (tblHead->MaxBytes.low,tblHead->MaxBytes.high));
   }
   else if (!strncmp (pszTag, "%total_bytes%", TagLen))
   {
     if (FieldWidth)
       sprintf (szFormat,"%%s%%%u.%us",FieldWidth,FieldWidth);
     else
       strcpy (szFormat,"%s%s");
     sprintf (outLine,szFormat,outLine, InsertThousandsSeparator
             (tblHead->TotalBytes.low,tblHead->TotalBytes.high));
   }
   else if (!strncmp (pszTag, "%proz_count%", TagLen))
   {
     if (tblHead->TotalCount)
     {
       dTemp = 100.0 * ((double)tblHead->MaxCount / tblHead->TotalCount);
       iLen =  2 + nDigits((int) dTemp);
       dTemp = 100.0 * ((double)tblRow->CurrentCount / tblHead->TotalCount);
     }
     else
     {
       dTemp = 0.0;
       iLen = 3;
     }
     FieldWidth = (FieldWidth > iLen) ? FieldWidth : iLen;
     sprintf (szFormat,"%%s%%%u.1f", FieldWidth );
     sprintf (outLine,szFormat,outLine, dTemp);
   }
   else if (!strncmp (pszTag, "%proz_bytes%", TagLen))
   {
     if (tblHead->TotalBytes.low || tblHead->TotalBytes.high)
     {
       dTemp = ByteCountProz (&tblHead->MaxBytes, &tblHead->TotalBytes);
       iLen =  2 + nDigits((int) dTemp);
       dTemp = ByteCountProz (&tblRow->CurrentBytes, &tblHead->TotalBytes);
     }
     else
     {
       dTemp = 0.0;
       iLen = 3;
     }
     FieldWidth = (FieldWidth > iLen) ? FieldWidth : iLen;
     sprintf (szFormat,"%%s%%%u.1f", FieldWidth );
     sprintf (outLine,szFormat,outLine, dTemp);
   }
   else
   {
     strncat (outLine,pszTag,PosLen);
   }
   inLine = pszTag + PosLen;
 }
 strcat (outLine, inLine);
 return;
}


